@@ -5,7 +5,7 @@
      Sundance ST201 "Alta"
    This file also contains the chip-independent diagnostic code.
 
-	Copyright 1998-2000 by Donald Becker.
+	Copyright 1998-2002 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -20,6 +20,7 @@
 
 	Support and updates available at
 	http://www.scyld.com/diag/index.html
+	http://scyld.com/expert/mii-status.html
 	http://scyld.com/expert/NWay.html
 
 	Common-sense licensing statement: Using any portion of this program in
@@ -28,10 +29,44 @@
 */
 
 static char *version_msg =
-"alta-diag.c:v2.00 4/19/2000 Donald Becker (becker@scyld.com)\n"
+"alta-diag.c:v2.03 11/18/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: etherdiag [-aEefFmqrRtvVwW] [-p <IOport>] [-A <media>] [-F <media>]\n";
+"Usage: alta-diag [-aEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+			Set a new hardware station address.  Typically diabled, 
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -57,7 +92,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -74,11 +109,12 @@ typedef unsigned char u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -95,23 +131,22 @@ struct option longopts[] = {
 	{ 0, 0, 0, 0 }
 };
 
-extern int    alta_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 extern int sundance_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 
 /* Chip-specific flags. Yes, it's grungy to have the enum here. */
 
 /* The table of known chips.
    Because of the bogus /proc/pci interface we must have both the exact
-   name and a PCI vendor/device IDs.
+   name from the kernel, a common name and the PCI vendor/device IDs.
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
-	char *proc_pci_name;
-	char *part_name;
+	const char *part_name;
+	const char *proc_pci_name;
 	int vendor, device, device_mask;
 	int flags;
 	int io_size;
-	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_num);
+	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_idx);
 } pcidev_tbl[] = {
 	{ "Sundance Technology Alta", "Sundance Technology Alta",
 	  0x1186, 0x1002, 0xffff, 0, 128, sundance_diag},
@@ -134,6 +169,7 @@ static unsigned int opt_flash_show = 0;
 static char	*opt_flash_dumpfile = NULL, *opt_flash_loadfile = NULL;
 
 static unsigned char new_hwaddr[6], set_hwaddr = 0;
+static int emergency_rewrite = 0;
 
 static int scan_proc_pci(int card_num);
 static int parse_media_type(const char *capabilities);
@@ -146,12 +182,11 @@ main(int argc, char **argv)
 {
 	int port_base = 0, chip_type = 0;
 	int errflag = 0, show_version = 0;
-	int emergency_rewrite = 0;
 	int c, longind;
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -166,7 +201,8 @@ main(int argc, char **argv)
 			if (new_default_media < 0)
 				errflag++;
 			break;
-		case 'G': opt_G++; opt_GPIO = strtol(optarg, NULL, 16); break;
+		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -182,7 +218,7 @@ main(int argc, char **argv)
 			}
 		case 'm': show_mii++;	 break;
 		case 'p':
-			port_base = strtol(optarg, NULL, 16);
+			port_base = strtoul(optarg, NULL, 16);
 			break;
 		case 'q': if (verbose) verbose--;		 break;
 		case 'r': opt_restart++;	break;
@@ -256,7 +292,8 @@ static char bogus_iobase[] =
 static char bogus_irq[] =
 "This chip has not been assigned a valid IRQ, and will not function.\n"
 " This must be fixed in the PCI BIOS setup.  The device driver has no way\n"
-" of changing the PCI IRQ settings.\n";
+" of changing the PCI IRQ settings.\n"
+" See  http://www.scyld.com/expert/irq-conflict.html  for more information.\n";
 
 static int scan_proc_bus_pci(int card_num)
 {
@@ -278,8 +315,9 @@ static int scan_proc_bus_pci(int card_num)
 				   &pci_bus, &pci_devid, &irq, &pciaddr0, &pciaddr1) <= 0)
 			break;
 		for (i = 0; pcidev_tbl[i].vendor; i++) {
-			if (pci_devid !=
-				(pcidev_tbl[i].vendor << 16) + pcidev_tbl[i].device)
+			if ((pci_devid >> 16) != pcidev_tbl[i].vendor
+				|| (pci_devid & pcidev_tbl[i].device_mask) !=
+				pcidev_tbl[i].device)
 				continue;
 			chip_idx = i;
 			card_cnt++;
@@ -337,8 +375,9 @@ static int scan_proc_pci(int card_num)
 					int i;
 					if (debug)
 						printf("Named ethernet controller %s.\n", chip_name);
-					for (i = 0; pcidev_tbl[i].proc_pci_name; i++)
-						if (strncmp(pcidev_tbl[i].proc_pci_name, chip_name,
+					for (i = 0; pcidev_tbl[i].part_name; i++)
+						if (pcidev_tbl[i].proc_pci_name  &&
+							strncmp(pcidev_tbl[i].proc_pci_name, chip_name,
 									strlen(pcidev_tbl[i].proc_pci_name))
 							== 0) {
 							state = 2;
@@ -401,7 +440,7 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtol(capabilities, NULL, 16)) <= 0xffff)
+	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -450,6 +489,9 @@ static int get_media_index(const char *name)
 
 /* The chip-specific section for the Sundance Alta. */
 
+#include <ctype.h>
+
+static int eeprom_addr_len(long ioaddr);
 static int read_eeprom(long ioaddr, int location, int addr_len);
 static void write_eeprom(long ioaddr, int index, int value, int addr_len);
 int mdio_read(long ioaddr, int phy_id, int location);
@@ -462,12 +504,16 @@ enum alta_offsets {
 	RxDMAStatus=0x0c, RxListPtr=0x10, RxDMACtrl=0x14, RxDescPoll=0x16,
 	LEDCtrl=0x1a, ASICCtrl=0x30,
 	EEData=0x34, EECtrl=0x36, TxThreshold=0x3c,
-	FlashAddr=0x40, FlashData=0x44, TxStatus=0x46,
+	FlashAddr=0x40, FlashData=0x44, TxStatus=0x46, DownCounter=0x48,
 	IntrClear=0x4a, IntrEnable=0x4c, IntrStatus=0x4e,
-	MACCtrl=0x50, StationAddr=0x54, MaxTxSize=0x5A, RxMode=0x5c, MIICtrl=0x5e,
+	MACCtrl0=0x50, MACCtrl1=0x52, StationAddr=0x54,
+	MaxTxSize=0x5A, RxMode=0x5c, MIICtrl=0x5e,
 	MulticastFilter0=0x60, MulticastFilter1=0x64,
 	RxOctetsLow=0x68, RxOctetsHigh=0x6a, TxOctetsLow=0x6c, TxOctetsHigh=0x6e,
-	TxFramesOK=0x70, RxFramesOK=0x72, RxMissed=0x79,
+	TxFramesOK=0x70, RxFramesOK=0x72, StatsCarrierError=0x74,
+	StatsLateColl=0x75, StatsMultiColl=0x76, StatsOneColl=0x77,
+	StatsTxDefer=0x78, RxMissed=0x79, StatsTxXSDefer=0x7a, StatsTxAbort=0x7b,
+	StatsBcastTx=0x7c, StatsBcastRx=0x7d, StatsMcastTx=0x7e, StatsMcastRx=0x7f,
 };
 
 /* The textual names of the interrupt indications. */
@@ -485,10 +531,10 @@ enum rx_mode_bits {
 };
 
 /* Values read from the EEPROM, and the new image. */
-#define EEPROM_SIZE 256
-static u16 eeprom_contents[EEPROM_SIZE];
-static u16 new_ee_contents[EEPROM_SIZE];
-#define EEPROM_SA_OFFSET	0x20
+#define MAX_EEPROM_SIZE 256
+static u16 eeprom_contents[MAX_EEPROM_SIZE];
+static u16 new_ee_contents[MAX_EEPROM_SIZE];
+#define EEPROM_SA_OFFSET	0x10 	/* In words! */
 #define EEPROM_CSUM_OFFSET	0	/* 0 means none. */
 
 /* Support for Flash operations. */
@@ -509,19 +555,28 @@ static rcvr_mode[] = {
 	{0x07, 0xff, "Normal unicast and all multicast"},
 	{0x15, 0xff, "Normal unicast and hashed multicast"},
 	{0x08, 0x08, "Promiscuous"},
+	{0x00, 0xff, "Uninitialized/disabled"},
 	{0x00, 0x00, "Unknown/invalid"},
 };
 
 int sundance_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 {
+	int chip_active = 0;
 	int rx_mode = inb(ioaddr + RxMode);
 	int tx_status = inb(ioaddr + TxStatus);
-	int mac_mode = inw(ioaddr + MACCtrl);
-	int i, eesize = 8;
+	int mac_mode = inw(ioaddr + MACCtrl0);
+	int ee_addr_len = 6, eeprom_size = 64;
+	int i;
 
 	if (mac_mode == 0xffff)
-		printf("\nThe appears not to exist!  The following info is probably"
-			   " bogus.\n\n");
+		printf("\nThe device appears not to exist!\n"
+			   "The following info is probably bogus.\n\n");
+	/* Start with the unique, and show the basic status. */
+	printf(" Station address ");
+	for (i = 0; i < 5; i++)
+		printf("%2.2x:", inb(ioaddr + StationAddr + i));
+	printf("%2.2x.\n", inb(ioaddr + StationAddr + i));
+
 	for (i = 0; rcvr_mode[i].mask; i++)
 		if ((rx_mode & rcvr_mode[i].mask) == rcvr_mode[i].val) break;
 	printf("  Receive mode is 0x%2.2x: %s.\n", rx_mode, rcvr_mode[i].name);
@@ -530,7 +585,7 @@ int sundance_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		   mac_mode & ~0x20 ? " NON-ETHERNET-STANDARD FEATURES SET" : "");
 	printf("  Tx status %2.2x, threshold %d.\n",
 		   tx_status, inw(ioaddr + TxThreshold));
-	if (tx_status) {
+	if (tx_status  &&  ! chip_active) {
 		printf(" Tx status stack:");
 		for (i = 0; i < 32; i++) {
 			printf(" %4.4x", inw(ioaddr + TxStatus));
@@ -550,10 +605,7 @@ int sundance_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 			for (i = 0; i < 11; i++)
 				if ((1<<i) & intr_status)
 					printf(" %s", intr_names[i]);
-		printf("\n  Station address ");
-		for (i = 0; i < 5; i++)
-			printf("%2.2x:", inb(ioaddr + StationAddr + i));
-		printf("%2.2x.\n", inb(ioaddr + StationAddr + i));
+		printf("\n");
 
 		if (opt_a > 1) {
 			printf("%s chip registers at %#lx",
@@ -570,55 +622,108 @@ int sundance_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		}
 	}
 
-	for (i = 0; i < EEPROM_SIZE; i++)
-		eeprom_contents[i] = read_eeprom(ioaddr, i, 8);
+	if (set_hwaddr || show_eeprom) {
+		ee_addr_len = eeprom_addr_len(ioaddr);
+		eeprom_size = 1 << ee_addr_len;
+		printf(" EEPROM address length %d, %d words.\n",
+			   ee_addr_len, eeprom_size);
+		for (i = 0; i < eeprom_size; i++)
+			eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
+	}
 
+	if (emergency_rewrite  &&  ! set_hwaddr)
+		printf("*** Emergency EEPROM rewrite is only valid when you also "
+			   "specify a new\n*** station address with -H <xx:xx:...>\n");
 	if (set_hwaddr) {
-		memcpy(new_ee_contents, eeprom_contents, eesize << 1);
-		for (i = 0; i < 3; i++) {
+		u16 backup_ee_contents[] = {
+			0x2af8, 0xc262, 0x1186, 0x1102, 0x0000, 0x0000, 0x0000, 0x0000,
+			0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+			0x5000, 0x01ba, 0xaf04, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
+			0x3400, 0x0100, 0x3702, 0x0300, 0x883a, 0x3f04, 0x0303, 0x0103,
+		};
+		memcpy(new_ee_contents, eeprom_contents, eeprom_size << 1);
+		if (emergency_rewrite)
+			memcpy(new_ee_contents, backup_ee_contents,
+				   sizeof backup_ee_contents);
+		for (i = 0; i < 3; i++)
 			new_ee_contents[i + EEPROM_SA_OFFSET] =
-				(new_hwaddr[i*2]<<8) + new_hwaddr[i*2+1];
-		}
+				(new_hwaddr[i*2+1]<<8) + new_hwaddr[i*2];
+#if 0							/* Only do limited write checking */
 		for (i = EEPROM_SA_OFFSET; i < EEPROM_SA_OFFSET + 3; i++)
 			if (new_ee_contents[i] != eeprom_contents[i])
-				write_eeprom(ioaddr, i, new_ee_contents[i], 8);
+				write_eeprom(ioaddr, i, new_ee_contents[i], ee_addr_len);
+#endif
 #if defined(EEPROM_CSUM_OFFSET)  &&  EEPROM_CSUM_OFFSET > 0
+		/* The checksum for the st201 is not documented. */
 		{
 			unsigned short sum = 0;
-			/* Recalculate the checksum. */
 			for (i = 0; i < EEPROM_CSUM_OFFSET; i++)
 				sum ^= new_ee_contents[i];
 			sum = (sum ^ (sum>>8)) & 0xff
 			new_ee_contents[EEPROM_CSUM_OFFSET] = sum;
-			write_eeprom(ioaddr, EEPROM_CSUM_OFFSET, sum);
+			write_eeprom(ioaddr, EEPROM_CSUM_OFFSET, sum, ee_addr_len);
 		}
 #endif
-		for (i = 0; i < eesize; i++)
-			eeprom_contents[i] = read_eeprom(ioaddr, i, 8);
+		if ( ! do_write_eeprom)
+			printf("---- These EEPROM writes will be not be done unless "
+				   "the '-w' flag is passed!\n");
+		for (i = 0; i < eeprom_size; i++)
+			if (new_ee_contents[i] != eeprom_contents[i]) {
+				if (do_write_eeprom)
+					write_eeprom(ioaddr, i, new_ee_contents[i],
+								 ee_addr_len);
+				else
+					printf("Would write %4.4x to replace %4.4x at %d.\n",
+						   new_ee_contents[i], eeprom_contents[i], i);
+				eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
+			}
+	}
+
+	if (show_eeprom > 1) {
+		int elided = 0;
+		printf("EEPROM contents (%d words):", eeprom_size);
+		for (i = 0; i < eeprom_size; i += 8) {
+			int j;
+			if (i != 0  &&  show_eeprom < 3  &&
+				memcmp(&eeprom_contents[i], &eeprom_contents[i-8], 8*2) == 0)
+				elided = 1;
+			else {
+				if (elided) {
+					printf("\n    ...");
+					elided = 0;
+				}
+				printf("\n   0x%2.2x: ", i);
+				for (j = 0; j < 8; j++)
+					printf(" %4.4x", eeprom_contents[i + j]);
+				if (show_eeprom > 2) {
+					printf("  ");
+					for (j = 0; j < 8; j++) {
+						int ew = eeprom_contents[i + j];
+						printf("%c%c",
+							   isprint(ew & 0xff) ? ew & 0xff : '_',
+							   isprint(ew >>   8) ? ew >> 8   : '_' );
+					}
+				}
+			}
+		}
+		printf("\n");
 	}
 
 	if (show_eeprom) {
 		u32 sum = 0;
 		u8 *ee = (u8*)eeprom_contents;
-		if (show_eeprom > 1) {
-			printf("EEPROM contents:");
-			for (i = 0; i < EEPROM_SIZE; i++) {
-				if ((i & 7) == 0)
-					printf("\n0x%3.3x: ", i);
-				printf(" %4.4x", eeprom_contents[i]);
-			}
-			printf("\n");
-		}
-		printf("EEPROM Subsystem IDs, Vendor %2.2x%2.2x Device %2.2x%2.2x.\n",
+		for (i = 0; i < 128-1; i++)
+			sum ^= eeprom_contents[i];
+
+		printf(" EEPROM Subsystem IDs, Vendor %2.2x%2.2x Device %2.2x%2.2x.\n",
 			   ee[5], ee[4], ee[7], ee[6]);
 		printf("  EEPROM Station address is ");
-		for (i = EEPROM_SA_OFFSET; i < EEPROM_SA_OFFSET + 5; i++)
+		for (i = EEPROM_SA_OFFSET*2; i < EEPROM_SA_OFFSET*2 + 5; i++)
 			printf("%2.2x:", ee[i]);
 		printf("%2.2x.\n", ee[i]);
-		printf("  Configuration %4.4x, ASIC Control %4.4x.\n",
-			    eeprom_contents[0], eeprom_contents[1]);
-		for (i = 0; i < EEPROM_SIZE-1; i++)
-			sum ^= eeprom_contents[i];
+		printf("  Configuration %4.4x, ASIC Control %4.4x.\n"
+			   "  Checksum %4.4x\n",
+			    eeprom_contents[0], eeprom_contents[1], sum);
 	}
 	if (show_mii) {
 		int phys[4], phy, phy_idx = 0;
@@ -670,6 +775,7 @@ int sundance_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 	if (opt_flash_loadfile  || opt_flash_dumpfile  ||  opt_flash_show)
 		printf("Flash operations not configured into this program.\n");
 	if (opt_flash_show) {
+		printf("The first few boot ROM bytes are:");
 		for (i = 0; i < 16; i++)
 			printf(" %2.2x", alta_flash_in(ioaddr, i));
 		printf(" (%8.8x -> %2.2x).\n",
@@ -680,6 +786,10 @@ int sundance_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 	return 0;
 }
 
+static int eeprom_addr_len(long ioaddr)
+{
+	return 8;
+}
 static int read_eeprom(long ioaddr, int location, int ignore_addr_len)
 {
 	int boguscnt = 1000;		/* Typical 190 ticks. */
@@ -687,11 +797,13 @@ static int read_eeprom(long ioaddr, int location, int ignore_addr_len)
 	do {
 		if (! (inw(ioaddr + EECtrl) & 0x8000)) {
 			if (debug)
-				printf("EEPROM read took %d ticks.\n", 1000 - boguscnt);
+				printf("EEPROM read took %d ticks to read %4.4x.\n",
+					   1000 - boguscnt, inw(ioaddr + EEData));
 			return inw(ioaddr + EEData);
 		}
 	} while (--boguscnt > 0);
-	printf("EEPROM read failed.\n");
+	fprintf(stderr, "Configuration EEPROM read of location %d failed, "
+			"status %4.4x.\n", location, inw(ioaddr + EEData));
 	return 0;
 }
 
@@ -703,20 +815,36 @@ static void write_eeprom(long ioaddr, int location, int value,
 		printf("Would write 0x%4.4x to offset %d.\n", value, location);
 		return;
 	}
-	outw(0x0100 | (location & 0xff), ioaddr + 0x36);
+	/* Write enable. */
+	outw(0x00C0, ioaddr + EECtrl);
 	do {
-		if (! (inw(ioaddr + 0x36) & 0x8000))
-			return;
+		if (! (inw(ioaddr + EECtrl) & 0x8000))
+			break;
 	} while (--boguscnt > 0);
-	fprintf(stderr, "EEPROM write of location %d with %#x failed!\n",
-			location, value);
+	outw(value, ioaddr + EEData);
+	outw(0x0100 | (location & 0xff), ioaddr + EECtrl);
+	do {
+		if (! (inw(ioaddr + EECtrl) & 0x8000)) {
+			if (debug)
+				fprintf(stderr, "EEPROM write of location %d with %#x took %d"
+						" ticks, status %4.4x\n",
+						location, inw(ioaddr + EEData),
+						10000- boguscnt, inw(ioaddr + EECtrl));
+			return;
+		}
+	} while (--boguscnt > 0);
+	fprintf(stderr, "EEPROM write of %#4.4x to location %d with failed "
+			"with status %#x.\n", location, value, inw(ioaddr + EECtrl));
+	outw(0x0000 | (location & 0xff), ioaddr + EECtrl);
 	return;
 }
 
 /* Read and write the MII registers using software-generated serial
    MDIO protocol.  It is just different enough from the serial EEPROM protocol
    to not share code.  The maxium data clock rate is 2.5 Mhz. */
-#define mdio_delay()  inb(mdio_addr) 	/* Extra bus turn-around as a delay. */
+#define mdio_out(value, mdio_addr) outb(value, mdio_addr)
+#define mdio_in(mdio_addr) inb(mdio_addr)
+#define mdio_delay(mdio_addr)  inb(mdio_addr)	/* Flush buffered write.*/
 #define MDIO_IO_OFFSET MIICtrl
 
 #define MDIO_SHIFT_CLK	0x01
@@ -733,10 +861,10 @@ static void mdio_sync(long ioaddr)
 	int i;
 
 	for (i = 32; i >= 0; i--) {
-		outl(MDIO_WRITE1, mdio_addr);
-		mdio_delay();
-		outl(MDIO_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_WRITE1, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	return;
 }
@@ -757,22 +885,22 @@ int mdio_read(long ioaddr, int phy_id, int location)
 		if (verbose > 3)		/* Debug: 5 */
 			printf("%d", (mii_cmd & (1 << i)) ? 1 : 0);
 
-		outb(MDIO_ENB | dataval, mdio_addr);
-		mdio_delay();
-		outb(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
-		if (verbose > 3) printf("%x ", (inb(mdio_addr) >> 16) & 0x0f);
-		mdio_delay();
+		mdio_out(MDIO_ENB | dataval, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
+		if (verbose > 3) printf(" %x", (mdio_in(mdio_addr) >> 16) & 0x0f);
+		mdio_delay(mdio_addr);
 	}
-	if (verbose > 3) printf("-> %x", (inb(mdio_addr) >> 16) & 0x0f);
+	if (verbose > 3) printf("-> %x", (mdio_in(mdio_addr) >> 16) & 0x0f);
 
 	/* Read the two transition, 16 data, and wire-idle bits. */
 	for (i = 19; i > 0; i--) {
-		outb(MDIO_ENB_IN, mdio_addr);
-		mdio_delay();
-		retval = (retval << 1) | ((inb(mdio_addr) & MDIO_DATA_READ) ? 1 : 0);
-		if (verbose > 3) printf(" %x", (inb(mdio_addr) >> 16) & 0x0f);
-		outb(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_ENB_IN, mdio_addr);
+		mdio_delay(mdio_addr);
+		retval = (retval << 1) | ((mdio_in(mdio_addr) & MDIO_DATA_READ) ? 1 : 0);
+		if (verbose > 3) printf(" %x", (mdio_in(mdio_addr) >> 16) & 0x0f);
+		mdio_out(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	if (verbose > 3) printf(" == %4.4x.\n", retval);
 	return (retval>>1) & 0xffff;
@@ -789,17 +917,17 @@ void mdio_write(long ioaddr, int phy_id, int location, int value)
 	/* Shift the command bits out. */
 	for (i = 31; i >= 0; i--) {
 		int dataval = (mii_cmd & (1 << i)) ? MDIO_WRITE1 : MDIO_WRITE0;
-		outb(MDIO_ENB | dataval, mdio_addr);
-		mdio_delay();
-		outb(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_ENB | dataval, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	/* Clear out extra bits. */
 	for (i = 2; i > 0; i--) {
-		outb(MDIO_ENB_IN, mdio_addr);
-		mdio_delay();
-		outb(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_ENB_IN, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	return;
 }
