@@ -4,7 +4,7 @@
    based on the Intel "Speedo3" chip series: the i82557, '558 and '559.
    These chips are used on the EtherExpress Pro100B, and EEPro PCI 10+.
 
-	Copyright 1998-2001 by Donald Becker.
+	Copyright 1998-2003 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -22,14 +22,49 @@
 
 	Common-sense licensing statement: Using any portion of this program in
 	your own program means that you must give credit to the original author
-	and release the resulting code under the GPL.
+	and release the resulting code under the GPL.  To use this code under
+	other terms requires an explicit license from the copyright holder.
 */
 
 static char *version_msg =
-"eepro100-diag.c:v2.04 5/21/2001 Donald Becker (becker@scyld.com)\n"
+"eepro100-diag.c:v2.12 4/15/2003 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: eepro100-diag [-aEefFmqrRtsvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: eepro100-diag [-aDEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type (with '-p'), use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:89:ab
+        Set a new hardware station address.  Typically disabled for safety.
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -43,8 +78,8 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
-
-#include <asm/types.h>
+#include <sys/types.h>
+#include <ctype.h>
 
 #if defined(__linux__)  &&  __GNU_LIBRARY__ == 1
 #include <asm/io.h>			/* Newer libraries use <sys/io.h> instead. */
@@ -52,28 +87,29 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
-extern int flash_show(long addr_ioaddr, long data_ioaddr);
-extern int flash_dump(long addr_ioaddr, long data_ioaddr, char *filename);
-extern int flash_program(long addr_ioaddr, long data_ioaddr, char *filename);
-extern int (*flash_in_hook)(long addr, int offset);
-extern void (*flash_out_hook)(long addr, int offset, int val);
+extern int flash_pci_rom_show(long addr_ioaddr, long data_ioaddr);
+extern int flash_pci_rom_dump(long addr_ioaddr, long data_ioaddr,
+							  const char *filename);
+extern int flash_pci_rom_program(long addr_ioaddr, long data_ioaddr,
+								 const char *filename);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -81,6 +117,7 @@ struct option longopts[] = {
 	{"show-mii",  0, 0, 'm'},	/* Dump MII management registers. */
 	{"port-base",  1, 0, 'p'},	/* Use the specified I/O address. */
 	{"quiet",	0, 0, 'q'},		/* Decrease verbosity */
+	{"restart",	0, 0, 'r'},		/* Restart autonegotiation. */
 	{"reset",	0, 0, 'R'},		/* Reset the transceiver. */
 	{"chip-type",  1, 0, 't'},	/* Assume the specified chip type index. */
 	{"test",	0, 0, 'T'},		/* Do register and SRAM test. */
@@ -94,7 +131,7 @@ extern int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 
 /* The table of known chips.
    Because of the bogus /proc/pci interface we must have both the exact
-   name and a PCI vendor/device IDs.
+   name from the kernel, a common name and the PCI vendor/device IDs.
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
@@ -112,9 +149,23 @@ struct pcidev_entry {
 	{ "Intel EtherExpressPro100+ Type 1029", 0,
 	  0x8086, 0x1029, 0xffff, 0, 0x20, eepro100_diag },
 	{ "Intel InBusiness i82559", 0,
-	  0x8086, 0x1030, 0xffff, 0, 0x20, eepro100_diag },
+	  0x8086, 0x1030, 0xffff, 0, 0x40, eepro100_diag },
 	{ "Intel i82562 Pro/100 V", 0,
-	  0x8086, 0x2449, 0xffff, 0, 0x20, eepro100_diag },
+	  0x8086, 0x2449, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VE (type 1031)", 0,
+	  0x8086, 0x1031, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VM-1038", 0,
+	  0x8086, 0x1038, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VM-1039", 0,
+	  0x8086, 0x1039, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VE-103a", 0,
+	  0x8086, 0x103A, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VE-103b", 0,
+	  0x8086, 0x103B, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VE-103d", 0,
+	  0x8086, 0x103D, 0xffff, 0, 0x40, eepro100_diag },
+	{ "Intel Pro/100 VE-unknown (1030 series)", 0,
+	  0x8086, 0x1030, 0xfff0, 0, 0x40, eepro100_diag },
 	{ 0, 0, 0, 0},
 };
 
@@ -150,7 +201,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -166,6 +217,7 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -396,6 +448,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -403,7 +456,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -434,16 +488,20 @@ struct { char *name; int value; } mediamap[] = {
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -451,6 +509,18 @@ static int get_media_index(const char *name)
 /* Chip-specific section. */
 
 /* The chip-specific section for the Intel EEPro100 diagnostic. */
+
+static char sleep_mode_msg[] =
+"   Sleep mode is enabled.  This is not recommended.\n"
+"   Under high load the card may not respond to\n"
+"   PCI requests, and thus cause a master abort.\n"
+"   To clear sleep mode use the '-G 0 -w -w -f' options.\n";
+
+static char no_xcvr_msg[] =
+"   This NIC has no transceiver programmed into the configuration EEPROM.\n"
+"   This is likely a manufacturing error.  It may be corrected with\n"
+"   software.  Contact the board vendor for an update.\n";
+
 static int read_eeprom(long ioaddr, int location, int addr_len);
 static void write_eeprom(long ioaddr, int index, int value, int addr_len);
 static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len);
@@ -459,6 +529,12 @@ static void eepro100_eeprom(unsigned short *ee_data);
 int mdio_read(long ioaddr, int phy_id, int location);
 void mdio_write(long ioaddr, int phy_id, int location, int value);
 
+int flash_pci_rom_show(long ioaddr, long data) __attribute__((weak));
+int flash_pci_rom_dump(long ioaddr, long data, const char *filename)
+	 __attribute__((weak));
+int flash_pci_rom_program(long ioaddr, long data, const char *filename)
+	 __attribute__((weak));
+
 
 /* Offsets to the various registers.
    Accesses need not be longword aligned. */
@@ -496,16 +572,25 @@ int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx)
 {
 	int i;
 	int phy0, phy1;
-	int ee_addr_size = 6;		/* Modified below. */
-	int eeprom_size = 64;
+	int ee_addr_len = 6;		/* Modified below. */
+	int eeprom_size = 64, eeprom_change = 0;
+	unsigned short status = inw(ioaddr + SCBStatus);
 
 	/* It's mostly safe to use the EEPro100 EEPROM and MDIO register during
 	   operation.  But warn the user, and make then pass '-f'. */
-	if (!opt_f  && (inw(ioaddr + SCBStatus) & 0x00fc) != 0x0000) {
-		printf("A potential i82557 chip has been found, but it appears to "
-			   "be active.\nEither shutdown the network, or use the"
-			   " '-f' flag.\n");
-		return 1;
+	if (!opt_f) {
+		if (status == 0xffff) {
+			printf(" * A recognized chip has been found, but it does not "
+				   "appear to exist in\n * I/O space.  Use the"
+				   " '-f' flag to see the register values anyway.\n");
+			return 1;
+		}
+		if ((status & 0x00fc) != 0x0000) {
+			printf("A potential i82557 chip has been found, but it appears to "
+				   "be active.\nEither shutdown the network, or use the"
+				   " '-f' flag.\n");
+			return 1;
+		}
 	}
 
 	if (verbose > 1 || show_regs) {
@@ -517,13 +602,11 @@ int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx)
 			"Suspended - no buffers", "No resources - no buffers", "Broken-11",
 			"Ready, but no buffers",  "Broken-13", "Broken-14", "Broken-15",
 		};
-		unsigned short status;
 
 		printf("i82557 chip registers at %#x:\n ", (int)ioaddr);
 		for (i = 0; i < 0x18; i += 4)
 			printf(" %8.8x", inl(ioaddr + i));
 		printf("\n");
-		status = inw(ioaddr + SCBStatus);
 		printf("  %snterrupt sources are pending.\n",
 			   (status & 0xff00) ? "I": "No i");
 		printf("   The transmit unit state is '%s'.\n",
@@ -544,14 +627,15 @@ int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx)
 	/* Read the EEPROM. */
 	{
 		int size_test = do_eeprom_cmd(ioaddr, (EE_READ_CMD << 8) << 16, 27);
-		ee_addr_size = (size_test & 0xffe0000) == 0xffe0000 ? 8 : 6;
-		eeprom_size = 1 << ee_addr_size;
+		ee_addr_len = (size_test & 0xffe0000) == 0xffe0000 ? 8 : 6;
+		eeprom_size = 1 << ee_addr_len;
 		if (debug)
 			printf("EEPROM size probe returned %#x, %d bit address.\n",
-				   size_test, ee_addr_size);
+				   size_test, ee_addr_len);
+		for (i = 0; i < eeprom_size; i++)
+			eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
+		memcpy(new_ee_contents, eeprom_contents, eeprom_size << 1);
 	}
-	for (i = 0; i < eeprom_size; i++)
-		eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_size);
 
 	phy0 = eeprom_contents[6] & 0x8000 ? -1 : eeprom_contents[6] & 0x1f;
 	phy1 = eeprom_contents[7] & 0x8000 ? -1 : eeprom_contents[7] & 0x1f;
@@ -570,13 +654,20 @@ int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx)
 					printf(" %4.4x", eeprom_contents[i + j]);
 					sum += eeprom_contents[i + j];
 				}
+				printf("  ");
+				for (j = 0; j < 8; j++) {
+					int ew = eeprom_contents[i + j];
+					printf("%c%c",
+						   isprint(ew & 0xff) ? ew & 0xff : '_',
+						   isprint(ew >>   8) ? ew >> 8   : '_' );
+				}
 				printf("\n");
 			} else if ( ! last_row_zeros) {
 				last_row_zeros = 1;
 				printf("      ...\n");
 			}
 		}
-		if (sum == 0xBaBa)
+		if (sum == 0xBaBa)		/* Bahhhh. */
 			printf(" The EEPROM checksum is correct.\n");
 		else
 			printf(" *****  The EEPROM checksum is INCORRECT!  *****\n"
@@ -586,29 +677,67 @@ int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx)
 	/* The user will usually want to see the interpreted EEPROM contents. */
 	if (verbose > 1 || show_eeprom)
 		eepro100_eeprom(eeprom_contents);
-	/* Warning: this is example code only!  Actual write code should
-	   verify the input values, verify that the writes actually
-	   occured, and never let the card get into a non-working state. */
-	if (do_write_eeprom  &&  opt_G) {
-		int checksum = 0;
-		if (do_write_eeprom < 2)
-			printf("Would write %x to the RPL-boot configuration.\n",
-				   opt_GPIO);
-		else {
-			printf("Writing %x to the RPL-boot configuration.\n",
-				   opt_GPIO);
-			write_eeprom(ioaddr, 0x20, opt_GPIO, ee_addr_size);
-			write_eeprom(ioaddr, 0x21, opt_GPIO >> 16, ee_addr_size);
-			eeprom_contents[0x20] = read_eeprom(ioaddr, 0x20, ee_addr_size);
-			eeprom_contents[0x21] = read_eeprom(ioaddr, 0x21, ee_addr_size);
+
+	/* These two "-G <n>" settings may change or be removed! */
+	/* Clear the sleep mode bit with "-G 0". */
+	if (opt_G  &&  opt_GPIO == 0  &&  (eeprom_contents[10] & 0x0002)) {
+		printf("Changing the chip configuration word (offset 10) to %#4.4x.\n",
+			   new_ee_contents[10]);
+		new_ee_contents[10] = eeprom_contents[10] & ~0x0002;
+		eeprom_change++;
+	}
+	if (opt_G  &&  opt_GPIO == 1  &&  eeprom_contents[6] == 0) {
+		printf("Changing the transceiver type to MII RJ45.\n");
+		new_ee_contents[5] = 0x0101;
+		new_ee_contents[6] = 0x4401;
+		eeprom_change++;
+	}
+
+#if ! defined(ENABLE_HWADDR_UPDATE)
+	if (emergency_rewrite)
+		fprintf(stderr, "Full EEPROM overwrite has been disabled"
+				" in this release of the diagnostic.\n");
+	if (set_hwaddr)
+		fprintf(stderr, "Modifying the station (MAC) address has been disabled"
+				" in this release of the diagnostic.\n");
+#else
+	/* Do not enable this code for end-users: it is too easy to misuse. */
+	if (emergency_rewrite  &&  ! set_hwaddr)
+		printf("*** Emergency EEPROM rewrite is only valid when you also "
+			   "specify a new\n*** station address with -H <xx:xx:...>\n");
+	if (set_hwaddr) {
+		if (emergency_rewrite)
+			memcpy(new_ee_contents, djb_eepro100_eeprom,
+				   sizeof djb_eepro100_eeprom);
+		for (i = 0; i < 3; i++)
+			new_ee_contents[i] =
+				(new_hwaddr[i*2+1]<<8) + new_hwaddr[i*2];
+		eeprom_change++;
+	}
+#endif
+	if (eeprom_change) {
+		u16 checksum = 0;
+		/* Recalculate the checksum. */
+		for (i = 0; i < eeprom_size - 1; i++)
+			checksum += new_ee_contents[i];
+		if (new_ee_contents[i] != 0xBABA - checksum) {
+			if (debug)
+				printf("The checksum at location %d must be "
+					   "updated from 0x%4.4X to 0x%4.4X.\n", i,
+					   eeprom_contents[i], 0xBABA - checksum);
+			new_ee_contents[i] = 0xBABA - checksum;
 		}
-		for (i = 0; i < eeprom_size - 1; i++) {
-			checksum += eeprom_contents[i];
-		}
-		if (debug)
-			printf("Writing the new checksum 0x%4.4X to location %d.\n",
-				   0xBABA - checksum, i);
-		write_eeprom(ioaddr, i, 0xBABA - checksum, ee_addr_size);
+		for (i = 0; i < eeprom_size; i++)
+			if (new_ee_contents[i] != eeprom_contents[i]) {
+				if (do_write_eeprom) {
+					printf("Writing %#4.4x to configuration EEPROM word %d.\n",
+						   new_ee_contents[i], i);
+					write_eeprom(ioaddr, i, new_ee_contents[i],
+								 ee_addr_len);
+				} else
+					printf("  Would write %4.4x to replace %4.4x at %d.\n",
+						   new_ee_contents[i], eeprom_contents[i], i);
+			}
 	}
 
 	/* Grrr, it turns out they the PHY is not always #0. */
@@ -616,41 +745,66 @@ int eepro100_diag(int vend_id, int dev_id, long ioaddr, int part_idx)
 		if (phy0 > 0 &&
 			mdio_read(ioaddr, phy0, 0) != 0xffff  && 
 			mdio_read(ioaddr, phy0, 0) != 0x0000) {
-			int mii_reg;
-			printf(" MII PHY #%d transceiver registers:", phy0);
-			for (mii_reg = 0; mii_reg < 32; mii_reg++)
-				printf("%s %4.4x", (mii_reg % 8) == 0 ? "\n " : "",
-					   mdio_read(ioaddr, phy0, mii_reg));
-			printf(".\n");
-#ifdef LIBMII
+			printf("Primary transceiver is MII PHY #%d.", phy0);
 			show_mii_details(ioaddr, phy0);
-#endif
 		}
 		if (phy1 > 0 &&
 			mdio_read(ioaddr, phy1, 0) != 0xffff  && 
 			mdio_read(ioaddr, phy1, 0) != 0x0000) {
-			int mii_reg;
-			printf(" Alternate MII PHY (#%d) transceiver registers:", phy1);
-			for (mii_reg = 0; mii_reg < 32; mii_reg++)
-				printf("%s %4.4x", (mii_reg % 8) == 0 ? "\n " : "",
-					   mdio_read(ioaddr, phy1, mii_reg));
-			printf(".\n");
-#ifdef LIBMII
+			printf("Alternate transceiver is MII PHY #%d.", phy0);
 			show_mii_details(ioaddr, phy1);
-#endif
-		}
-		if (phy0 > 0 &&  nway_advertise > 0) {
-			printf(" Setting the media capability advertisement register of "
-				   "PHY #%d to 0x%4.4x.\n", phy0, nway_advertise | 1);
-			mdio_write(ioaddr, phy0, 4, nway_advertise | 1);
 		}
 	}
+	if (phy0 > 0 &&  nway_advertise > 0) {
+		printf("Setting the media capability advertisement register of "
+			   "PHY #%d to 0x%4.4x.\n", phy0, nway_advertise | 1);
+		mdio_write(ioaddr, phy0, 4, nway_advertise | 1);
+	}
 	if (show_mii > 1) {			/* Monitor MII status */
 		monitor_mii(ioaddr, phy0);
 	}
+
+	/* Flash operations. */
+	if (opt_flash_show)
+		flash_pci_rom_show(ioaddr, 0);
+	if (opt_flash_dumpfile)
+		if (flash_pci_rom_dump(ioaddr, 0, opt_flash_dumpfile) < 0) {
+			fprintf(stderr, "Failed to save the old Flash BootROM image "
+					"into file '%s'.\n", opt_flash_dumpfile);
+			return 3;
+		}
+	if (opt_flash_loadfile)
+		if (flash_pci_rom_program(ioaddr, 0, opt_flash_loadfile) < 0) {
+			fprintf(stderr, "Failed to load the new Flash BootROM image "
+					"from file '%s'.\n", opt_flash_loadfile);
+			return 4;
+		}
+
 	return 0;
 }
 
+/*
+  Reading and programming the EEPro100 flash requires a kernel extension
+  that allocates a memory region and mmaps the PCI boot ROM space.
+  The current (mid-2002) code that does this only works on a few kernel
+  versions, and thus is not enabled by default.
+*/
+int flash_pci_rom_show(long ioaddr, long data)
+{
+	printf("Flash operations not configured into this program.\n");
+	return -1;
+}
+int flash_pci_rom_dump(long ioaddr, long data, const char *filename)
+{
+	printf("Flash operations not configured into this program.\n");
+	return -1;
+}
+int flash_pci_rom_program(long ioaddr, long data, const char *filename)
+{
+	printf("Flash operations not configured into this program.\n");
+	return -1;
+}
+
 
 /* Serial EEPROM section.
    A "bit" grungy, but we work our way through bit-by-bit :->. */
@@ -774,7 +928,10 @@ void mdio_write(long ioaddr, int phy_id, int location, int value)
 	return;
 }
 
-#ifndef LIBMII
+/* Simple alternate functions if libmii is not used. */
+extern int monitor_mii(long ioaddr, int phy_id) __attribute__((weak));
+extern int show_mii_details(long ioaddr, int phy_id) __attribute__((weak));
+
 int monitor_mii(long ioaddr, int phy_id)
 {
 	int i;
@@ -795,7 +952,17 @@ int monitor_mii(long ioaddr, int phy_id)
 	}
 	return 0;
 }
-#endif  /* not LIBMII */
+
+int show_mii_details(long ioaddr, int phy_id)
+{
+	int mii_reg;
+	printf(" MII PHY #%d transceiver registers:", phy_id);
+	for (mii_reg = 0; mii_reg < 32; mii_reg++)
+		printf("%s %4.4x", (mii_reg % 8) == 0 ? "\n  " : "",
+			   mdio_read(ioaddr, phy_id, mii_reg));
+	printf(".\n");
+	return 0;
+}
 
 
 /* PHY media interface chips. */
@@ -842,10 +1009,10 @@ static void eepro100_eeprom(unsigned short *eeprom)
 	if (eeprom[7] & 0x0700)
 		printf("    Secondary interface chip %s, PHY %d.\n",
 			   phys[(eeprom[7]>>8)&7], phy1);
-	if (eeprom[10] & 0x0004)
-		printf("   Sleep mode is enabled.  This is not recommended."
-			   "  Under high load the\n""     card may not respond to"
-			   " PCI requests, and cause a master abort.\n");
+	if (eeprom[6] == 0)
+		printf(no_xcvr_msg);
+	if (eeprom[10] & 0x0002)
+		printf(sleep_mode_msg);
 
 	return;
 }
