@@ -4,7 +4,7 @@
    based on the SMC83C170 series EPIC/100 chip, as used on the SMC EtherPowerII
    boards.
 
-	Copyright 1998-2001 by Donald Becker.
+	Copyright 1998-2002 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -33,7 +33,7 @@
 */
 
 static char *version_msg =
-"epic-diag.c:v2.01 1/6/2000 Donald Becker (becker@scyld.com)\n"
+"epic-diag.c:v2.03 6/22/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
 "Usage: epic-diag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
@@ -104,9 +104,9 @@ struct pcidev_entry {
 	int io_size;
 	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_idx);
 } pcidev_tbl[] = {
-	{"SMSC EPIC/100 83c170", "SMC 9432 TX", 0x10B8, 0x0005, 0xffffffff,
+	{"SMSC EPIC/100 83c170", "SMC 9432 TX", 0x10B8, 0x0005, 0xffff,
 	 0, 32, epic_diag},
-	{"SMSC EPIC/C 83c175", 0, 0x10B8, 0x0006, 0xffffffff, 0, 32, epic_diag},
+	{"SMSC EPIC/C 83c175", 0, 0x10B8, 0x0006, 0xffff, 0, 32, epic_diag},
 	{ 0, 0, 0, 0},
 };
 
@@ -121,11 +121,10 @@ unsigned int opt_GPIO = 0;		/* General purpose I/O setting. */
 int do_write_eeprom = 0, do_test = 0;
 int nway_advertise = 0, fixed_speed = -1;
 int new_default_media = -1;
-#if defined(LIBFLASH)		/* Valid with libflash only. */
 static unsigned int opt_flash_show = 0;
 static char	*opt_flash_dumpfile = NULL, *opt_flash_loadfile = NULL;
-#endif
 static unsigned char new_hwaddr[6], set_hwaddr = 0;
+static int emergency_rewrite = 0;
 
 static int scan_proc_pci(int card_num);
 static int parse_media_type(const char *capabilities);
@@ -139,7 +138,6 @@ main(int argc, char **argv)
 {
 	int port_base = 0, chip_type = 0;
 	int errflag = 0, show_version = 0;
-	int emergency_rewrite = 0;
 	int c, longind;
 	int card_num = 0;
 	extern char *optarg;
@@ -159,7 +157,7 @@ main(int argc, char **argv)
 			if (new_default_media < 0)
 				errflag++;
 			break;
-		case 'G': opt_G++; opt_GPIO = strtol(optarg, NULL, 16); break;
+		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -175,7 +173,7 @@ main(int argc, char **argv)
 			}
 		case 'm': show_mii++;	 break;
 		case 'p':
-			port_base = strtol(optarg, NULL, 16);
+			port_base = strtoul(optarg, NULL, 16);
 			break;
 		case 'q': if (verbose) verbose--;		 break;
 		case 'r': opt_restart++;	break;
@@ -185,11 +183,9 @@ main(int argc, char **argv)
 		case 'V': show_version++;	break;
 		case 'w': do_write_eeprom++;	break;
 		case 'W': opt_watch++;		break;
-#if defined(LIBFLASH)		/* Valid with libflash only. */
 		case 'B': opt_flash_show++;	break;
 		case 'L': opt_flash_loadfile = optarg;	break;
 		case 'S': opt_flash_dumpfile = optarg;	break;
-#endif
 		case '?':
 			errflag++;
 		}
@@ -274,8 +270,9 @@ static int scan_proc_bus_pci(int card_num)
 				   &pci_bus, &pci_devid, &irq, &pciaddr0, &pciaddr1) <= 0)
 			break;
 		for (i = 0; pcidev_tbl[i].vendor; i++) {
-			if (pci_devid !=
-				(pcidev_tbl[i].vendor << 16) + pcidev_tbl[i].device)
+			if ((pci_devid >> 16) != pcidev_tbl[i].vendor
+				|| (pci_devid & pcidev_tbl[i].device_mask) !=
+				pcidev_tbl[i].device)
 				continue;
 			chip_idx = i;
 			card_cnt++;
@@ -398,7 +395,7 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtol(capabilities, NULL, 16)) <= 0xffff)
+	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -509,7 +506,7 @@ int epic_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 	int chip_active = 0, chip_lowpower = 0;
 	unsigned int rx_ctrl = inl(ioaddr + RxCtrl);
 	unsigned int genctrl = inl(ioaddr + GENCTL);
-	int eeprom_size = 64, eeprom_addr_size;
+	int eeprom_size = 64, eeprom_addr_size = 6;
 	int i;
 
 	/* It's mostly safe to examine the registers and EEPROM during
@@ -718,6 +715,12 @@ int epic_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 #endif
 	}
 
+	/* The EPIC chips do not have flash boot ROM access through I/O space,
+	   only through the mapped PCI BIOS extension space.
+	   No write-enable is needed. */
+	if (opt_flash_loadfile  || opt_flash_dumpfile  ||  opt_flash_show)
+		printf("Flash operations not configured into this program.\n");
+
 	if (chip_lowpower)
 		outl(genctrl | 0x0008, ioaddr + GENCTL);
 	return 0;
@@ -905,14 +908,30 @@ static void parse_eeprom(unsigned short *eeprom)
 	int i;
 	unsigned char sum = 0;
 
-	printf("Parsing the EEPROM of a EPIC/100:\n Station Address ");
-	for (i = 0; i < 5; i++)
-		printf("%2.2X:", p[i]);
-	printf("%2.2X.\n Board name '%.20s', revision %d.\n", p[5], p + 88, p[6]);
 	for (i = 0; i < 8; i++)
 		sum += p[i];
-	printf("  Calculated checksum is %2.2x, %scorrect.\n", sum, sum ? "in":"");
-	printf(" Subsystem ID Vendor/Device %4.4x/%4.4x.\n", eeprom[6], eeprom[7]);
+	printf("Interpreted EEPROM contents of a SMSC EPIC:\n Station Address ");
+	/* Little-endian assumption here. */
+	for (i = 0; i < 5; i++)
+		printf("%2.2X:", p[i]);
+	printf("%2.2X.\n"
+		   "  Calculated checksum on station address is %2.2x, %scorrect.\n"
+		   "  Board revision %d, name '%.20s'.\n"
+		   "  PCI Subsystem ID Vendor/Device %4.4x/%4.4x.\n"
+		   "  PCI Max latency %d  Min grant %d (requested values in 250ns "
+		   "units)\n"
+		   "  Non-volatile Control setting %4.4x:\n"
+		   "    PCI BAR 1 (memory mapped register) is %s.\n"
+		   "    GPIO pin 1 is an %s, initial value %d.\n"
+		   "    GPIO pin 2 is an %s, initial value %d.\n"
+		   "\n",
+		   eeprom[2] >> 8, sum, sum ? "in" : "",
+		   eeprom[3] & 0xff, p + 88,
+		   eeprom[6], eeprom[7], eeprom[5] >> 8, eeprom[5] & 0xff,
+		   eeprom[4], eeprom[4] & 0x01 ? "enabled" : "disabled",
+		   eeprom[4] & 0x04 ? "output" : "input", eeprom[4] & 0x10 ? 1:0,
+		   eeprom[4] & 0x08 ? "output" : "input", eeprom[4] & 0x20 ? 1:0
+		   );
 	return;
 }
 
