@@ -4,7 +4,7 @@
    based on the following chips:
 	Myson MTD803.
 
-	Copyright 1998-2001 by Donald Becker.
+	Copyright 1998-2002 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -28,10 +28,44 @@
 */
 
 static char *version_msg =
-"myson-diag.c:v1.00 5/15/2001 Donald Becker (becker@scyld.com)\n"
+"myson-diag.c:v1.07 8/07/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: myson-diag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: myson-diag [-aEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+			Set a new hardware station address.  Typically diabled, 
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -56,7 +90,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -73,11 +107,12 @@ typedef unsigned char u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -155,7 +190,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -171,6 +206,7 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -606,9 +642,9 @@ int myson_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 			   "specify a new\n*** station address with -H <xx:xx:...>\n");
 	if (set_hwaddr) {
 		u16 backup_ee_contents[] = {
-			0x7312, 0x0080, 0x0000, 0x0000, 0x2006, 0x1516, 0x0803, 0x0000,
+			0x7310, 0x0083, 0x0000, 0x0000, 0x4020, 0x1516, 0x0803, 0x0000,
 			0x0200, 0xb02a, 0x9927, 0x0000, 0x0200, 0xb02a, 0x9927, 0x0000,
-			0x02FA, 0x0000, };
+			0x00F2, 0x0000, };
 		memcpy(new_ee_contents, eeprom_contents, eeprom_size << 1);
 		if (emergency_rewrite)
 			memcpy(new_ee_contents, backup_ee_contents,
