@@ -1,10 +1,12 @@
-/* natsemi-diag.c: Diagnostic and setup core code for Linux ethercards.
+/* natsemi-diag.c: National Semiconductor DP83815 diagnostic and setup.
 
    This is a diagnostic and EEPROM setup program for Ethernet adapters
    based on the
 	National Semiconductor DP83810 / 83815
 
-	Copyright 1998-2001 by Donald Becker.
+	Written 1998-2002 by Donald Becker.
+	Copyright Donald Becker and Scyld Computing Corporation
+
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -28,10 +30,48 @@
 */
 
 static char *version_msg =
-"natsemi-diag.c:v2.04 1/8/2001 Donald Becker (becker@scyld.com)\n"
+"natsemi-diag.c:v2.07 12/17/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: natsemi-diag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: natsemi-diag [-aEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      natsemi-diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+			Set a new hardware station address.  Typically diabled, 
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+ To read and write the boot BIOS extension Flash ROM
+   -B       Show the first few bytes of the ROM
+   -L FILE  Load the Flash from FILE.
+   -S FILE  Store the Flash image to FILE.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -45,6 +85,8 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <ctype.h>
 
 /* The following are required only with unaligned field accesses. */
 #include <asm/types.h>
@@ -56,7 +98,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -67,17 +109,18 @@ extern int (*flash_in_hook)(long addr, int offset);
 extern void (*flash_out_hook)(long addr, int offset, int val);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -100,7 +143,7 @@ extern int natsemi_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 
 /* The table of known chips.
    Because of the bogus /proc/pci interface we must have both the exact
-   name and a PCI vendor/device IDs.
+   name from the kernel, a common name and the PCI vendor/device IDs.
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
@@ -124,7 +167,7 @@ unsigned int opt_a = 0,					/* Show-all-interfaces flag. */
 	opt_watch = 0,
 	opt_G = 0;
 unsigned int opt_GPIO = 0;		/* General purpose I/O setting. */
-int do_write_eeprom = 0, do_test = 0, emergency_rewrite = 0;
+int do_write_eeprom = 0, do_test = 0;
 int nway_advertise = 0, fixed_speed = -1;
 int new_default_media = -1;
 /* Valid with libflash only. */
@@ -132,6 +175,7 @@ static unsigned int opt_flash_show = 0;
 static char	*opt_flash_dumpfile = NULL, *opt_flash_loadfile = NULL;
 
 static unsigned char new_hwaddr[6], set_hwaddr = 0;
+static int emergency_rewrite = 0;
 
 static int scan_proc_pci(int card_num);
 static int parse_media_type(const char *capabilities);
@@ -147,7 +191,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:TvVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -162,7 +206,8 @@ main(int argc, char **argv)
 			if (new_default_media < 0)
 				errflag++;
 			break;
-		case 'G': opt_G++; opt_GPIO = strtol(optarg, NULL, 16); break;
+		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -178,12 +223,13 @@ main(int argc, char **argv)
 			}
 		case 'm': show_mii++;	 break;
 		case 'p':
-			port_base = strtol(optarg, NULL, 16);
+			port_base = strtoul(optarg, NULL, 16);
 			break;
 		case 'q': if (verbose) verbose--;		 break;
 		case 'r': opt_restart++;	break;
 		case 'R': opt_reset++;		break;
 		case 't': chip_type = atoi(optarg);	break;
+		case 'T': do_test++;		break;
 		case 'v': verbose++;		break;
 		case 'V': show_version++;	break;
 		case 'w': do_write_eeprom++;	break;
@@ -275,8 +321,9 @@ static int scan_proc_bus_pci(int card_num)
 				   &pci_bus, &pci_devid, &irq, &pciaddr0, &pciaddr1) <= 0)
 			break;
 		for (i = 0; pcidev_tbl[i].vendor; i++) {
-			if (pci_devid !=
-				(pcidev_tbl[i].vendor << 16) + pcidev_tbl[i].device)
+			if ((pci_devid >> 16) != pcidev_tbl[i].vendor
+				|| (pci_devid & pcidev_tbl[i].device_mask) !=
+				pcidev_tbl[i].device)
 				continue;
 			chip_idx = i;
 			card_cnt++;
@@ -392,6 +439,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -399,7 +447,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtol(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -409,37 +458,37 @@ static int parse_media_type(const char *capabilities)
    0x0800	Power up autosense (check speed only once)
    0x8000	Dynamic Autosense
 */
-/* A table of media names to indices.  This matches the Digital Tulip
-   SROM numbering, primarily because that is the most complete list.
-   Other chips will have to map these number to their internal values.
+/* A table of media names to EEPROM setting.
+   These match the natsemi internal values.
 */
 struct { char *name; int value; } mediamap[] = {
-	{ "10baseT", 0 },
-	{ "10base2", 1 },
-	{ "AUI", 2 },
-	{ "100baseTx", 3 },
-	{ "10baseT-FDX", 0x204 },
-	{ "100baseTx-FDX", 0x205 },
-	{ "100baseT4", 6 },
-	{ "100baseFx", 7 },
-	{ "100baseFx-FDX", 8 },
-	{ "MII", 11 },
-	{ "Autosense", 0x0800 },
+	{ "Force 10baseT half duplex", 0 },
+	{ "Advertise 10 Mbps half and full duplex", 1 },
+	{ "Force 100 Mbps half duplex", 2 },
+	{ "Advertise 10/100 Mbps half duplex", 3 },
+	{ "Force 10 Mbps, full duplex", 4 },
+	{ "Advertise 100 Mbps half and full duplex", 5 },
+	{ "Force 100 Mbps full duplex", 6 },
+	{ "Advertise 10/100 Mbps half and full duplex", 7 },
 	{ 0, 0 },
 };
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -452,7 +501,7 @@ static void natsemi_eeprom(unsigned short *ee_data);
 
 /* Offsets to the DP83815 registers.*/
 enum register_offsets {
-	ChipCmd=0x00, ChipConfig=0x04, EECtrl=0x08, PCIBusCfg=0x0C,
+	ChipCmd=0x00, ChipConfig=0x04, EECtrl=0x08, TestControl=0x0C,
 	IntrStatus=0x10, IntrMask=0x14, IntrEnable=0x18,
 	TxRingPtr=0x20, TxConfig=0x24,
 	RxRingPtr=0x30, RxConfig=0x34,
@@ -461,6 +510,18 @@ enum register_offsets {
 	RxPktErrs=0x60, RxMissed=0x68, RxCRCErrs=0x64,
 };
 
+/* The textual names of the interrupt indications. */
+static const char *intr_names[] ={
+	"Rx Done event", "Rx event", "Receive error", "Rx started",
+	"Receiver idle", "Rx overrun", "Tx complete", "Tx packet in FIFO",
+	"Transmit error", "Tx queue emptied", "Tx underrun",
+	"Statistics counters full", "Driver software intr", "Wake-up packet",
+	"Link change wakeup", "Error summary",
+	"Rx status overrun",  "Intr-17", "Intr-18", "Intr-19",
+	"PCI target abort",  "PCI master abort", "PCI system error",
+	"PCI parity error", "Receiver reset done", "Transmitter reset done",
+};
+
 /* A table for emitting the configuration of a register. */
 struct config_name { int val, mask; const char*name;};
 static struct config_name rcvr_mode[] = {
@@ -472,17 +533,20 @@ static struct config_name rcvr_mode[] = {
 	{0x00, 0x00, "Unknown/invalid"},
 };
 
+/* Autonegotation selection bits in the EEPROM and ChipCmd. */
+static const char *aneg_sel[8] = {
+	"10 Mbps, half", "10 Mbps half and full",
+	"100 Mbps half", "10/100 Mbps half",
+	"10 Mbps, full", "100 Mbps half and full",
+	"100 Mbps full", "10/100 Mbps half and full" };
+
 /* Read the EEPROM and MII Management Data I/O (MDIO) interfaces.
    The EEPROM code is for the common 93c06/46 EEPROMs with 6 bit addresses. */
 
 /* Delay between EEPROM clock transitions.
-   No extra delay is needed with 33Mhz PCI, but future 66Mhz access may need
-   a delay.  Note that pre-2.0.34 kernels had a cache-alignment bug that
-   made udelay() unreliable.
-   The old method of using an ISA access as a delay, __SLOW_DOWN_IO__, is
-   depricated.
+   This flushes the write buffer to prevent quick double-writes.
 */
-#define eeprom_delay(ee_addr)	inl(ee_addr)
+#define eeprom_delay(ee_addr)	inl(ee_addr); inl(ee_addr)
 
 enum EEPROM_Ctrl_Bits {
 	EE_ShiftClk=0x04, EE_DataIn=0x01, EE_ChipSelect=0x08, EE_DataOut=0x02,
@@ -490,18 +554,20 @@ enum EEPROM_Ctrl_Bits {
 #define EE_Write0 (EE_ChipSelect)
 #define EE_Write1 (EE_ChipSelect | EE_DataIn)
 
-/* The EEPROM commands include the alway-set leading bit. */
+/* The EEPROM commands always start with 01.. preamble bits.
+   Commands are prepended to the variable-length address. */
 enum EEPROM_Cmds { EE_WriteCmd=5, EE_ReadCmd=6, EE_EraseCmd=7, };
 
-#define EEPROM_SIZE 64
+/* Values read from the EEPROM, and a new image to write. */
+#define EEPROM_SIZE 256
 static unsigned short eeprom_contents[EEPROM_SIZE];
 static unsigned short new_ee_contents[EEPROM_SIZE];
 #define EEPROM_SA_OFFSET	7
 #define EE_OFFSET		EECtrl	/* Register offset in I/O space. */
 
-/* This executes a generic EEPROM command, typically a write or write enable.
-   It returns the data output from the EEPROM, and thus may also be used for
-   reads. */
+/* Execute a generic EEPROM command.
+   Return all data output from the EEPROM, and thus may be used for
+   EEPROM sizing, read, erase or write. */
 static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 {
 	unsigned retval = 0;
@@ -518,10 +584,10 @@ static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 		outl(dataval, ee_addr);
 		eeprom_delay(ee_addr);
 		if (debug > 2)
-			printf("%X", inb(ee_addr) & 15);
+			printf("%X", inl(ee_addr) & 15);
 		outl(dataval | EE_ShiftClk, ee_addr);
 		eeprom_delay(ee_addr);
-		retval = (retval << 1) | ((inb(ee_addr) & EE_DataOut) ? 1 : 0);
+		retval = (retval << 1) | ((inl(ee_addr) & EE_DataOut) ? 1 : 0);
 	} while (--cmd_len >= 0);
 	outl(EE_ChipSelect, ee_addr);
 
@@ -556,10 +622,8 @@ static int bit_reverse_16(int val)
 /* The abstracted functions for EEPROM access. */
 static int read_eeprom(long ioaddr, int location, int addr_len)
 {
-	/* Bit-reverse the result.  Damn designers mixed big-endian and LE. */
-	return bit_reverse_16(do_eeprom_cmd(ioaddr, ((EE_ReadCmd << addr_len)
-												 | location) << 16,
-										3 + addr_len + 16));
+	return do_eeprom_cmd(ioaddr, ((EE_ReadCmd << addr_len) | location) << 16,
+						 3 + addr_len + 16) & 0xffff;
 }
 
 static void write_eeprom(long ioaddr, int index, int value, int addr_len)
@@ -580,9 +644,10 @@ static void write_eeprom(long ioaddr, int index, int value, int addr_len)
 	i = eeprom_busy_poll(ee_ioaddr);
 	if (debug)
 		printf(" Write finished after %d ticks.\n", i);
-	/* Disable programming.  Note: this command is not instantaneous, but
-	   we check for busy before the next write. */
+	/* Disable programming.  Note: this command is not instantaneous, so
+	   we check for busy before the next op. */
 	do_eeprom_cmd(ioaddr, (0x40 << (addr_len-4)), 3 + addr_len);
+	eeprom_busy_poll(ee_ioaddr);
 }
 
 /*  MII transceiver control section.
@@ -618,93 +683,219 @@ static void natsemi_flash_out(long ioaddr, int offset, int val) {
 
 int natsemi_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 {
+	unsigned int chip_active = inl(ioaddr + IntrEnable);
+	unsigned int chip_config = inl(ioaddr + ChipConfig);
+	unsigned int rx_mode = inl(ioaddr + RxFilterAddr);
+	int ee_addr_len = 6, eeprom_size = 64, eeprom_change = 0;
 	int i;
 
-	if (verbose || show_regs) {
-		unsigned intr_status = inw(ioaddr + IntrStatus),
-			rx_mode = inl(ioaddr + RxFilterAddr);
+	if (chip_active == 0xffffffff && !opt_f) {
+		printf(" * A recognized chip has been found, but it does not "
+			   "appear to exist in\n * I/O space.  Use the"
+			   " '-f' flag to see the register values anyway.\n");
+		return 1;
+	}
 
-		if (opt_a > 1) {
-			printf("NatSemi chip registers at %#lx", ioaddr);
-			for (i = 0; i < 0x100; i += 4) {
-				if ((i & 0x1f) == 0)
-					printf("\n 0x%3.3X:", i);
-				printf(" %8.8x", inl(ioaddr + i));
-			}
-			printf("\n");
+	/* Always show the basic status. */
+	printf(" Natsemi 83815 series with station address ");
+	for (i = 0; i < 6; i+=2) {
+		unsigned short rf_data;
+		outw(i, ioaddr + RxFilterAddr);
+		rf_data = inl(ioaddr + RxFilterData);
+		printf("%2.2x:%2.2x%c", rf_data & 0xff, (rf_data >> 8) & 0xff,
+			   i < 4 ? ':' : '\n');
+	}
+	printf(" Transceiver setting %s %s duplex.\n",
+		   chip_config & 0x2000 ? "Autonegotation advertise" : "force",
+		   aneg_sel[(chip_config>>13) & 7]);
+
+	if (opt_G) {
+		printf("Setting the configuration register to %8.8x.\n", opt_GPIO);
+		outl(opt_GPIO, ioaddr + ChipConfig);
+	}
+
+	if (verbose > 1 || show_regs) {
+		unsigned int intr_status = inl(ioaddr + IntrStatus);
+		char dont_read[8] = {0x10,0x00,0x01,0xff,0x00,0xff,0x00,0x00};
+
+		if (chip_active && !opt_f) {
+		  printf(" This device appears to be active, so some registers"
+				 " will not be read.\n"
+				 " To see all register values use the '-f' flag.\n");
+		} else
+			chip_active = 0;		/* Ignore the chip status with -f */
+
+		printf("%s chip registers at %#lx",
+			   pcidev_tbl[part_idx].part_name, ioaddr);
+		for (i = 0; i < pcidev_tbl[part_idx].io_size; i += 4) {
+			if ((i & 0x1f) == 0)
+				printf("\n 0x%3.3X:", i);
+			if (chip_active && (dont_read[i>>5]) & (1<<((i>>2) & 7)))
+				printf(" ********");
+			else
+				printf(" %8.8x", (int)inl(ioaddr + i));
+		}
+		printf("\n");
+		if (intr_status == 0xffffffff && !opt_f) {
+			printf(" * A recognized chip has been found, but it does not "
+				   "appear to exist in\n * I/O space.  Use the"
+				   " '-f' flag to see the register values anyway.\n");
+			return 1;
+		}
+
+		printf("  %snterrupt sources are pending (%8.8x).\n",
+			   (intr_status & 0x03ff) ? "I": "No i", intr_status);
+		if (intr_status) {
+			for (i = 0; i < sizeof(intr_names)/sizeof(intr_names[0]); i++)
+				if (intr_status & (1<<i))
+					printf("   %s indication.\n", intr_names[i]);
 		}
-		printf("  %snterrupt sources are pending (%4.4x).\n",
-			   (intr_status & 0x03ff) ? "I": "No i", inl(ioaddr + IntrStatus));
-		printf("  Rx filter address is %#x, data is %4.4x.\n",
-			   inl(ioaddr + RxFilterAddr), inl(ioaddr + RxFilterData));
 		for (i = 0; rcvr_mode[i].mask; i++)
 			if ((rx_mode & rcvr_mode[i].mask) == rcvr_mode[i].val) break;
-		printf("  Receive  mode is 0x%2.2x: %s.\n", rx_mode, rcvr_mode[i].name);
-		for (i = 0; i < 16; i+=2) {
-			outw(i, ioaddr + RxFilterAddr);
-			printf("   Rx filter index %d is %4.4x\n",
-				   i, inw(ioaddr + RxFilterData));
-		}
-		outl(rx_mode, ioaddr + RxFilterAddr);
-	}
-
-	if (show_eeprom) {
-		int ee_addr_len = do_eeprom_cmd(ioaddr, EE_ReadCmd << (6+16),
-										3 + 6 + 16) & 0x10000 ? 8 : 6;
-		printf(" EEPROM address length %d.\n", ee_addr_len);
-
-		for (i = 0; i < EEPROM_SIZE; i++)
-			eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
-
-		if (show_eeprom > 1) {
-			printf("EEPROM contents:");
-			for (i = 0; i < EEPROM_SIZE; i++) {
-				if ((i&7) == 0)
-					printf("\n   0x%3.3x: ", i);
-				printf(" %4.4x", eeprom_contents[i]);
+		printf("  Receive mode is 0x%2.2x: %s.\n", rx_mode, rcvr_mode[i].name);
+		if (opt_f || ! chip_active || opt_a > 1) {
+			printf("  Rx filter contents:  ");
+			for (i = 0; i < 16; i+=2) {
+				outw(i, ioaddr + RxFilterAddr);
+				printf(" %4.4x", (int)inl(ioaddr + RxFilterData));
 			}
 			printf("\n");
 		}
-		natsemi_eeprom(eeprom_contents);
-		if (emergency_rewrite  &&  ! set_hwaddr)
-			printf("*** Emergency EEPROM rewrite is only valid when you also "
-				   "specify a new\n*** station address with -H <xx:xx:...>\n");
-		if (set_hwaddr) {
-			u16 backup_ee_contents[] = {
-				0x1385, 0xf311, 0x0b34, 0x41f3, 0x0000, 0x0000, 0x0000,
-				0x0000, 0x0000, 0x0000, 0x1915, };
-			memcpy(new_ee_contents, eeprom_contents, sizeof eeprom_contents);
-			if (emergency_rewrite)
-				memcpy(new_ee_contents, backup_ee_contents,
-					   sizeof backup_ee_contents);
-			/* Grrr, damn designers dropped a bit and "fixed" the hardware
-			   by documenting the EEPROM layout to work that way. */
-			new_ee_contents[6] &= ~0x8000;
-			new_ee_contents[6] |= new_hwaddr[0] & 1 ? 0x8000 : 0;
-			new_ee_contents[7] = (new_hwaddr[0] >> 1) + (new_hwaddr[1]<<7)
-				+ (new_hwaddr[2]<<15);
-			new_ee_contents[8] = (new_hwaddr[2] >> 1) + (new_hwaddr[3]<<7)
-				+ (new_hwaddr[4]<<15);
-			new_ee_contents[9] = (new_hwaddr[4] >> 1) + (new_hwaddr[5]<<7)
-				+ (new_ee_contents[9] & 0x8000);
-			for (i = 0; i < 16; i++)
-				if (new_ee_contents[i] != eeprom_contents[i]) {
-					if (do_write_eeprom)
-						write_eeprom(ioaddr, i, new_ee_contents[i],
-									 ee_addr_len);
-					else
-						printf("Would write %4.4x to replace %4.4x at %d.\n",
-							   new_ee_contents[i], eeprom_contents[i], i);
-				}
+	}
+	outl(rx_mode, ioaddr + RxFilterAddr);
+
+	if (do_test && chip_active && !opt_f)
+		printf(" Running these diagnostic test is not recommended while the "
+			   "device is active.\n"
+			   " To run them anyway, use the '-f' flag.\n");
+	else if (do_test) {
+		int result;
+		outl(0x80, ioaddr + TestControl);
+		for (i = 0; i < 10000; i++) { 		/* Typical 260 iterations. */
+			if (inl(ioaddr + TestControl) & 0x40)
+				break;
 		}
+		result = inl(ioaddr + TestControl);
+		if (debug)
+			printf(" Internal SRAM test took %d ticks.\n", i);
+		printf(" Internal SRAM test result is %4.4x: %s.\n",
+			   result, result & 0x38 ? "failed" : "passed");
+		if (result & 0x38) {
+			if (result & 0x20)
+				fprintf(stderr, " *** Rx FIFO test failed!\n");
+			if (result & 0x10)
+				fprintf(stderr, " *** Tx FIFO test failed!\n");
+			if (result & 0x10)
+				fprintf(stderr, " *** Rx Filter test failed!\n");
+			if ( ! opt_f)
+				return 2;
+		}
+		outl(0x02, ioaddr + TestControl);
+		for (i = 0; i < 10000; i++) { 		/* Typical 2500 iterations. */
+			if ( ! (inl(ioaddr + TestControl) & 0x02))
+				break;
+		}
+		result = inl(ioaddr + TestControl);
+		if (debug)
+			printf(" EEPROM internal test took %d ticks.\n", i);
+		printf(" Configuration EEPROM checksum test result is %4.4x: %s.\n",
+			   result, result & 1 ? "failed" : "passed");
+	}
+	if (set_hwaddr || show_eeprom || new_default_media >= 0) {
+		ee_addr_len = do_eeprom_cmd(ioaddr, EE_ReadCmd << (6+16),
+										3 + 6 + 16) & 0x10000 ? 8 : 6;
+		eeprom_size = 1 << ee_addr_len;
+		printf(" EEPROM address length %d, %d words.\n",
+			   ee_addr_len, eeprom_size);
+		/* Bit-reverse the result.  Damn designers mixed big-endian and LE. */
+		for (i = 0; i < eeprom_size; i++)
+			eeprom_contents[i] =
+				bit_reverse_16(read_eeprom(ioaddr, i, ee_addr_len));
+		memcpy(new_ee_contents, eeprom_contents, sizeof eeprom_contents);
 	}
 
+	if (emergency_rewrite  &&  ! set_hwaddr)
+		printf("*** Emergency EEPROM rewrite is only valid when you also "
+			   "specify a new\n*** station address with -H <xx:xx:...>\n");
+
+	if (set_hwaddr) {
+		u16 backup_ee_contents[] = {
+			0x1385, 0xf311, 0x0b34, 0x41f3, 0x0000, 0x0000, 0x0000, 0x0000,
+			0x0000, 0x0000, 0x1915, };
+		if (emergency_rewrite)
+			memcpy(new_ee_contents, backup_ee_contents,
+				   sizeof backup_ee_contents);
+		/* Grrr, damn designers dropped a bit and "fixed" the hardware
+		   by documenting the EEPROM layout to work that way. */
+		new_ee_contents[6] &= ~0x8000;
+		new_ee_contents[6] |= new_hwaddr[0] & 1 ? 0x8000 : 0;
+		new_ee_contents[7] = (new_hwaddr[0] >> 1) + (new_hwaddr[1]<<7)
+			+ (new_hwaddr[2]<<15);
+		new_ee_contents[8] = (new_hwaddr[2] >> 1) + (new_hwaddr[3]<<7)
+			+ (new_hwaddr[4]<<15);
+		new_ee_contents[9] = (new_hwaddr[4] >> 1) + (new_hwaddr[5]<<7)
+			+ (new_ee_contents[9] & 0x8000);
+		eeprom_change++;
+	}
+	if (new_default_media >= 0) {
+		new_ee_contents[3] &= ~0x0070;
+		new_ee_contents[3] |=  (new_default_media << 4) & 0x00F0;
+		eeprom_change++;
+	}
+	if (eeprom_change) {
+		int sum = 0;
+		/* Calculate the new checksum. */
+		for (i = 0; i < 11; i++) {
+			unsigned short tmp = bit_reverse_16(new_ee_contents[i]);
+			sum += tmp + (tmp>>8);
+		}
+		new_ee_contents[i] = 0xaa00 + (bit_reverse_16(0xaa - sum + 1) >> 8);
+		for (i = 0; i < 16; i++)
+			if (new_ee_contents[i] != eeprom_contents[i]) {
+				if (do_write_eeprom)
+					write_eeprom(ioaddr, i, new_ee_contents[i],
+								 ee_addr_len);
+				else
+					printf("Would write %4.4x to replace %4.4x at %d.\n",
+						   new_ee_contents[i], eeprom_contents[i], i);
+			}
+	}
+
+	if (show_eeprom > 1) {
+		printf("EEPROM contents (%d words):", eeprom_size);
+		for (i = 0; i < eeprom_size; i += 8) {
+			int j;
+			printf("\n0x%2.2x: ", i);
+			for (j = 0; j < 8; j++)
+				printf(" %4.4x", eeprom_contents[i + j]);
+			if (show_eeprom > 2) {
+				printf("  ");
+				for (j = 0; j < 8; j++) {
+					int ew = eeprom_contents[i + j];
+					printf("%c%c",
+						   isprint(ew & 0xff) ? ew & 0xff : '_',
+						   isprint(ew >>   8) ? ew >> 8   : '_' );
+				}
+			}
+		}
+		printf("\n");
+	}
+	if (show_eeprom)
+		natsemi_eeprom(eeprom_contents);
+
 	if (show_mii) {
+#ifdef LIBMII
+		show_mii_details(ioaddr, 1);
+		if (show_mii > 1)
+			monitor_mii(ioaddr, 1);
+#else
 		int mii_reg;
+		printf(" Internal PHY with MII management register.");
 		for (mii_reg = 0; mii_reg < 32; mii_reg++)
 			printf("%s %4.4x", (mii_reg % 8) == 0 ? "\n  " : "",
 				   mdio_read(ioaddr, 1, mii_reg));
 		printf(".\n");
+#endif
 	}
 
 #ifdef LIBFLASH
@@ -742,32 +933,40 @@ int natsemi_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 
 static void natsemi_eeprom(unsigned short *ee_data)
 {
-	int i, sum = 0x55;
+	int i, sum = 0;
+	unsigned int val;
 
 	printf("Decoded EEPROM contents:\n"
 		   "  PCI Subsystem IDs -- Vendor %#4.4x, Device %#4.4x.\n"
 		   "  PCI timer settings -- minimum grant %d, maximum latency %d.\n",
 		   ee_data[0], ee_data[1], ee_data[2] >> 8, ee_data[2] & 0xff);
-	printf("  Wake-On-LAN password "
-		   "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n",
-		   ee_data[4] & 0xff, ee_data[4]>>8,
-		   ee_data[5] & 0xff, ee_data[5]>>8,
-		   ee_data[6] & 0xff, ee_data[6]>>8);
 	printf("  Ethernet MAC Station Address "
 		   "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n",
 		   (ee_data[7]<<1) & 0xff, (ee_data[7]>>7) & 0xff,
 		   ((ee_data[8]<<1) + (ee_data[7]>>15)) & 0xff, (ee_data[8]>>7)&0xff, 
 		   ((ee_data[9]<<1) + (ee_data[8]>>15)) & 0xff, (ee_data[9]>>7)&0xff);
-	for (i = 0; i < 11; i++)
-		sum += ee_data[i] + (ee_data[i]>>8);
-	sum = -sum & 0xff;
-	printf("  EEPROM active region checksum read as %4.4x, vs %2.2x55 "
+	printf("  Wake-On-LAN password "
+		   "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n",
+		   ee_data[4] & 0xff, ee_data[4]>>8,
+		   ee_data[5] & 0xff, ee_data[5]>>8,
+		   ee_data[6] & 0xff, ee_data[6]>>8);
+	val = (ee_data[3] >> 4) & 0x0F;
+	printf("  Transceiver setting 0x--%x-: %s %s duplex.\n",
+		   val, val & 1 ? "advertise" : "force",
+		   aneg_sel[val & 7]);
+	if (val & 8)
+		printf("   Flow control enabled.\n");
+	for (i = 0; i < 11; i++) {
+	    unsigned short tmp = bit_reverse_16(ee_data[i]);
+	    sum += tmp + (tmp>>8);
+	}
+	sum = bit_reverse_16(0xaa - sum + 1) >> 8;
+	printf("  EEPROM active region checksum read as %4.4x, vs aa%2.2x "
 		   "calculated value.\n",
 		   ee_data[11], sum);
 	return;
 }
 
-
 
 /*
  * Local variables:
