@@ -3,7 +3,7 @@
    This is a diagnostic and EEPROM setup program for PCI adapters
    based on the National Semiconductor DP83820 series.
 
-	Written 1998-2001 by Donald Becker.
+	Written 1998-2002 by Donald Becker.
 	Copyright Donald Becker and Scyld Computing Corporation
 
 	This software may be used and distributed according to the terms of
@@ -29,10 +29,44 @@
 */
 
 static char *version_msg =
-"ns820-diag.c:v2.02 5/15/2001 Donald Becker (becker@scyld.com)\n"
+"ns820-diag.c:v2.04 12/2/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: ns820-diag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: ns820-diag [-aEefFGhmqrRtTvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      ns820-diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+			Set a new hardware station address.  Typically diabled, 
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -46,6 +80,7 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <ctype.h>
 
 /* The following are required only with unaligned field accesses. */
 #include <asm/types.h>
@@ -57,7 +92,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -74,11 +109,12 @@ typedef unsigned char u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -101,7 +137,7 @@ extern int  ns820_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 
 /* The table of known chips.
    Because of the bogus /proc/pci interface we must have both the exact
-   name and a PCI vendor/device IDs.
+   name from the kernel, a common name and the PCI vendor/device IDs.
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
@@ -113,7 +149,7 @@ struct pcidev_entry {
 	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_idx);
 } pcidev_tbl[] = {
 	{"National Semiconductor DP83820 series", 0,
-	 0x100b, 0x0022, 0xffff, 0, 256, ns820_diag},
+	 0x100B, 0x0022, 0xffff, 0, 256, ns820_diag},
 	{ 0, 0, 0, 0},
 };
 
@@ -150,7 +186,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:TvVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -166,6 +202,7 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -187,6 +224,7 @@ main(int argc, char **argv)
 		case 'r': opt_restart++;	break;
 		case 'R': opt_reset++;		break;
 		case 't': chip_type = atoi(optarg);	break;
+		case 'T': do_test++;		break;
 		case 'v': verbose++;		break;
 		case 'V': show_version++;	break;
 		case 'w': do_write_eeprom++;	break;
@@ -396,6 +434,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -403,7 +442,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -434,16 +474,20 @@ struct { char *name; int value; } mediamap[] = {
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -459,7 +503,7 @@ void mdio_write(long ioaddr, int phy_id, int location, int value);
 
 /* Offsets to the various registers. */
 enum register_offsets {
-	ChipCmd=0x00, ChipConfig=0x04, EECtrl=0x08, PCIBusCfg=0x0C,
+	ChipCmd=0x00, ChipConfig=0x04, EECtrl=0x08, TestControl=0x0C,
 	IntrStatus=0x10, IntrMask=0x14, IntrEnable=0x18,
 	TxRingPtr=0x20, TxRingPtrHi=0x24, TxConfig=0x28, GeneralIO=0x2C,
 	RxRingPtr=0x30, RxRingPtrHi=0x34, RxConfig=0x38,
@@ -529,24 +573,37 @@ static struct config_name rcvr_mode[] = {
 
 int ns820_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 {
-	int chip_active = inl(ioaddr + IntrEnable);
-	int ee_addr_len = 6, eeprom_size = 64;
-	unsigned rx_mode = inl(ioaddr + RxFilterAddr);
+	unsigned int chip_active = inl(ioaddr + IntrEnable);
+	unsigned int rx_mode = inl(ioaddr + RxFilterAddr);
+	int ee_addr_len = 6, eeprom_size = 64, eeprom_change = 0;
 	int i;
 
+	if (chip_active == 0xffffffff && !opt_f) {
+		printf(" * A recognized chip has been found, but it does not "
+			   "appear to exist in\n * I/O space.  Use the"
+			   " '-f' flag to see the register values anyway.\n");
+		return 1;
+	}
+
 	/* Always show the basic status. */
 	printf(" Natsemi 83820 series with station address ");
 	for (i = 0; i < 6; i+=2) {
 		unsigned short rf_data;
 		outw(i, ioaddr + RxFilterAddr);
-		rf_data = inw(ioaddr + RxFilterData);
-		printf("%2.2x:%2.2x%c", rf_data & 0xff, rf_data >> 8,
+		rf_data = inl(ioaddr + RxFilterData);
+		printf("%2.2x:%2.2x%c", rf_data & 0xff, (rf_data >> 8) & 0xff,
 			   i < 4 ? ':' : '\n');
 	}
-	outl(rx_mode, ioaddr + RxFilterAddr);
 
-	if (verbose || show_regs) {
-		unsigned intr_status, rx_mode = inl(ioaddr + RxFilterAddr);
+	if (opt_G) {
+		printf("Setting the configuration register to %8.8x.\n", opt_GPIO);
+		outl(opt_GPIO, ioaddr + ChipConfig);
+	}
+
+	if (verbose > 1 || show_regs) {
+		unsigned int intr_status = inl(ioaddr + IntrStatus);
+		unsigned int tx_config = inl(ioaddr + TxConfig);
+		char dont_read[8] = {0x10,0x00,0x01,0x00,0x00,0x00,0x00,0x00};
 
 		if (chip_active && !opt_f) {
 		  printf(" This device appears to be active, so some registers"
@@ -555,32 +612,25 @@ int ns820_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		} else
 			chip_active = 0;		/* Ignore the chip status with -f */
 
-		if (opt_a > 1) {
-			/* If reading some registers hoses normal chip operation,
-			   perhaps clearing error counters or interrupts, add those
-			   registers to this bitmap. */
-			char dont_read[8] = {0x10,0x00,0x01,0x00,0x00,0x00,0x00,0x00};
-			printf("%s chip registers at %#lx",
-				   pcidev_tbl[part_idx].part_name, ioaddr);
-			for (i = 0; i < pcidev_tbl[part_idx].io_size; i += 4) {
-				if ((i & 0x1f) == 0)
-					printf("\n 0x%3.3X:", i);
-				if (chip_active && (dont_read[i>>5]) & (1<<((i>>2) & 7)))
-					printf(" ********");
-				else
-					printf(" %8.8x", inl(ioaddr + i));
-			}
-			printf("\n");
+		printf("%s chip registers at %#lx",
+			   pcidev_tbl[part_idx].part_name, ioaddr);
+		for (i = 0; i < pcidev_tbl[part_idx].io_size; i += 4) {
+			if ((i & 0x1f) == 0)
+				printf("\n 0x%3.3X:", i);
+			if (chip_active && (dont_read[i>>5]) & (1<<((i>>2) & 7)))
+				printf(" ********");
+			else
+				printf(" %8.8x", (int)inl(ioaddr + i));
 		}
-		intr_status = inl(ioaddr + IntrStatus);
-		if (intr_status == 0xffff && !opt_f) {
+		printf("\n");
+		if (intr_status == 0xffffffff && !opt_f) {
 			printf(" * A recognized chip has been found, but it does not "
 				   "appear to exist in\n * I/O space.  Use the"
 				   " '-f' flag to see the register values anyway.\n");
 			return 1;
 		}
 
-		printf("  %snterrupt sources are pending (%4.4x).\n",
+		printf("  %snterrupt sources are pending (%8.8x).\n",
 			   (intr_status & 0x03ff) ? "I": "No i", intr_status);
 		if (intr_status) {
 			for (i = 0; i < sizeof(intr_names)/sizeof(intr_names[0]); i++)
@@ -589,20 +639,63 @@ int ns820_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		}
 		for (i = 0; rcvr_mode[i].mask; i++)
 			if ((rx_mode & rcvr_mode[i].mask) == rcvr_mode[i].val) break;
-		printf("  Receive  mode is 0x%2.2x: %s.\n",
-			   rx_mode, rcvr_mode[i].name);
+		printf("  Receive mode is 0x%2.2x: %s.\n", rx_mode, rcvr_mode[i].name);
 		if (opt_f || ! chip_active || opt_a > 1) {
 			printf("  Rx filter contents:  ");
 			for (i = 0; i < 16; i+=2) {
 				outw(i, ioaddr + RxFilterAddr);
-				printf(" %4.4x", inw(ioaddr + RxFilterData));
+				printf(" %4.4x", (int)inl(ioaddr + RxFilterData));
 			}
 			printf("\n");
-			outl(rx_mode, ioaddr + RxFilterAddr);
 		}
+		printf("  Tx threshold %d bytes, FIFO fill %d bytes.\n",
+			   (tx_config & 0xff) << 5, (tx_config & 0xff00) >> 3); 
+	}
+	outl(rx_mode, ioaddr + RxFilterAddr);
+
+	if (do_test && chip_active && !opt_f)
+		printf(" Running these diagnostic test is not recommended while the "
+			   "device is active.\n"
+			   " To run them anyway, use the '-f' flag.\n");
+	else if (do_test) {
+		int result;
+		outl(0x0400, ioaddr + TestControl);
+		for (i = 0; i < 10000; i++) { 		/* Typical 1050 iterations. */
+			if (inl(ioaddr + TestControl) & 0x0200)
+				break;
+		}
+		result = inl(ioaddr + TestControl);
+		if (debug)
+			printf(" Internal SRAM test took %d ticks.\n", i);
+		printf(" Internal SRAM test result is %4.4x: %s.\n",
+			   result, result & 0x01B8 ? "failed" : "passed");
+		if (result & 0x01B8) {
+			if (result & 0x0100)
+				fprintf(stderr, " *** Rx status FIFO test failed!\n");
+			if (result & 0x0080)
+				fprintf(stderr, " *** Rx data FIFO test failed!\n");
+			if (result & 0x20)
+				fprintf(stderr, " *** Tx data FIFO test failed!\n");
+			if (result & 0x0010)
+				fprintf(stderr, " *** Rx multicast filter test failed!\n");
+			if (result & 0x0008)
+				fprintf(stderr, " *** Rx match filter test failed!\n");
+			if ( ! opt_f)
+				return 2;
+		}
+		outl(0x02, ioaddr + TestControl);
+		for (i = 0; i < 10000; i++) { 		/* Typical 3350 iterations. */
+			if ( ! (inl(ioaddr + TestControl) & 0x02))
+				break;
+		}
+		result = inl(ioaddr + TestControl);
+		if (debug)
+			printf(" EEPROM internal test took %d ticks.\n", i);
+		printf(" Configuration EEPROM checksum test result is %4.4x: %s.\n",
+			   result, result & 1 ? "failed" : "passed");
 	}
 
-	if (set_hwaddr || show_eeprom) {
+	if (set_hwaddr || show_eeprom || new_default_media >= 0) {
 		ee_addr_len = do_eeprom_cmd(ioaddr, EE_ReadCmd << (6+16),
 										3 + 6 + 16) & 0x10000 ? 8 : 6;
 		eeprom_size = 1 << ee_addr_len;
@@ -610,23 +703,30 @@ int ns820_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 			   ee_addr_len, eeprom_size);
 		for (i = 0; i < eeprom_size; i++)
 			eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
+		memcpy(new_ee_contents, eeprom_contents, sizeof eeprom_contents);
 	}
 
 	if (emergency_rewrite  &&  ! set_hwaddr)
 		printf("*** Emergency EEPROM rewrite is only valid when you also "
 			   "specify a new\n*** station address with -H <xx:xx:...>\n");
+
 	if (set_hwaddr) {
 		u16 backup_ee_contents[] = {
-			0x4900, 0x1186, 0x340b, 0x0200, 0x0000, 0x00b4, 0x0000,
-			0x0000, 0x0000, 0x24a7, 0x37ba, 0x5000, };
-		int sum = 0x55;
-		memcpy(new_ee_contents, eeprom_contents, sizeof eeprom_contents);
+			0x4900, 0x1186, 0x340b, 0x0200, 0x0000, 0x00b4, 0x0000, 0x0000,
+			0x0000, 0x24a7, 0x37ba, 0x5000, };
 		if (emergency_rewrite)
 			memcpy(new_ee_contents, backup_ee_contents,
 				   sizeof backup_ee_contents);
+		/* NatSemi corrected the endian mix-up and dropped bits of the
+		   earlier series. */
 		new_ee_contents[12] = new_hwaddr[0] + (new_hwaddr[1]<<8);
 		new_ee_contents[11] = new_hwaddr[2] + (new_hwaddr[3]<<8);
 		new_ee_contents[10] = new_hwaddr[4] + (new_hwaddr[5]<<8);
+		eeprom_change++;
+	}
+	if (eeprom_change) {
+		int sum = 0x55;
+		/* Calculate the new checksum. */
 		for (i = 0; i < 13; i++)
 			sum += new_ee_contents[i] + (new_ee_contents[i]>>8);
 		new_ee_contents[13] = ((-sum & 0xff) << 8) | 0x0055;
@@ -642,11 +742,21 @@ int ns820_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 	}
 
 	if (show_eeprom > 1) {
-		printf("EEPROM contents:");
-		for (i = 0; i < eeprom_size; i++) {
-			if ((i & 7) == 0)
-				printf("\n   0x%3.3x: ", i);
-			printf(" %4.4x", eeprom_contents[i]);
+		printf("EEPROM contents (%d words):", eeprom_size);
+		for (i = 0; i < eeprom_size; i += 8) {
+			int j;
+			printf("\n0x%2.2x: ", i);
+			for (j = 0; j < 8; j++)
+				printf(" %4.4x", eeprom_contents[i + j]);
+			if (show_eeprom > 2) {
+				printf("  ");
+				for (j = 0; j < 8; j++) {
+					int ew = eeprom_contents[i + j];
+					printf("%c%c",
+						   isprint(ew & 0xff) ? ew & 0xff : '_',
+						   isprint(ew >>   8) ? ew >> 8   : '_' );
+				}
+			}
 		}
 		printf("\n");
 	}
@@ -758,7 +868,7 @@ static void natsemi820_eeprom(unsigned short *ee_data)
 /* Delay between EEPROM clock transitions.
    This flushes the write buffer to prevent quick double-writes.
 */
-#define eeprom_delay(ee_addr)	inl(ee_addr)
+#define eeprom_delay(ee_addr)	inl(ee_addr); inl(ee_addr)
 
 enum EEPROM_Ctrl_Bits {
 	EE_ShiftClk=0x04, EE_DataIn=0x01, EE_ChipSelect=0x08, EE_DataOut=0x02,
@@ -766,9 +876,9 @@ enum EEPROM_Ctrl_Bits {
 #define EE_Write0 (EE_ChipSelect)
 #define EE_Write1 (EE_ChipSelect | EE_DataIn)
 
-/* This executes a generic EEPROM command, typically a write or write enable.
-   It returns the data output from the EEPROM, and thus may also be used for
-   reads. */
+/* Execute a generic EEPROM command.
+   Return all data output from the EEPROM, and thus may be used for
+   EEPROM sizing, read, erase or write. */
 static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 {
 	unsigned retval = 0;
@@ -834,9 +944,10 @@ static void write_eeprom(long ioaddr, int index, int value, int addr_len)
 	i = eeprom_busy_poll(ee_ioaddr);
 	if (debug)
 		printf(" Write finished after %d ticks.\n", i);
-	/* Disable programming.  Note: this command is not instantaneous, but
-	   we check for busy before the next write. */
+	/* Disable programming.  Note: this command is not instantaneous, so
+	   we check for busy before the next op. */
 	do_eeprom_cmd(ioaddr, (0x40 << (addr_len-4)), 3 + addr_len);
+	eeprom_busy_poll(ee_ioaddr);
 }
 
 /* MII - MDIO  (Media Independent Interface - Management Data I/O) accesses. */
@@ -844,21 +955,24 @@ static void write_eeprom(long ioaddr, int index, int value, int addr_len)
 /*  Unlike the 83815 series, the 820 series has only a serial interface to
 	an external transceiver.
 
-	We Read and write the MII registers using software-generated serial
+	Read and write the MII registers using software-generated serial
 	MDIO protocol.  It is just different enough from the serial EEPROM
 	protocol to not share code.
 
 	The maxium data clock rate is 2.5 Mhz, which can be exceeded unless
 	we force out buffered PCI writes.
 */
-#define mdio_delay()  inl(mdio_addr)
+#define mdio_out(value, mdio_addr) outl(value, mdio_addr)
+#define mdio_in(mdio_addr) inl(mdio_addr)
+#define mdio_delay(mdio_addr) mdio_in(mdio_addr)
+
 #define MDIO_IO_OFFSET EECtrl
 
 #define MDIO_SHIFT_CLK	0x40
 #define MDIO_DATA_BIT	0x10
 /* Derived values. */
-#define MDIO_DATA_WRITE0 0x20
-#define MDIO_DATA_WRITE1 0x30
+#define MDIO_WRITE0 0x20
+#define MDIO_WRITE1 0x30
 #define MDIO_ENB		0x00
 #define MDIO_ENB_IN		0x00
 #define MDIO_DATA_READ	0x10
@@ -870,10 +984,10 @@ static void mdio_sync(long ioaddr)
 	int i;
 
 	for (i = 32; i >= 0; i--) {
-		outl(MDIO_ENB | MDIO_DATA_WRITE1, mdio_addr);
-		mdio_delay();
-		outl(MDIO_ENB | MDIO_DATA_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_ENB | MDIO_WRITE1, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB | MDIO_WRITE1 | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	return;
 }
@@ -881,9 +995,8 @@ static void mdio_sync(long ioaddr)
 int mdio_read(long ioaddr, int phy_id, int location)
 {
 	long mdio_addr = ioaddr + MDIO_IO_OFFSET;
-	int i;
-	int read_cmd = (0xf6 << 10) | (phy_id << 5) | location;
-	int retval = 0;
+	int mii_cmd = (0xf6 << 10) | (phy_id << 5) | location;
+	int i, retval = 0;
 
 	if (verbose > 2)		/* Debug: 5 */
 		printf(" mdio_read(%#lx, %d, %d)..", ioaddr, phy_id, location);
@@ -891,26 +1004,26 @@ int mdio_read(long ioaddr, int phy_id, int location)
 	mdio_sync(ioaddr);
 	/* Shift the read command bits out. */
 	for (i = 17; i >= 0; i--) {
-		int dataval = (read_cmd & (1 << i)) ? MDIO_DATA_WRITE1 : MDIO_DATA_WRITE0;
+		int dataval = (mii_cmd & (1 << i)) ? MDIO_WRITE1 : MDIO_WRITE0;
 		if (verbose > 3)		/* Debug: 5 */
-			printf("%d", (read_cmd & (1 << i)) ? 1 : 0);
+			printf("%d", (mii_cmd & (1 << i)) ? 1 : 0);
 
-		outl(MDIO_ENB | dataval, mdio_addr);
-		mdio_delay();
-		outl(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
-		if (verbose > 3) printf(" %x", (inl(mdio_addr) >> 16) & 0x0f);
-		mdio_delay();
+		mdio_out(MDIO_ENB | dataval, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
+		if (verbose > 3) printf(" %x", (mdio_in(mdio_addr) >> 16) & 0x0f);
+		mdio_delay(mdio_addr);
 	}
-	if (verbose > 3) printf("-> %x", (inl(mdio_addr) >> 16) & 0x0f);
+	if (verbose > 3) printf("-> %x", (mdio_in(mdio_addr) >> 16) & 0x0f);
 
 	/* Read the two transition, 16 data, and wire-idle bits. */
 	for (i = 19; i > 0; i--) {
-		outl(MDIO_ENB_IN, mdio_addr);
-		mdio_delay();
-		retval = (retval << 1) | ((inl(mdio_addr) & MDIO_DATA_READ) ? 1 : 0);
-		if (verbose > 3) printf(" %x", (inl(mdio_addr) >> 16) & 0x0f);
-		outl(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_ENB_IN, mdio_addr);
+		mdio_delay(mdio_addr);
+		retval = (retval << 1) | ((mdio_in(mdio_addr) & MDIO_DATA_READ) ? 1 : 0);
+		if (verbose > 3) printf(" %x", (mdio_in(mdio_addr) >> 16) & 0x0f);
+		mdio_out(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	if (verbose > 3) printf(" == %4.4x.\n", retval);
 	return (retval>>1) & 0xffff;
@@ -919,25 +1032,25 @@ int mdio_read(long ioaddr, int phy_id, int location)
 void mdio_write(long ioaddr, int phy_id, int location, int value)
 {
 	long mdio_addr = ioaddr + MDIO_IO_OFFSET;
+	int mii_cmd = (0x5002 << 16) | (phy_id << 23) | (location<<18) | value;
 	int i;
-	int cmd = (0x5002 << 16) | (phy_id << 23) | (location<<18) | value;
 
 	/* Establish sync by sending 32 logic ones. */
 	mdio_sync(ioaddr);
 	/* Shift the command bits out. */
 	for (i = 31; i >= 0; i--) {
-		int dataval = (cmd & (1 << i)) ? MDIO_DATA_WRITE1 : 0;
-		outl(MDIO_ENB | dataval, mdio_addr);
-		mdio_delay();
-		outl(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		int dataval = (mii_cmd & (1 << i)) ? MDIO_WRITE1 : MDIO_WRITE0;
+		mdio_out(MDIO_ENB | dataval, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB | dataval | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	/* Clear out extra bits. */
 	for (i = 2; i > 0; i--) {
-		outl(MDIO_ENB_IN, mdio_addr);
-		mdio_delay();
-		outl(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
-		mdio_delay();
+		mdio_out(MDIO_ENB_IN, mdio_addr);
+		mdio_delay(mdio_addr);
+		mdio_out(MDIO_ENB_IN | MDIO_SHIFT_CLK, mdio_addr);
+		mdio_delay(mdio_addr);
 	}
 	return;
 }
