@@ -4,7 +4,7 @@
    based on the AMD PCnet/PCI series chips.
    Note: The EEPROM setup code is not implemented.
 
-	Copyright 1998-2000 by Donald Becker.
+	Copyright 1998-2001 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -28,7 +28,7 @@
 */
 
 static char *version_msg =
-"pcnet-diag.c:v2.00 4/19/2000 Donald Becker (becker@scyld.com)\n"
+"pcnet-diag.c:v2.02 3/31/2001 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
 "Usage: pcnet-diag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
@@ -103,17 +103,19 @@ enum pcnet_flags { PCNET_HAS_MII };
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
-	char *proc_pci_name;
-	char *part_name;
+	const char *part_name;
+	const char *proc_pci_name;
 	int vendor, device, device_mask;
 	int flags;
 	int io_size;
-	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_num);
+	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_idx);
 } pcidev_tbl[] = {
-	{ "AMD PCnet/PCI series", "AMD PCnet32, unknown type",
-	  0x1022, 0x0000, 0xffff, 0, 32, pcnet_diag },
-	{ "AMD PCnet/PCI 978", "AMD PCnet-Home 79c978 Homenet",
+	{ "AMD PCnet-PCI 79c970", "AMD 79C970",
+	  0x1022, 0x2000, 0xffff, 0, 32, pcnet_diag },
+	{ "AMD PCnet-Home 79c978 Homenet", 0,
 	  0x1022, 0x2001, 0xffff, 0, 32, pcnet_diag },
+	{ "AMD PCnet/PCI series", "AMD PCnet32, unknown type",
+	  0x1022, 0x0000, 0, 0, 32, pcnet_diag },
 	{ 0, 0, 0, 0},
 };
 
@@ -254,7 +256,8 @@ static char bogus_iobase[] =
 static char bogus_irq[] =
 "This chip has not been assigned a valid IRQ, and will not function.\n"
 " This must be fixed in the PCI BIOS setup.  The device driver has no way\n"
-" of changing the PCI IRQ settings.\n";
+" of changing the PCI IRQ settings.\n"
+" See  http://www.scyld.com/expert/irq-conflict.html  for more information.\n";
 
 static int scan_proc_bus_pci(int card_num)
 {
@@ -276,7 +279,7 @@ static int scan_proc_bus_pci(int card_num)
 				   &pci_bus, &pci_devid, &irq, &pciaddr0, &pciaddr1) <= 0)
 			break;
 		for (i = 0; pcidev_tbl[i].vendor; i++) {
-			if (pci_devid !=
+			if ((pci_devid & (pcidev_tbl[i].device_mask | 0xffff0000)) !=
 				(pcidev_tbl[i].vendor << 16) + pcidev_tbl[i].device)
 				continue;
 			chip_idx = i;
@@ -335,8 +338,9 @@ static int scan_proc_pci(int card_num)
 					int i;
 					if (debug)
 						printf("Named ethernet controller %s.\n", chip_name);
-					for (i = 0; pcidev_tbl[i].proc_pci_name; i++)
-						if (strncmp(pcidev_tbl[i].proc_pci_name, chip_name,
+					for (i = 0; pcidev_tbl[i].part_name; i++)
+						if (pcidev_tbl[i].proc_pci_name  &&
+							strncmp(pcidev_tbl[i].proc_pci_name, chip_name,
 									strlen(pcidev_tbl[i].proc_pci_name))
 							== 0) {
 							state = 2;
@@ -424,7 +428,7 @@ struct { char *name; int value; } mediamap[] = {
 	{ "100baseFx", 7 },
 	{ "100baseFx-FDX", 8 },
 	{ "MII", 11 },
-	{ "HomePNA", 16 },
+	{ "HomePNA", 18 },
 	{ "Autosense", 0x0800 },
 	{ 0, 0 },
 };
