@@ -3,7 +3,7 @@
    This is a diagnostic and EEPROM setup program for Ethernet adapters
    based on the RealTek RTL8129 and RTL8139 chips.
 
-	Copyright 1998-2001 by Donald Becker.
+	Copyright 1998-2003 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -24,14 +24,53 @@
 
 	Common-sense licensing statement: Using any portion of this program in
 	your own program means that you must give credit to the original author
-	and release the resulting code under the GPL.
+	and release the resulting code under the GPL.  To use this code under
+	other terms requires an explicit license from the copyright holder.
 */
 
 static char *version_msg =
-"rtl8139-diag.c:v2.03 5/15/2001 Donald Becker (becker@scyld.com)\n"
+"rtl8139-diag.c:v2.11 4/22/2003 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: rtl8139-diag [-aEefFmqrRtsvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: rtl8139-diag [-aDEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:89:ab
+        Set a new hardware station address.  Typically disabled for safety.
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+ To read and write the boot BIOS extension Flash ROM
+   -B       Show the first few bytes of the ROM
+   -L FILE  Load the Flash from FILE.
+   -S FILE  Store the Flash image to FILE.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -45,6 +84,8 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <ctype.h>
 
 #if defined(__linux__)  &&  __GNU_LIBRARY__ == 1
 #include <asm/io.h>			/* Newer libraries use <sys/io.h> instead. */
@@ -52,7 +93,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -63,17 +104,18 @@ extern int (*flash_in_hook)(long addr, int offset);
 extern void (*flash_out_hook)(long addr, int offset, int val);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -81,6 +123,7 @@ struct option longopts[] = {
 	{"show-mii",  0, 0, 'm'},	/* Dump MII management registers. */
 	{"port-base",  1, 0, 'p'},	/* Use the specified I/O address. */
 	{"quiet",	0, 0, 'q'},		/* Decrease verbosity */
+	{"restart",	0, 0, 'r'},		/* Restart autonegotiation. */
 	{"reset",	0, 0, 'R'},		/* Reset the transceiver. */
 	{"chip-type",  1, 0, 't'},	/* Assume the specified chip type index. */
 	{"test",	0, 0, 'T'},		/* Do register and SRAM test. */
@@ -115,10 +158,12 @@ struct pcidev_entry {
 	  0x1113, 0x1211, 0xffff, 0, 128, rtl81x9_diag},
 	{ "Accton MPX5030 (RealTek RTL8139)", 0,
 	  0x1113, 0x1211, 0xffff, 0, 128, rtl81x9_diag },
-	{ "D-Link DFE-538TX (RealTek RTL8139)", 0,
-	  0x1186, 0x1300, 0xffff, 0, 128, rtl81x9_diag },
+	{ "D-Link DFE-530TX+ (RealTek RTL8139C)", 0,
+	  0x1186, 0x1300, 0xffff, 0, 256, rtl81x9_diag },
 	{ "RealTek (unknown chip type)", 0,
 	  0x10ec, 0x8100, 0xff00, 0, 128, rtl81x9_diag },
+	{ "Compaq HNE-300 (RTL8139C)", 0,
+	  0x021b, 0x8139, 0xffff, 0, 128, rtl81x9_diag },
 	{ 0, 0, 0, 0},
 };
 
@@ -154,7 +199,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -170,6 +215,7 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -400,6 +446,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -407,7 +454,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -438,16 +486,20 @@ struct { char *name; int value; } mediamap[] = {
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -455,10 +507,13 @@ static int get_media_index(const char *name)
 /* Chip-specific section. */
 
 /* The chip-specific section for the RealTek 8129/8139 diagnostic. */
+static int eeprom_addr_len(long ioaddr);
 static int read_eeprom(long ioaddr, int location, int addr_len);
-static void parse_eeprom(unsigned short *eeprom);
+static void write_eeprom(long ioaddr, int index, int value, int addr_len);
+#if defined(ENABLE_EEWRITE)
 static int do_update(long ioaddr, unsigned short *ee_values,
-			  int index, char *field_name, int new_value);
+			  int index, char *field_name, int new_value, int ee_addr_len);
+#endif
 static void mdio_sync(long ioaddr);
 int mdio_read(long ioaddr, int phy_id, int location);
 
@@ -486,18 +541,76 @@ enum RTL8129_registers {
 	FlashReg=0x54, GPPinData=0x58, GPPinDir=0x59, MII_SMI=0x5A, HltClk=0x5B,
 	MultiIntr=0x5C, TxSummary=0x60,
 	MII_BMCR=0x62, MII_BMSR=0x64, NWayAdvert=0x66, NWayLPAR=0x68,
-	NWayExpansion=0x6A, FlashAccess=0x74,
+	NWayExpansion=0x6A, FlashAccess=0xD4,
 };
 
-/* Values read from the EEPROM, and the new image. */
+/* Utility functions for showing register states. */
+/* Table formats.  Tables are zero terminated. */
+struct bits_to_name { int bits;  const char *msg; };
+struct bitfield_to_name { int val, mask; const char *msg;};
+
+/* Use MSG_TBL to describe the bits set in VAL. */
+void msg_if_set(const int val, const struct bits_to_name msg_tbl[])
+{
+	int i;
+	for (i = 0; msg_tbl[i].bits; i++)
+		if (msg_tbl[i].bits & val)
+			printf(" %s\n", msg_tbl[i].msg);
+}
+
+/* Find the first matching MSG_TBL entry to describe VAL.
+   Return the message from that entry.
+   The message table is terminated with an entry using a zero mask, and
+   the string from entry is the no-match default.
+ */
+const char *msg_match(const int val, const struct bitfield_to_name msg_tbl[])
+{
+	int i;
+	for (i = 0; msg_tbl[i].mask; i++)
+		if ((val & msg_tbl[i].mask) == msg_tbl[i].val)
+			break;
+	return msg_tbl[i].msg;
+}
+
+/* Chip version from TxConfig. */
+static const struct bitfield_to_name chip_ver_table[] = {
+	{ 0x60000000, 0xfc800000, "rtl8139"},
+	{ 0x70000000, 0xfc800000, "rtl8139A"},
+	{ 0x70800000, 0xfc800000, "rtl8139A-G"},
+	{ 0x78800000, 0xfc800000, "rtl8139B-G"}, /* No manual entry! */
+	{ 0x78000000, 0xfc800000, "rtl8139B"},
+	{ 0x7C000000, 0xfc800000, "rtl8130"},
+	{ 0x74000000, 0xfc800000, "rtl8139C"},
+	{ 0x74800000, 0xfc800000, "rtl8139C+"},
+	{ 0x00000000, 0x00000000, "Unknown version"},
+};
+/* Receiver filter in RxConfig. */
+static const struct bitfield_to_name rcvr_mode[] = {
+	{0x0a, 0x3f, "Unicast and no multicast"},
+	{0x0e, 0x3f, "Normal unicast and hashed multicast"},
+	{0x0f, 0x3f, "Promiscuous"},
+	{0x00, 0x3f, "Reception disabled"},
+	{0x00, 0x00, "Unknown/invalid"},
+};
+/* Tx status registers. */
+static const struct bits_to_name tx_status_table[] = {
+	{ 0x80000000, "  Tx carrier lost"},
+	{ 0x40000000, "  Tx aborted"},
+	{ 0x20000000, "  Tx out-of-window collision"},
+	{ 0x00004000, "  Tx FIFO underrun"},
+	{ 0, 0},
+};
+
+static int rtl81x9_eeprom(long ioaddr, int flags);
+/* Values read from the EEPROM, and a new image to write. */
 static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len);
-#define EE_READ_CMD		(6)
-#define EEPROM_SIZE 64			/* Really the maximum index we read. */
-unsigned short eeprom_contents[EEPROM_SIZE];
-unsigned short new_ee_contents[EEPROM_SIZE];
 
 #ifdef LIBFLASH
-/* Support for Flash operations. */
+/* Support for Flash operations.
+   These functions are exported to the flash library.
+   Note that early chip versions use FlashAccess offset 0x74, while current
+   chips use 0xD4.
+ */
 static int rtl_flash_in(long ioaddr, int offset) {
 	outl(0x1C0000 | (offset & 0x1ffff), ioaddr + FlashAccess);
 	return inl(ioaddr + FlashAccess) >> 24;
@@ -511,17 +624,17 @@ static void rtl_flash_out(long ioaddr, int offset, int val) {
 int rtl81x9_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 {
 	int flags = pcidev_tbl[part_idx].flags; 		/* Capabilities. */
+	unsigned intr_status = inw(ioaddr + IntrStatus);
+	int chip_cmd = inb(ioaddr + ChipCmd);
 	int chip_active = 0;
 	int i;
 
 	/* It's mostly safe to examine the registers and EEPROM during
 	   operation.  But warn the user, and make then pass '-f'. */
-	if ((inb(ioaddr + ChipCmd) & 0x000C) != 0x0000)
+	if ((chip_cmd & 0x000C) != 0x0000)
 		chip_active = 1;
 
 	if (show_regs) {
-		unsigned intr_status;
-
 		if (chip_active && !opt_f) {
 			printf("The RealTek chip appears to be active, so some registers"
 				   " will not be read.\n"
@@ -536,14 +649,59 @@ int rtl81x9_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 			printf(" %8.8x", inl(ioaddr + i));
 		}
 		printf(".\n");
+	}
 
-		intr_status = inw(ioaddr + IntrStatus);
-		if (intr_status == 0xffff && !opt_f) {
-			printf(" * A recognized chip has been found, but it does not "
-				   "appear to exist in\n * I/O space.  Use the"
-				   " '-f' flag to see the register values anyway.\n");
-			return 1;
-		}
+	if (intr_status == 0xffff && !opt_f) {
+		printf(" * A recognized chip has been found, but it does not "
+			   "appear to exist in\n * I/O space.  Use the"
+			   " '-f' flag to show the status anyway.\n");
+		return 1;
+	}
+
+	/* Always show the basic status. */
+	printf("Realtek station address ");
+	for (i = 0; i < 5; i++)
+		printf("%2.2x:", inb(ioaddr + i));
+	printf("%2.2x, ", inb(ioaddr + i));
+
+	{
+		int tx_config = inl(ioaddr + TxConfig);
+		int rx_config = inl(ioaddr + RxConfig);
+		unsigned char cfg0 = inb(ioaddr + Config0);
+		unsigned char cfg1 = inb(ioaddr + Config1);
+		const char *xcvr_mode[] = {
+			"MII", "an invalid transceiver", "MII/symbol",
+			"4B/5B scambler"};
+
+		printf("chip type '%s'.\n"
+			   "  Receiver configuration: %s\n"
+			   "     Rx FIFO threshold %d bytes, maximum burst %d bytes, "
+			   "%dKB ring\n"
+			   "  Transmitter %sabled with %s settings, maximum burst"
+			   " %d bytes.\n",
+			   msg_match(tx_config, chip_ver_table),
+			   msg_match(rx_config, rcvr_mode),
+			   16 << ((rx_config >> 13) & 7), 16 << ((rx_config >> 8) & 7),
+			   8 << ((rx_config >> 11) & 3),
+			   chip_cmd & 0x04 ? "en" : "dis",
+			   (tx_config & 0x30700f0) ? "NONSTANDARD!" : "normal",
+			   16 << ((tx_config >> 8) & 7));
+		if (opt_a)
+			for (i = 0; i < 4; i++) {
+				int tx_status = inl(ioaddr + TxStat0 + i*4);
+				printf("    Tx entry #%d status %8.8x %scomplete, %d bytes.\n",
+					   i, tx_status, tx_status & 0x8000 ? "" : "in",
+					   tx_status & 0x7ff);
+				msg_if_set(tx_status, tx_status_table);
+			}
+		printf("  Flow control: Tx %sabled  Rx %sabled.\n",
+			   inb(ioaddr + GPPinData) & 0x80 ? "en" : "dis",
+			   inb(ioaddr + GPPinData) & 0x40 ? "en" : "dis");
+		printf("  The chip configuration is 0x%2.2x 0x%2.2x, %s %s-duplex"
+			   " mode.\n",
+			   cfg0, cfg1,
+			   cfg0 & 0x20 ? "10baseT" : xcvr_mode[cfg0>>6],
+			   cfg1 & 0x40 ? "full" : "half");
 		printf("  %snterrupt sources are pending.\n",
 			   (intr_status & 0x03ff) ? "I": "No i");
 		if (intr_status) {
@@ -551,64 +709,9 @@ int rtl81x9_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 				if (intr_status & (1<<i))
 					printf("   %s indication.\n", intr_names[i]);
 		}
-		{
-			unsigned char cfg0 = inb(ioaddr + Config0);
-			unsigned char cfg1 = inb(ioaddr + Config1);
-			const char *xcvr_mode[] = {
-				"MII", "an invalid transceiver", "MII/symbol",
-				"4B/5B scambler"};
-			printf(" The chip configuration is 0x%2.2x 0x%2.2x, %s %s-duplex"
-				   " mode.\n",
-				   cfg0, cfg1,
-				   cfg0 & 0x20 ? "10baseT" : xcvr_mode[cfg0>>6],
-				   cfg1 & 0x40 ? "full" : "half");
-		}
 	}
 
-	/* Read the EEPROM. */
-	{
-		int testval = do_eeprom_cmd(ioaddr,
-									(((EE_READ_CMD<<8)|0) << 16) | 0xffff,
-									3 + 8 + 16);
-		int addr_size = (read_eeprom(ioaddr, 0, 6) & 0xffff) == 0x8129 ? 6 : 8;
-		printf("EEPROM size test returned %d, %#x / %#x.\n", addr_size,
-			   testval, read_eeprom(ioaddr, 0xff, 8));
-		for (i = 0; i < EEPROM_SIZE; i++)
-			eeprom_contents[i] = read_eeprom(ioaddr, i, addr_size);
-	}
-
-	/* The user will usually want to see the interpreted EEPROM contents. */
-	if (show_eeprom)
-		parse_eeprom(eeprom_contents);
-	if (show_eeprom > 1) {
-		unsigned short sum = 0;
-		printf("EEPROM contents:");
-		for (i = 0; i < EEPROM_SIZE; i++) {
-			printf("%s %4.4x", (i & 7) == 0 ? "\n ":"",
-				   eeprom_contents[i]);
-			sum += eeprom_contents[i];
-		}
-		printf("\n The word-wide EEPROM checksum is %#4.4x.\n", sum);
-	}
-	if (set_hwaddr) {
-		for (i = 0; i < 3; i++) {
-			do_update(ioaddr, eeprom_contents, 7 + i, "Station address",
-					  (new_hwaddr[i*2]<<8) + new_hwaddr[i*2+1]);
-		}
-	}
-	if (new_default_media) {
-		int new_media = eeprom_contents[6] & ~0x0031;
-		switch (new_default_media) {
-		case 0x0000: new_media |= 0x0000; break;			/* 10baseT */
-		case 0x0204: new_media |= 0x0001; break;			/* 10baseT-FDX */
-		case 0x0003: new_media |= 0x0020; break;			/* 100baseTx */
-		case 0x0205: new_media |= 0x0021; break;			/* 100baseTx-FDX */
-		case 0x0800: new_media |= 0x0011; break;			/* Autosense */
-		default: new_media = eeprom_contents[6];			/* Unchanged */
-		}
-		do_update(ioaddr, eeprom_contents, 6, "Default Media",
-				  (new_hwaddr[i*2]<<8) + new_hwaddr[i*2+1]);
-	}
+	rtl81x9_eeprom(ioaddr, flags);
 
 	/* Show up to four (not just the on-board) PHYs. */
 	if (show_mii  &&  (flags & HasMII)) {
@@ -652,13 +755,13 @@ int rtl81x9_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 			   "   False carrier sense counter   0x%4.4x.\n"
 			   "   NWay test register            0x%4.4x.\n"
 			   "   Receive frame error count     0x%4.4x.\n",
-			   inw(ioaddr + MII_BMSR), inw(ioaddr + MII_BMCR),
+			   inw(ioaddr + MII_BMCR), inw(ioaddr + MII_BMSR),
 			   inw(ioaddr + NWayAdvert), inw(ioaddr + NWayLPAR),
 			   inw(ioaddr + NWayExpansion), inw(ioaddr + 0x6C),
 			   inw(ioaddr + 0x6E), inw(ioaddr + 0x70), inw(ioaddr + 0x72));
 #ifdef LIBMII
 		if (show_mii > 1) {
-			show_mii_details(ioaddr, -1);
+			show_mii_details(ioaddr, 32);
 			if (show_mii > 2) monitor_mii(ioaddr, 32);
 		}
 #endif
@@ -698,9 +801,153 @@ int rtl81x9_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 	return 0;
 }
 
+
+/* Configuration EEPROM contents and updates. */
+#define MAX_EEPROM_SIZE 256		/* In 16 bit words. */
+#define EEPROM_SA_OFFSET	7
+unsigned short eeprom_contents[MAX_EEPROM_SIZE];
+unsigned short new_ee_contents[MAX_EEPROM_SIZE];
+
+
+static void rtl8139_eeprom_decode(unsigned short *eeprom)
+{
+	unsigned char *p = (void *)eeprom;
+	int i, sum = 0;
+
+	printf("Decoded EEPROM contents:\n"
+		   "   PCI IDs -- Vendor %#4.4x, Device %#4.4x.\n"
+		   "   PCI Subsystem IDs -- Vendor %#4.4x, Device %#4.4x.\n"
+		   "   PCI timer settings -- minimum grant %d, maximum latency %d.\n"
+		   "  General purpose pins --  direction 0x%2.2x  value 0x%2.2x.\n"
+		   "  Station Address ",
+		   eeprom[1], eeprom[2], eeprom[3], eeprom[4],
+		   p[10], p[11], p[13], p[12] );
+	for (i = 14; i < 19; i++)
+		printf("%2.2X:", p[i]);
+	printf("%2.2X.\n"
+		   "  Configuration register 0/1 -- 0x%2.2x / 0x%2.2x.\n",
+		   p[i], p[21], p[22]);
+	for (i = 0; i < 24; i++)
+		sum += p[i];
+	printf(" EEPROM active region checksum is %4.4x.\n", sum);
+	return;
+}
+
+static int rtl81x9_eeprom(long ioaddr, int flags)
+{
+	int ee_addr_len = 6, eeprom_size = 64;
+	int i;
+
+	/* Read the EEPROM. */
+	if (set_hwaddr || show_eeprom || new_default_media >= 0) {
+		ee_addr_len = eeprom_addr_len(ioaddr);
+		eeprom_size = 1 << ee_addr_len;
+
+		for (i = 0; i < eeprom_size; i++)
+			eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
+		memcpy(new_ee_contents, eeprom_contents, eeprom_size << 1);
+	}
+
+	/* The user will usually want to see the interpreted EEPROM contents. */
+	if (show_eeprom)
+		rtl8139_eeprom_decode(eeprom_contents);
+	if (show_eeprom > 1) {
+		int previous_row_same = 0;
+
+		printf("EEPROM contents (%d words):", eeprom_size);
+		for (i = 0; i < eeprom_size; i += 8) {
+			int j;
+			if (i == 0 ||		/* Always show first row. */
+				(memcmp(eeprom_contents+i, eeprom_contents+i - 8, 16) != 0)) {
+				previous_row_same = 0;
+				printf("\n 0x%2.2x: ", i);
+				for (j = 0; j < 8; j++)
+					printf(" %4.4x", eeprom_contents[i + j]);
+				if (show_eeprom > 2) {
+					printf("  ");
+					for (j = 0; j < 8; j++) {
+						int ew = eeprom_contents[i + j];
+						printf("%c%c",
+							   isprint(ew & 0xff) ? ew & 0xff : '_',
+							   isprint(ew >>   8) ? ew >> 8   : '_' );
+					}
+				}
+			} else if ( ! previous_row_same) {
+				previous_row_same = 1;
+				printf("\n      ...");
+			}
+		}
+		printf("\n");
+	}
+	if (emergency_rewrite  &&  ! set_hwaddr)
+		printf("*** Emergency EEPROM rewrite is only valid when you also "
+			   "specify a new\n*** station address with -H <xx:xx:...>\n");
+	if (set_hwaddr) {
+		u16 backup_ee_contents[] = {
+			0x8129, 0x10ec, 0x8139, 0x10ec, 0x8139, 0x4020, 0xe110, 0x2301,
+			0x6745, 0xab89, 0x4d10, 0xf7c2, 0x8001, 0xb388, 0x58fa, 0x0708,
+			0xd843, 0xa438, 0xd843, 0xa438, 0xd843, 0xa438, 0xd843, 0xa438, };
+		if (emergency_rewrite)
+			memcpy(new_ee_contents, backup_ee_contents,
+				   sizeof backup_ee_contents);
+		for (i = 0; i < 3; i++)
+			new_ee_contents[i + EEPROM_SA_OFFSET] =
+				(new_hwaddr[i*2+1]<<8) + new_hwaddr[i*2];
+	}
+	if (new_default_media >= 0) {
+		int new_media = eeprom_contents[6] & ~0x0031;
+		switch (new_default_media) {
+		case 0x0000: new_media |= 0x0000; break;			/* 10baseT */
+		case 0x0204: new_media |= 0x0001; break;			/* 10baseT-FDX */
+		case 0x0003: new_media |= 0x0020; break;			/* 100baseTx */
+		case 0x0205: new_media |= 0x0021; break;			/* 100baseTx-FDX */
+		case 0x0800: new_media |= 0x0011; break;			/* Autosense */
+		default:
+			printf(" Unsupported new media type (%4.4x), leaving the setting"
+				   " unchanged.\n", new_default_media);
+			new_media = eeprom_contents[6];			/* Unchanged */
+		}
+		if (new_ee_contents[6] != new_media) {
+			printf("  Changing the default media setting to %4.4x.\n",
+				   new_media);
+			new_ee_contents[6] = new_media;
+		}
+	}
+	/* To set another EEPROM value do  new_ee_contents[xxx] = new_value; */
+	if (set_hwaddr || new_default_media >= 0) {
+		int not_written_msg = 0;
+		for (i = 0; i < eeprom_size; i++)
+			if (new_ee_contents[i] != eeprom_contents[i]) {
+				if (do_write_eeprom)
+					write_eeprom(ioaddr, i, new_ee_contents[i],
+								 ee_addr_len);
+				else {
+					printf("   Would write %4.4x to replace 0x%4.4x at "
+						   "offset %d.\n",
+						   new_ee_contents[i], eeprom_contents[i], i);
+					not_written_msg++;
+				}
+			}
+		if (not_written_msg)
+			printf(" To actually write these values pass the '-w' flag.\n"
+				   "  Note that this is a permanent action!\n"
+				   "  You should record the original EEPROM values in"
+				   " a safe place.\n");
+	}
+	return 0;
+}
 
 /* Serial EEPROM section. */
 
+/* The EEPROM commands always start with 01.. preamble bits.
+   Commands are prepended to the variable-length address. */
+enum EEPROM_Cmds { EE_WriteCmd=5, EE_ReadCmd=6, EE_EraseCmd=7, };
+
+enum EEPROM_Ctrl_Bits {
+	EE_ShiftClk=0x04, EE_ChipSelect=0x88, EE_DataOut=0x02, EE_DataIn=0x01,
+	EE_Write0=0x88, EE_Write1=0x8A,
+};
+
 /*  EEPROM_Ctrl bits. */
 #define EE_SHIFT_CLK	0x04	/* EEPROM shift clock. */
 #define EE_CS			0x08	/* EEPROM chip select. */
@@ -711,51 +958,15 @@ int rtl81x9_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 #define EE_ENB			(0x80 | EE_CS)
 
 /* Delay between EEPROM clock transitions.
-   No extra delay is needed with 33Mhz PCI, but 66Mhz may change this.
- */
-
-#define eeprom_delay()	inl(ee_addr)
+   This forces out buffered PCI writes.
+*/
+#define eeprom_delay(ee_addr)	inb(ee_addr)
 
 /* The EEPROM commands include the alway-set leading bit. */
 #define EE_WRITE_CMD	(5)
 #define EE_READ_CMD		(6)
 #define EE_ERASE_CMD	(7)
 
-static int read_eeprom(long ioaddr, int location, int addr_len)
-{
-	int i;
-	unsigned retval = 0;
-	long ee_addr = ioaddr + Cfg9346;
-	int read_cmd = location | (EE_READ_CMD << addr_len);
-
-	outb(EE_ENB & ~EE_CS, ee_addr);
-	outb(EE_ENB, ee_addr);
-
-	/* Shift the read command bits out. */
-	for (i = 4 + addr_len; i >= 0; i--) {
-		int dataval = (read_cmd & (1 << i)) ? EE_DATA_WRITE : 0;
-		outb(EE_ENB | dataval, ee_addr);
-		eeprom_delay();
-		outb(EE_ENB | dataval | EE_SHIFT_CLK, ee_addr);
-		eeprom_delay();
-		retval = (retval << 1) | ((inb(ee_addr) & EE_DATA_READ) ? 1 : 0);
-	}
-	outb(EE_ENB, ee_addr);
-	eeprom_delay();
-
-	for (i = 16; i > 0; i--) {
-		outb(EE_ENB | EE_SHIFT_CLK, ee_addr);
-		eeprom_delay();
-		retval = (retval << 1) | ((inb(ee_addr) & EE_DATA_READ) ? 1 : 0);
-		outb(EE_ENB, ee_addr);
-		eeprom_delay();
-	}
-
-	/* Terminate the EEPROM access. */
-	outb(~EE_CS, ee_addr);
-	return retval;
-}
-
 /* This executes a generic EEPROM command, typically a write or write enable.
    It returns the data output from the EEPROM, and thus may also be used for
    reads. */
@@ -767,64 +978,106 @@ static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 	if (debug > 1)
 		printf(" EEPROM op 0x%x: ", cmd);
 
-	outb(EE_ENB | EE_SHIFT_CLK, ee_addr);
+	outb(EE_ChipSelect | EE_ShiftClk, ee_addr);
 
 	/* Shift the command bits out. */
 	do {
-		short dataval = (cmd & (1 << cmd_len)) ? EE_WRITE_1 : EE_WRITE_0;
+		short dataval = (cmd & (1 << cmd_len)) ? EE_Write1 : EE_Write0;
 		outb(dataval, ee_addr);
-		eeprom_delay();
+		eeprom_delay(ee_addr);
 		if (debug > 2)
 			printf("%X", inl(ee_addr) & 15);
-		outb(dataval | EE_SHIFT_CLK, ee_addr);
-		eeprom_delay();
-		retval = (retval << 1) | ((inb(ee_addr) & EE_DATA_READ) ? 1 : 0);
+		outb(dataval | EE_ShiftClk, ee_addr);
+		eeprom_delay(ee_addr);
+		retval = (retval << 1) | ((inb(ee_addr) & EE_DataIn) ? 1 : 0);
 	} while (--cmd_len >= 0);
-	outb(EE_ENB, ee_addr);
+	outb(EE_ChipSelect, ee_addr);
 
 	/* Terminate the EEPROM access. */
-	outb(EE_ENB & ~EE_CS, ee_addr);
-	outb(~EE_CS, ee_addr);
+	outb(0x80, ee_addr);
+	outb(0, ee_addr);
 	if (debug > 1)
 		printf(" EEPROM result is 0x%5.5x.\n", retval);
 	return retval;
 }
 
+/* Wait for the EEPROM to finish what it is doing.
+   Note that we take the already computed ioaddr of the control register. */
+static int eeprom_busy_poll(long ee_ioaddr)
+{
+	int i;
+	outb(EE_ChipSelect, ee_ioaddr);
+	for (i = 0; i < 10000; i++)			/* Typical 2000 ticks */
+		if (inb(ee_ioaddr) & EE_DataIn)
+			break;
+	return i;
+}
+
+/* The abstracted functions for EEPROM access. */
+
+/* Return the number of address bits this EEPROM accepts.
+   Note that we cannot count on a leading transition bit, and must look
+   explicitly for the initial value.
+   This approach might fail with a blank or corrupted EEPROM.
+ */
+static int eeprom_addr_len(long ioaddr)
+{
+	int testval = do_eeprom_cmd(ioaddr,
+								(((EE_READ_CMD<<8)|0) << 16) | 0xffff,
+								3 + 8 + 16);
+	int addr_size = (read_eeprom(ioaddr, 0, 6) & 0xffff) == 0x8129 ? 6 : 8;
+
+	if (debug)
+		printf("   Testvalue %x address size %d, test read 0x%4.4x.\n",
+			   testval, addr_size, read_eeprom(ioaddr, 0, 6));
+
+	return addr_size;
+}
+
+static int read_eeprom(long ioaddr, int location, int addr_len)
+{
+	return do_eeprom_cmd(ioaddr, ((EE_ReadCmd << addr_len) | location) << 16,
+						 3 + addr_len + 16) & 0xffff;
+}
+
+static void write_eeprom(long ioaddr, int index, int value, int addr_len)
+{
+	long ee_ioaddr = ioaddr + Cfg9346;
+	int i;
+
+	/* Poll for previous op finished. */
+	eeprom_busy_poll(ee_ioaddr);
+
+	/* Enable programming modes. */
+	do_eeprom_cmd(ioaddr, (0x4f << (addr_len-4)), 3 + addr_len);
+	/* Do the actual write. */ 
+	do_eeprom_cmd(ioaddr,
+				  (((EE_WriteCmd<<addr_len) | index) << 16) | (value & 0xffff),
+				  3 + addr_len + 16);
+	i = eeprom_busy_poll(ee_ioaddr);
+	if (debug)
+		printf(" Write finished after %d ticks.\n", i);
+	/* Disable programming.  Note: this command is not instantaneous, so
+	   we check for busy before the next op. */
+	do_eeprom_cmd(ioaddr, (0x40 << (addr_len-4)), 3 + addr_len);
+	eeprom_busy_poll(ee_ioaddr);
+}
+
+#if defined(ENABLE_EEWRITE)
 /* Caution!  This routine works!
    It will permanently change the EEPROM contents.
    "You could put an eye out with that thing." */
-static int do_update(long ioaddr, unsigned short *ee_values,
-					 int index, char *field_name, int new_value)
+static int do_update(long ioaddr, unsigned short *ee_values, int index,
+					 char *field_name, int new_value, int ee_addr_len)
 {
 	if (ee_values[index] == new_value)
 		return 0;
 	if (do_write_eeprom) {
-		int ee_addr_size = read_eeprom(ioaddr, 0, 8) == 0x8129 ? 8 : 6;
-		int i;
-		u16 newval;
 		printf("Writing new %s entry 0x%4.4x to offset %d.\n",
 			   field_name, new_value, index);
-		/* Enable programming modes. */
-		do_eeprom_cmd(ioaddr, (0x4f << (ee_addr_size-4)), 3+ee_addr_size);
-		/* Do the actual write. */ 
-		do_eeprom_cmd(ioaddr,
-					  (((EE_WRITE_CMD<<ee_addr_size)|index) << 16) | new_value,
-					  3 + ee_addr_size + 16);
-		/* Poll for write finished. */
-		outb(EE_ENB, ioaddr + Cfg9346);
-		for (i = 0; i < 10000; i++)			/* Typical 2000 ticks */
-			if (inb(ioaddr + Cfg9346) & EE_DATA_READ)
-				break;
-		if (debug)
-			printf(" Write finished after %d ticks.\n", i);
-		/* Disable programming. */
-		do_eeprom_cmd(ioaddr, (0x40 << (ee_addr_size-4)), 3 + ee_addr_size);
-		/* And read the result. */
-		newval = do_eeprom_cmd(ioaddr,
-							   (((EE_READ_CMD<<ee_addr_size)|index) << 16)
-							   | 0xffff, 3 + ee_addr_size + 16);
+		write_eeprom(ioaddr, index, new_value, ee_addr_len);
 		printf("  New %s value at offset %d is %4.4x.\n",
-			   field_name, index, newval);
+			   field_name, index, read_eeprom(ioaddr, index, ee_addr_len));
 	} else
 		printf(" Would write new %s entry 0x%4.4x to offset %d, the "
 			   "current value is 0x%4.4x.\n",
@@ -832,6 +1085,7 @@ static int do_update(long ioaddr, unsigned short *ee_values,
 	ee_values[index] = new_value;
 	return 1;
 }
+#endif
 
 
 /* MII serial management: mostly bogus for now. */
@@ -938,29 +1192,6 @@ void mdio_write(long ioaddr, int phy_id, int location, int value)
 	}
 	return;
 }
-
-static void parse_eeprom(unsigned short *eeprom)
-{
-	unsigned char *p = (void *)eeprom;
-	int i, sum = 0;
-
-	printf("Parsing the EEPROM of a RealTek chip:\n"
-		   "  PCI IDs -- Vendor %#4.4x, Device %#4.4x, Subsystem %#4.4x.\n"
-		   "  PCI timer settings -- minimum grant %d, maximum latency %d.\n"
-		   "  General purpose pins --  direction 0x%2.2x  value 0x%2.2x.\n"
-		   "  Station Address ",
-		   eeprom[1], eeprom[2], eeprom[3], p[10], p[11], p[13], p[12] );
-	for (i = 14; i < 19; i++)
-		printf("%2.2X:", p[i]);
-	printf("%2.2X.\n"
-		   "  Configuration register 0/1 -- 0x%2.2x / 0x%2.2x.\n",
-		   p[i], p[21], p[22]);
-	for (i = 0; i < 24; i++)
-		sum += p[i];
-	printf(" EEPROM active region checksum is %4.4x.\n", sum);
-	return;
-}
-
 
 /*
  * Local variables:
