@@ -6,7 +6,7 @@
 	Work-alikes from Lite-On, LinkSys, Macronix, ASIX, Compex, STmicro,
 	ADMtek, Davicom and (ugh) Xircom.
 
-	Copyright 1998-2000 by Donald Becker.
+	Copyright 1998-2003 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -26,14 +26,54 @@
 
 	Common-sense licensing statement: Using any portion of this program in
 	your own program means that you must give credit to the original author
-	and release the resulting code under the GPL.
+	and release the resulting code under the GPL.  To use this code under
+	other terms requires an explicit license from the copyright holder.
 */
 
 static char *version_msg =
-"tulip-diag.c:v2.08 5/15/2001 Donald Becker (becker@scyld.com)\n"
+"tulip-diag.c:v2.17 5/6/2003 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: tulip-diag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: tulip-diag [-aDEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type (with '-p'), use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -F, --new-interface N	Set the default transceiver type.
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:89:ab
+        Set a new hardware station address.  Typically disabled for safety.
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+ To read and write the boot BIOS extension Flash ROM
+   -B       Show the first few bytes of the ROM
+   -L FILE  Load the Flash from FILE.
+   -S FILE  Store the Flash image to FILE.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -46,8 +86,9 @@ static char *usage_msg =
 #include <fcntl.h>
 #include <getopt.h>
 #include <string.h>
-#include <strings.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <ctype.h>
 
 #include <asm/types.h>
 #include <asm/unaligned.h>
@@ -58,7 +99,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -69,17 +110,18 @@ extern int (*flash_in_hook)(long addr, int offset);
 extern void (*flash_out_hook)(long addr, int offset, int val);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -87,6 +129,7 @@ struct option longopts[] = {
 	{"show-mii",  0, 0, 'm'},	/* Dump MII management registers. */
 	{"port-base",  1, 0, 'p'},	/* Use the specified I/O address. */
 	{"quiet",	0, 0, 'q'},		/* Decrease verbosity */
+	{"restart",	0, 0, 'r'},		/* Restart autonegotiation. */
 	{"reset",	0, 0, 'R'},		/* Reset the transceiver. */
 	{"chip-type",  1, 0, 't'},	/* Assume the specified chip type index. */
 	{"verbose",	0, 0, 'v'},		/* Verbose mode */
@@ -156,6 +199,14 @@ struct pcidev_entry {
 	0x13d1, 0xab02, 0xffff, COMET_MACADDR, 256, tulip_diag },
   { "ADMtek AL985 Centaur (Linksys CardBus)", 0,
 	0x13d1, 0xab03, 0xffff, COMET_MACADDR, 256, tulip_diag },
+  { "ADMtek AL985 Centaur (OEM CardBus)", 0,
+	0x13d1, 0xab08, 0xffff, COMET_MACADDR, 256, tulip_diag },
+  { "ADMtek AL985 Centaur (Vendor 0x13d1 Unknown version)", 0,
+	0x13d1, 0xab00, 0xfff0, COMET_MACADDR, 256, tulip_diag },
+  { "Linksys PCM200 v4 CardBus (ADMtek AL985 Centaur)", 0,
+	0x1737, 0xab08, 0xffff, COMET_MACADDR, 256, tulip_diag },
+  { "Linksys PCM200 v3 CardBus (ADMtek AL985 Centaur)", 0,
+	0x1737, 0xab09, 0xffff, COMET_MACADDR, 256, tulip_diag },
   { "STMicro STE10/100 Comet", 0,
 	0x104a, 0x0981, 0xffff, COMET_MACADDR, 256, tulip_diag },
   { "STMicro STE10/100A Comet", 0,
@@ -170,6 +221,8 @@ struct pcidev_entry {
 	0x1282, 0x9100, 0xfff0, CSR12_IS_GPIO | IS_DAVICOM, 128, tulip_diag },
   { "Conexant LANfinity", 0,
 	0x14f1, 0x1803, 0xffff, IS_CONEXANT, 256, tulip_diag },
+  { "3Com 3cSOHO100B-TX (ADMtek Centaur)", 0,
+	0x10b7, 0x9300, 0xffff, COMET_MACADDR, 256, tulip_diag },
 	{ 0, 0, 0, 0},
 };
 
@@ -204,7 +257,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -220,6 +273,7 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -450,6 +504,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -457,7 +512,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -489,19 +545,23 @@ struct { char *name; int value; } mediamap[] = {
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0) {
 			if (debug)
 				fprintf(stderr, "Using media index %d for '%s'.\n", i, name);
 			return i;
 		}
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -515,16 +575,16 @@ static void mdio_sync(long ioaddr);
 static void setup_nway_xcvr(long ioaddr);
 static void tulip_eeprom(long ioaddr, int flags);
 static int read_eeprom(long ioaddr, int location, int addr_len);
-static void parse_eeprom(unsigned char *ee_data, int part_idx);
+static void parse_eeprom(long ioaddr, unsigned char *ee_data, int part_idx);
 static void liteon_eeprom(unsigned char *ee_data, int part_idx);
 static void admtek_eeprom(unsigned char *ee_data, int part_idx);
 static void conexant_eeprom(unsigned char *ee_data, int part_idx);
 static void davicom_eeprom(unsigned char *ee_data, int part_idx);
 static unsigned int calculate_checksum1(u16 *eeprom_contents, int len);
 static unsigned int ether_crc_le(void *ptr, int length);
-int do_update(long ioaddr, unsigned short *ee_values, int index,
+int do_update(long ioaddr, u_int16_t *ee_values, int index,
 			  char *field_name, int new_value);
-static void check_for_intel_cb(long ioaddr, unsigned short *eeprom_contents);
+static void check_for_intel_cb(long ioaddr, u_int16_t *eeprom_contents);
 
 
 /* Offsets to the Command and Status Registers, "CSRs".  All accesses
@@ -564,8 +624,8 @@ static int current_part_idx = 0; /* Global, for mdio_{read,write,sync}() */
 static int default_media_offset = -1;
 
 /* Values read from the EEPROM, and the new image to write. */
-static unsigned short eeprom_contents[EEPROM_SIZE];
-static unsigned short new_ee_contents[EEPROM_SIZE];
+static u_int16_t eeprom_contents[EEPROM_SIZE];
+static u_int16_t new_ee_contents[EEPROM_SIZE];
 
 /* Support for Flash operations. */
 static int tulip_flash_in(long ioaddr, int offset) {
@@ -608,21 +668,21 @@ int tulip_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		for (i = 0; i < 64; i += 8)
 			printf(" %8.8x", inl(ioaddr + i));
 		printf("\n");
-	}
-	if (show_regs && (!if_active || opt_f)) {
-		int num_regs = pcidev_tbl[part_idx].io_size;
-		printf(" 0x40:");
-		for (; i < 128; i += 8)
-			printf(" %8.8x", inl(ioaddr + i));
-		if (i < num_regs) {
-			/* Extended registers are _not_ quadword aligned. */
-			printf("\n Extended registers:");
-			for (; i < num_regs; i += 4) {
-				if (i % 32 == 0) printf("\n %2.2x:", i);
+		if (!if_active || opt_f) {
+			int num_regs = pcidev_tbl[part_idx].io_size;
+			printf(" 0x40:");
+			for (; i < 128; i += 8)
 				printf(" %8.8x", inl(ioaddr + i));
+			if (i < num_regs) {
+				/* Extended registers are _not_ quadword aligned. */
+				printf("\n Extended registers:");
+				for (; i < num_regs; i += 4) {
+					if (i % 32 == 0) printf("\n 0x%2.2x:", i);
+					printf(" %8.8x", inl(ioaddr + i));
+				}
 			}
+			printf("\n");
 		}
-		printf("\n");
 	}
 	if (!opt_f  &&  inl(ioaddr + CSR5) == 0xffffffff) {
 		printf(" * A recognized chip has been found, but it does not appear"
@@ -650,10 +710,9 @@ int tulip_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 				   (csr6 & 0x00800000 ? "100mbps-SYM/PCS" : "MII"),
 				   (csr6 & 0x01800000)==0x01800000 ? " 100baseTx scrambler":"",
 				   csr6 & 0x0200 ? "full" : "half");
-		printf(" Transmit %s, Receive %s, %s-duplex.\n",
+		printf(" Transmit %s, Receive %s.\n",
 			   csr6 & 0x2000 ? "started" : "stopped",
-			   csr6 & 0x0002 ? "started" : "stopped",
-			   csr6 & 0x0200 ? "full" : "half");
+			   csr6 & 0x0002 ? "started" : "stopped");
 		printf("  The Rx process state is '%s'.\n",
 			   rx_state[(csr5 >> 17) & 7]);
 		printf("  The Tx process state is '%s'.\n",
@@ -677,8 +736,16 @@ int tulip_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		}
 	}
 
-	if (flags & HAS_NWAY)
-		printf("  The NWay status register is %8.8x.\n", inl(ioaddr + CSR12));
+	if (flags & HAS_NWAY) {
+		int csr12 = inl(ioaddr + CSR12);
+		printf("  The NWay status register is %8.8x.\n", csr12);
+		if (csr12 & 0x8000)
+			printf("    Link Partner ability: %4.4x.\n", csr12>>16);
+		if ((csr12 & 0x0002) == 0)
+			printf("    100baseTx link good.\n");
+		if ((csr12 & 0x0003) == 0)
+			printf("    10baseT link good.\n");
+	}
 
 	if (flags & COMET_MACADDR) {
 		printf("  Comet MAC address registers %8.8x %8.8x\n"
@@ -711,11 +778,13 @@ int tulip_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 		printf(".\n");
 	}
 
+	tulip_eeprom(ioaddr, flags);
 	if (opt_GPIO) {
 		printf("Setting the GPIO register %8.8x.\n", opt_GPIO);
 		outl(opt_GPIO, ioaddr + CSR15);
+		printf("The GPIO register is now %8.8x.\n", inl(ioaddr + CSR15));
+		return 0;
 	}
-	tulip_eeprom(ioaddr, flags);
 
 	/* Show up to four (not just the on-board) PHYs. */
 	if ((has_mii && verbose) || show_mii) {
@@ -808,7 +877,6 @@ int tulip_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 	return 0;
 }
 
-#include <ctype.h>
 static void tulip_eeprom(long ioaddr, int flags)
 {
 	int eeprom_words = 64;
@@ -830,12 +898,12 @@ static void tulip_eeprom(long ioaddr, int flags)
 	} else if (flags & PNIC_EEPROM) {
 		for (i = 0; i < eeprom_words; i++) {
 			int value, boguscnt = 100000;
-			unsigned short sum;
+			u_int16_t sum;
 			outl(0x600 | i, ioaddr + 0x98);
 			do
 				value = inl(ioaddr + CSR9);
 			while (value < 0  && --boguscnt > 0);
-			((unsigned short *)eeprom_contents)[i] = value;
+			((u_int16_t *)eeprom_contents)[i] = value;
 			sum += value & 0xffff;
 		}
 	} else {
@@ -864,10 +932,10 @@ static void tulip_eeprom(long ioaddr, int flags)
 			admtek_eeprom((unsigned char *)eeprom_contents, flags);
 		else if (flags & IS_CONEXANT)
 			conexant_eeprom((unsigned char *)eeprom_contents, flags);
-		else if (flags & IS_CONEXANT)
+		else if (flags & IS_DAVICOM)
 			davicom_eeprom((unsigned char *)eeprom_contents, flags);
 		else
-			parse_eeprom((unsigned char *)eeprom_contents, flags);
+			parse_eeprom(ioaddr, (unsigned char *)eeprom_contents, flags);
 
 		if (show_eeprom > 1) {
 			int block_crc = (calculate_checksum1(eeprom_contents, 8)>>8) & 0xff;
@@ -884,8 +952,8 @@ static void tulip_eeprom(long ioaddr, int flags)
 					for (j = 0; j < 8; j++) {
 						int ew = eeprom_contents[i + j];
 						printf("%c%c",
-							   isalnum(ew & 0xff) ? ew & 0xff : '_',
-							   isalnum(ew >>   8) ? ew >> 8   : '_' );
+							   isprint(ew & 0xff) ? ew & 0xff : '_',
+							   isprint(ew >>   8) ? ew >> 8   : '_' );
 					}
 				}
 			}
@@ -907,8 +975,10 @@ static void tulip_eeprom(long ioaddr, int flags)
 }
 
 /* Reading a serial EEPROM is a "bit" grungy, but we work our way through:->.*/
-/* This code is a "nasty" timing loop, but PC compatible machines are
-   *supposed* to delay an ISA-compatible period for the SLOW_DOWN_IO macro.  */
+/*
+ * This is not precisely a "delay" between EEPROM clock transitions, but
+ * rather avoiding buffered PCI writes.
+ */
 #define eeprom_delay()	inl(ee_addr)
 
 /*  EEPROM_Ctrl bits. */
@@ -951,6 +1021,7 @@ static int read_eeprom(long ioaddr, int location, int addr_len)
 		retval = (retval << 1) | ((inl(ee_addr) & EE_DATA_READ) ? 1 : 0);
 	}
 	outl(EE_ENB, ee_addr);
+	eeprom_delay();
 	if (debug > 2)
 		printf(" :%X:", inl(ee_addr) & 15);
 
@@ -1007,7 +1078,8 @@ static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 /* Read and write the MII registers using software-generated serial
    MDIO protocol.  It is just different enough from the EEPROM protocol
    to not share code.  The maxium data clock rate is 2.5 Mhz. */
-#define mdio_delay()  inl(mdio_addr)	/* Extra bus turn-around as a delay. */
+/* Use a bus turn-around to avoid PCI write buffers pairing writes. */
+#define mdio_delay()  inl(mdio_addr)
 
 #define MDIO_SHIFT_CLK	0x10000
 #define MDIO_DATA_WRITE0 0x00000
@@ -1227,7 +1299,7 @@ static void setup_nway_xcvr(long ioaddr)
 
 #define CRC1_POLYNOMIAL 0x07		/* x^8 + x^2 + x^1 + 1 */
 static unsigned int
-calculate_checksum1(unsigned short *eeprom, int len)
+calculate_checksum1(u_int16_t *eeprom, int len)
 {
 	u16 crc = 0xffff;
 	int i, bit;
@@ -1261,7 +1333,7 @@ static unsigned int ether_crc_le(void *ptr, int length)
 	return crc;
 }
 
-int do_update(long ioaddr, unsigned short *ee_values,
+int do_update(long ioaddr, u_int16_t *ee_values,
 			  int index, char *field_name, int new_value)
 {
 	if (ee_values[index] == new_value)
@@ -1319,7 +1391,7 @@ struct mediainfo {
 	unsigned char *mii;
 };
 
-static void parse_eeprom(unsigned char *ee_data, int flags)
+static void parse_eeprom(long ioaddr, unsigned char *ee_data, int flags)
 {
 	unsigned char csr12;
 	unsigned char *p;
@@ -1393,10 +1465,14 @@ static void parse_eeprom(unsigned char *ee_data, int flags)
 				p += 1;
 		} else if (p[0] & 0x80) { /* Extended block */
 			int blk_type = p[1];
-			if (show_eeprom)
-				printf("  Media %s, block type %d, length %d.\n",
-					   p[1] & 1  ?  "MII"  :  medianame[p[2] & 31],
-					   blk_type, p[0] & 0x7f);
+			static const char * const blockname[8] = {
+				0, "MII", 0, "MII", 0, "Device Reset", "PHY Shutdown",
+				"PNA Magic" };
+			printf("  Media %s, block type %d, length %d.\n",
+				   blk_type > 8 ? "<Unknown/Invalid>"
+				   : blockname[blk_type] ?  blockname[blk_type]
+				   :  medianame[p[2] & 31],
+				   blk_type, p[0] & 0x7f);
 			switch (blk_type) {
 			case 0:
 				printf("   AUI or SYM transceiver for %s (media type %d).\n"
@@ -1501,9 +1577,21 @@ static void parse_eeprom(unsigned char *ee_data, int flags)
 				for( i = 0; i < p[2]; i++)
 					printf(" %2.2x%2.2x", p[i*2 + 4], p[i*2 + 3]);
 				printf(".\n");
+				if (opt_reset) {
+					printf("  Performing a transceiver reset.\n");
+					for (i = 0; i < p[2]; i++)
+						outw(((p[i*2 + 4]<<8) + p[i*2 + 3]) << 16,
+							 ioaddr + CSR15);
+					printf("  New CSR15 value is %8.8x.\n",
+						   inl(ioaddr + CSR15));
+				}
 				break;
 			case 6:
-				printf("   Disconnect reset, sequence length %d:", p[3]);
+				printf("   Disconnect reset for%s%s%s%s sequence length %d:",
+					   p[2] & 1 ? "Link fail, " : "",
+					   p[2] & 2 ? "D1 sleep, " : "",
+					   p[2] & 4 ? "D2 sleep, " : "",
+					   p[2] & 8 ? "D3 sleep, " : "", p[3]);
 				for( i = 0; i < p[3]; i++)
 					printf(" %2.2x%2.2x", p[i*2 + 5], p[i*2 + 4]);
 				printf(".\n");
@@ -1571,7 +1659,7 @@ admtek_media[] = { {0x0000, "10baseT"}, {0x0001, "BNC"}, {0x0002, "AUI"},
 				   {0x0400, "Autosense"}, {0xFFFF, "Default"}, {0, ""}, };
 static void admtek_eeprom(unsigned char *ee_data, int part_idx)
 {
-	unsigned short *eew = (void *)ee_data;
+	u_int16_t *eew = (void *)ee_data;
 	int i;
 	printf("  Ethernet MAC Station Address "
 		   "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n",
@@ -1607,7 +1695,17 @@ static void davicom_eeprom(unsigned char *ee_data, int part_idx)
 		   ~ether_crc_le(ee_data, 126) & 0xffff, ee_data[63]);
 }
 
-static void check_for_intel_cb(long ioaddr, unsigned short *eeprom_contents)
+static const char msg_no_station_addr[] =
+"  However this card's station address could not be located in Card
+ Information Structure, therefore no corrected EEPROM table information can
+ be written.
+";
+
+static const char msg_bogus_intel_eeprom[] =
+"You have an Intel CardBus card with an incomplete EEPROM!
+";
+
+static void check_for_intel_cb(long ioaddr, u_int16_t *eeprom_contents)
 {
 	unsigned const char cis_addr_prefix[4] = {0x22, 0x08, 0x04, 0x06};
 	unsigned const short media_ctrl_tbl[] = {
@@ -1617,11 +1715,13 @@ static void check_for_intel_cb(long ioaddr, unsigned short *eeprom_contents)
 	unsigned char cis_addr_tuple[10];
 	int i;
 
-	printf("You have an Intel CardBus card with an incomplete EEPROM.\n");
+	if (eeprom_contents[9] == 0x0103)
+		return;
+	printf(msg_bogus_intel_eeprom);
 	for (i = 0; i < 10; i++)
 		cis_addr_tuple[i] = tulip_flash_in(ioaddr, 0xeb + i);
 	if (memcmp(cis_addr_tuple, cis_addr_prefix, 4)) {
-		printf(" I could not locate the station address.\n");
+		printf(msg_no_station_addr);
 		return;
 	}
 	printf("The station address is ");
