@@ -2,9 +2,9 @@
 
    This is a diagnostic and EEPROM setup program for PCI adapters
    based on the following chips:
-	VIA Rhine vt86c100 and vt3043 Ethernet controller.
+	VIA Rhine vt86c100, vt3043, vt6102, and vt6105 Ethernet controllers.
 
-	Copyright 1998-2001 by Donald Becker.
+	Copyright 1998-2002 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -31,10 +31,48 @@
 */
 
 static char *version_msg =
-"via-diag.c:v2.06 5/22/2001 Donald Becker (becker@scyld.com)\n"
+"via-diag.c:v2.09 12/28/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: etherdiag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: via-diag [-aEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent configuration EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+			Set a new hardware station address.  Typically diabled, 
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+ To read and write the boot BIOS extension Flash ROM
+   -B       Show the first few bytes of the ROM
+   -L FILE  Load the Flash from FILE.
+   -S FILE  Store the Flash image to FILE.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -48,6 +86,8 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <ctype.h>
 
 /* The following are required only with unaligned field accesses. */
 #include <asm/types.h>
@@ -59,7 +99,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -70,17 +110,18 @@ extern int (*flash_in_hook)(long addr, int offset);
 extern void (*flash_out_hook)(long addr, int offset, int val);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -104,7 +145,7 @@ enum rhine_flags { CanFlash=1, };
 
 /* The table of known chips.
    Because of the bogus /proc/pci interface we must have both the exact
-   name and a PCI vendor/device IDs.
+   name from the kernel, a common name and the PCI vendor/device IDs.
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
@@ -121,8 +162,12 @@ struct pcidev_entry {
 	{"VIA VT3043 Rhine",
 	 " Ethernet controller: VIA Technologies Unknown device",
 	 0x1106, 0x3043, 0xffff, 0, 128, rhine_diag},
-	{"VIA VT3065 Rhine-II", 0,
-	 0x1106, 0x3065, 0xffff, CanFlash, 128, rhine_diag},
+	{"VIA VT6102 Rhine-II (type 3065)", 0,
+	 0x1106, 0x3065, 0xffff, CanFlash, 256, rhine_diag},
+	{"VIA VT6105 Rhine-III (type 3053 prototype)", 0,
+	 0x1106, 0x3053, 0xffff, CanFlash, 256, rhine_diag},
+	{"VIA VT6105 Rhine-III (type 3106)", 0,
+	 0x1106, 0x3106, 0xffff, CanFlash, 256, rhine_diag},
 	{ 0, 0, 0, 0},
 };
 
@@ -159,7 +204,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -175,6 +220,7 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -405,6 +451,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -412,7 +459,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -437,23 +485,26 @@ struct { char *name; int value; } mediamap[] = {
 	{ "100baseFx", 7 },
 	{ "100baseFx-FDX", 8 },
 	{ "MII", 11 },
-	{ "HomePNA", 18 },
 	{ "Autosense", 0x0800 },
 	{ 0, 0 },
 };
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -462,8 +513,10 @@ static int get_media_index(const char *name)
 
 /* The chip-specific section for the VIA Rhine. */
 
+static int eeprom_addr_len(long ioaddr);
 static int read_eeprom(long ioaddr, int location, int addr_len);
 static void write_eeprom(long ioaddr, int index, int value, int addr_len);
+static void rhine_eeprom(unsigned short *ee_data);
 int mdio_read(long ioaddr, int phy_id, int location);
 void mdio_write(long ioaddr, int phy_id, int location, int value);
 
@@ -629,6 +682,7 @@ int rhine_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 			eeprom_contents[0x1e] = eeprom_contents[0x1f] = 0x73;
 			outb(0x00, ioaddr + EECtrl);
 		} else {
+			ee_addr_len = eeprom_addr_len(ioaddr);
 			eeprom_size = 1 << ee_addr_len;
 			for (i = 0; i < eeprom_size; i++)
 				eeprom_contents[i] = read_eeprom(ioaddr, i, ee_addr_len);
@@ -658,14 +712,28 @@ int rhine_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 
 	if (show_eeprom > 1) {
 		printf("EEPROM contents%s:",
-			   inb(ioaddr + EECtrl)&0x80 ? " (Assumed from chip registers)":"");
-		for (i = 0; i < eeprom_size; i++) {
-			if ((i&15) == 0)
-				printf("\n0x%3.3x: ", i + 0x100);
-			printf(" %2.2x", eeprom_contents[i]);
+			   inb(ioaddr + EECtrl) & 0x80 ?
+			   " (Assumed from chip registers)" : "");
+		for (i = 0; i < eeprom_size; i += 8) {
+			int j;
+			printf("\n0x%2.2x: ", i);
+			for (j = 0; j < 8; j++)
+				printf(" %4.4x", eeprom_contents[i + j]);
+			if (show_eeprom > 2) {
+				printf("  ");
+				for (j = 0; j < 8; j++) {
+					int ew = eeprom_contents[i + j];
+					printf("%c%c",
+						   isprint(ew & 0xff) ? ew & 0xff : '_',
+						   isprint(ew >>   8) ? ew >> 8   : '_' );
+				}
+			}
 		}
 		printf("\n");
 	}
+	if (show_eeprom)
+		rhine_eeprom(eeprom_contents);
+
 	if (show_mii) {
 		int phys[4], phy, phy_idx = 0;
 		for (phy = 0; phy < 32 && phy_idx < 4; phy++) {
@@ -727,11 +795,9 @@ int rhine_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 /* Serial EEPROM section.
    A "bit" grungy, but we work our way through bit-by-bit :->. */
 
-/* The EEPROM commands include the alway-set leading bit.
-   They must be shifted by the number of address bits. */
-#define EE_WRITE_CMD	(5)
-#define EE_READ_CMD		(6)
-#define EE_ERASE_CMD	(7)
+/* The EEPROM commands always start with 01.. preamble bits.
+   Commands are prepended to the variable-length address. */
+enum EEPROM_Cmds { EE_WriteCmd=5, EE_ReadCmd=6, EE_EraseCmd=7, };
 
 /*  EEPROM_Ctrl bits.
     Some implementations have a data pin direction bit instead of
@@ -747,12 +813,18 @@ int rhine_diag(int vendor_id, int device_id, long ioaddr, int part_idx)
 #define EE_WRITE_1		(EE_CS | EE_DATA_WRITE)
 
 #define EE_OFFSET		EECtrl	/* Register offset in I/O space. */
-/* Delay between EEPROM clock transitions. */
+enum EEPROM_Ctrl_Bits {
+	EE_ShiftClk=0x04, EE_ChipSelect=0x08, EE_DataOut=0x02, EE_DataIn=0x01,
+	EE_Write0=0x08, EE_Write1=0x0A,
+};
+/* Delay between EEPROM clock transitions.
+   This forces out buffered PCI writes.
+*/
 #define eeprom_delay(ee_addr)	inb(ee_addr)
 
-/* This executes a generic EEPROM command, typically a write or write enable.
-   It returns the data output from the EEPROM, and thus may also be used for
-   reads. */
+/* Execute a generic EEPROM command.
+   Return all data output from the EEPROM, and thus may be used for
+   EEPROM sizing, read, erase or write. */
 static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 {
 	unsigned retval = 0;
@@ -761,20 +833,20 @@ static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 	if (debug > 1)
 		printf(" EEPROM op 0x%x: ", cmd);
 
-	outb(EE_ENB | EE_SHIFT_CLK, ee_addr);
+	outb(EE_ChipSelect | EE_ShiftClk, ee_addr);
 
 	/* Shift the command bits out. */
 	do {
-		short dataval = (cmd & (1 << cmd_len)) ? EE_WRITE_1 : EE_WRITE_0;
+		short dataval = (cmd & (1 << cmd_len)) ? EE_Write1 : EE_Write0;
 		outb(dataval, ee_addr);
 		eeprom_delay(ee_addr);
 		if (debug > 2)
 			printf("%X", inb(ee_addr) & 15);
-		outb(dataval | EE_SHIFT_CLK, ee_addr);
+		outb(dataval | EE_ShiftClk, ee_addr);
 		eeprom_delay(ee_addr);
-		retval = (retval << 1) | ((inb(ee_addr) & EE_DATA_READ) ? 1 : 0);
+		retval = (retval << 1) | ((inb(ee_addr) & EE_DataIn) ? 1 : 0);
 	} while (--cmd_len >= 0);
-	outb(EE_ENB, ee_addr);
+	outb(EE_ChipSelect, ee_addr);
 
 	/* Terminate the EEPROM access. */
 	outb(EE_ENB & ~EE_CS, ee_addr);
@@ -783,41 +855,82 @@ static int do_eeprom_cmd(long ioaddr, int cmd, int cmd_len)
 	return retval;
 }
 
+/* Wait for the EEPROM to finish what it is doing. */
+static int eeprom_busy_poll(long ee_ioaddr)
+{
+	int i;
+	outb(EE_ChipSelect, ee_ioaddr);
+	for (i = 0; i < 10000; i++)			/* Typical 2000 ticks */
+		if (inb(ee_ioaddr) & EE_DataIn)
+			break;
+	return i;
+}
+
 /* The abstracted functions for EEPROM access. */
+/* Return the number of address bits this EEPROM accepts. */
+static int eeprom_addr_len(long ioaddr)
+{
+	return do_eeprom_cmd(ioaddr, EE_ReadCmd << (6+16), 3 + 6 + 16)
+		& 0x10000 ? 8 : 6;
+}
+
 static int read_eeprom(long ioaddr, int location, int addr_len)
 {
-	return do_eeprom_cmd(ioaddr, ((EE_READ_CMD << addr_len) | location) << 16,
+	return do_eeprom_cmd(ioaddr, ((EE_ReadCmd << addr_len) | location) << 16,
 						 3 + addr_len + 16) & 0xffff;
 }
 
-
 static void write_eeprom(long ioaddr, int index, int value, int addr_len)
 {
 	long ee_ioaddr = ioaddr + EE_OFFSET;
 	int i;
 
 	/* Poll for previous op finished. */
-	outb(EE_ENB, ee_ioaddr);
-	for (i = 0; i < 10000; i++)			/* Typical 2000 ticks */
-		if (inb(ee_ioaddr) & EE_DATA_READ)
-			break;
+	eeprom_busy_poll(ee_ioaddr);
+
 	/* Enable programming modes. */
 	do_eeprom_cmd(ioaddr, (0x4f << (addr_len-4)), 3 + addr_len);
 	/* Do the actual write. */ 
 	do_eeprom_cmd(ioaddr,
-				  (((EE_WRITE_CMD<<addr_len) | index)<<16) | (value & 0xffff),
+				  (((EE_WriteCmd<<addr_len) | index) << 16) | (value & 0xffff),
 				  3 + addr_len + 16);
-	/* Poll for write finished. */
-	outb(EE_ENB, ee_ioaddr);
-	for (i = 0; i < 10000; i++)			/* Typical 2000 ticks */
-		if (inb(ee_ioaddr) & EE_DATA_READ)
-			break;
+	i = eeprom_busy_poll(ee_ioaddr);
 	if (debug)
 		printf(" Write finished after %d ticks.\n", i);
-	/* Disable programming (note: this could take extra time. */
+	/* Disable programming.  Note: this command is not instantaneous, so
+	   we check for busy before the next op. */
 	do_eeprom_cmd(ioaddr, (0x40 << (addr_len-4)), 3 + addr_len);
+	eeprom_busy_poll(ee_ioaddr);
 }
 
+#define decode_ee_fmt \
+"Decoded EEPROM contents:\n"\
+"   Ethernet MAC Station Address %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n"\
+"   PHY at MII #%d.\n"\
+"   PCI Subsystem IDs -- Vendor %#4.4x, Device %#4.4x.\n"\
+"   PCI IDs -- Vendor %#4.4x, Device %#4.4x.\n"\
+"   PCI timer settings -- minimum grant %d, maximum latency %d.\n"\
+"  Checksum computed as %2.2x, read as %2.2x.\n"
+
+static void rhine_eeprom(unsigned short *ee_data)
+{
+	int i, checksum = 0;
+	for (i = 0; i < 31; i++)
+		checksum = ((char *)ee_data)[i];
+	printf(decode_ee_fmt,
+		   ee_data[0] & 0xff, ee_data[0]>>8,
+		   ee_data[1] & 0xff, ee_data[1]>>8,
+		   ee_data[2] & 0xff, ee_data[2]>>8,
+		   ee_data[3] & 0xff,						/* PHY */
+		   ee_data[5], ee_data[4], 					/* Subsystem */
+		   ee_data[7], ee_data[6], 					/* Subsystem */
+		   ee_data[11] & 0xff, ee_data[11]>>8,		/* PCI bus config */
+		   checksum & 0xff, ee_data[15] >> 8);
+	return;
+}
+
+
+
 /* MII - MDIO  (Media Independent Interface - Management Data I/O) accesses. */
 
 int mdio_read(long ioaddr, int phy_id, int regnum)
