@@ -4,7 +4,7 @@
    based on the 3Com Vortex, Boomerang and Cyclone chips, as used on the
    3Com 3c590/595/900/905 PCI EtherLink XL adapters.
 
-   Copyright 1997-2001 by Donald Becker.
+	Copyright 1997-2003 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -17,24 +17,66 @@
 	 410 Severn Ave., Suite 210
 	 Annapolis MD 21403
 
-	Support and updates available at
-	 http://www.scyld.com/diag/index.html
+	Updates and additional information are available at
+	http://www.scyld.com/diag/index.html
+	http://scyld.com/expert/mii-status.html
+	http://scyld.com/expert/NWay.html
+
+	Common-sense licensing statement: Using any portion of this program in
+	your own program means that you must give credit to the original author
+	and release the resulting code under the GPL.  To use this code under
+	other terms requires an explicit license from the copyright holder.
 
 	References
 	 3Com Vortex Engineering Release Specification
 	 3Com Boomerang modifications (unreleased)
-	 http://scyld.com/expert/mii-status.html
-	 http://scyld.com/expert/NWay.html
-     http://www.national.com/pf/DP/DP83840.html
+	 http://www.national.com/pf/DP/DP83840.html
 */
 
 static char *version_msg =
-"vortex-diag.c:v2.05 5/15/2001 Donald Becker (becker@scyld.com)\n"
+"vortex-diag.c:v2.14 12/28/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: vortex-diag [-aDeEfFgGmpPqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"Usage: vortex-diag [-aDEefFgGhmPqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
 "  For details and other options see http://www.scyld.com/diag/index.html\n";
 
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      vortex-diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+        Set a new hardware station address.  Read the documentation first!
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+ To read and write the boot BIOS extension Flash ROM
+   -B       Show the first few bytes of the ROM
+   -L FILE  Load the Flash from FILE.
+   -S FILE  Store the Flash image to FILE.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
+
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
 #warning  See the last lines of the source file.
@@ -47,13 +89,15 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <ctype.h>
 #if defined(__linux__)  &&  __GNU_LIBRARY__ == 1
 #include <asm/io.h>			/* Newer libraries use <sys/io.h> instead. */
 #else
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -64,17 +108,18 @@ extern int (*flash_in_hook)(long addr, int offset);
 extern void (*flash_out_hook)(long addr, int offset, int val);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -82,7 +127,8 @@ struct option longopts[] = {
 	{"show-mii",  0, 0, 'm'},	/* Dump MII management registers. */
 	{"port-base",  1, 0, 'p'},	/* Use the specified I/O address. */
 	{"quiet",	0, 0, 'q'},		/* Decrease verbosity */
-	{"Reset",	0, 0, 'R'},		/* Reset the transceiver. */
+	{"reset",	0, 0, 'R'},		/* Reset the transceiver. */
+	{"restart",	0, 0, 'r'},		/* Restart autonegotiation. */
 	{"chip-type",  1, 0, 't'},	/* Assume the specified chip type index. */
 	{"test",	0, 0, 'T'},		/* Do register and SRAM test. */
 	{"verbose",	0, 0, 'v'},		/* Verbose mode */
@@ -94,9 +140,12 @@ struct option longopts[] = {
 extern int  vortex_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 
 /* Chip-specific flags. Yes, it's grungy to have the enum here. */
-enum { HAS_FLASH_BUG=1, ROADRUNNER=2 };
+enum { HAS_FLASH_BUG=1, ROADRUNNER=2, IS_CYCLONE=4 };
 
-/* This table is searched in order: place specific entries followed by
+/* The table of known chips.
+   Because of the bogus /proc/pci interface we must have both the exact
+   name from the kernel, a common name and the PCI vendor/device IDs.
+   This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
 	const char *part_name;
@@ -109,49 +158,49 @@ struct pcidev_entry {
 	{"3Com Generic Vortex/Boomerag/Cyclone", 0,
 	 0x10B7, 0x9999, 0xffff, 0, 256, vortex_diag},
 	{"3Com 3c590 Vortex 10Mbps", "3Com Vortex (rev ",
-	 0x10B7, 0x5900, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x5900, 0xffff, 0, 32, vortex_diag},
 	{"3c592 EISA 10mbps Demon/Vortex", 0,
-	 0x10B7, 0x5920, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x5920, 0xffff, 0, 32, vortex_diag},
 	{"3Com 3c595 Vortex 10/100baseTx", "3Com Vortex (rev ",
-	 0x10B7, 0x5950, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x5950, 0xffff, 0, 32, vortex_diag},
 	{"3Com 3c595 Vortex 10/100baseT4", 0,
-	 0x10B7, 0x5951, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x5951, 0xffff, 0, 32, vortex_diag},
 	{"3Com 3c595 Vortex 10/100baseT-MII", 0,
-	 0x10B7, 0x5952, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x5952, 0xffff, 0, 32, vortex_diag},
 	{"3c597 EISA Fast Demon/Vortex", 0,
-	 0x10B7, 0x5970, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x5970, 0xffff, 0, 32, vortex_diag},
 	{"3c900 Boomerang 10baseT", "3Com 3C900",
-	 0x10B7, 0x9000, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9000, 0xffff, 0, 64, vortex_diag},
 	{"3c900 Boomerang 10Mbps Combo", 0,
-	 0x10B7, 0x9001, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9001, 0xffff, 0, 64, vortex_diag},
 	{"3c900 Cyclone 10Mbps TPO", 0,
-	 0x10B7, 0x9004, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9004, 0xffff, IS_CYCLONE, 256, vortex_diag},
 	{"3c900 Cyclone 10Mbps Combo", 0,
-	 0x10B7, 0x9005, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9005, 0xffff, IS_CYCLONE, 256, vortex_diag},
 	{"3c900 Cyclone 10Mbps TPC", 0,
-	 0x10B7, 0x9006, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9006, 0xffff, IS_CYCLONE, 256, vortex_diag},
 	{"3c905 Boomerang 100baseTx", "3Com 3C905",
-	 0x10B7, 0x9050, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9050, 0xffff, 0, 64, vortex_diag},
 	{"3c905 Boomerang 100baseT4", 0,
-	 0x10B7, 0x9051, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x9051, 0xffff, 0, 64, vortex_diag},
 	{"3c905B Cyclone 100baseTx", 0,
-	 0x10B7, 0x9055, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x9055, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c905B Cyclone 10/100/BNC", "3Com 3C905BNC",
-	 0x10B7, 0x9058, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x9058, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c905B Cyclone 100baseFx", "3Com 3C905B-FX",
-	 0x10B7, 0x905A, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x905A, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c905C Tornado 100baseTx", 0,
-	 0x10B7, 0x9200, 0xff00, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x9200, 0xff00, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c980 Cyclone, server edition", 0,
-	 0x10B7, 0x9800, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x9800, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c982 Dual Port server NIC", 0,
-	 0x10B7, 0x9805, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x9805, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c980 series, server edition", 0,
-	 0x10B7, 0x9800, 0xfff0, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x9800, 0xfff0, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3Com 3cSOHO100-TX", 0,
-	 0x10B7, 0x7646, 0xff00, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x7646, 0xff00, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3c450 HomePNA Cyclone", 0,
-	 0x10B7, 0x4500, 0xffff, 0, 256, vortex_diag},
+	 0x10B7, 0x4500, 0xffff, IS_CYCLONE, 256, vortex_diag},
 	{"3c555 Laptop Hurricane",	0,
 	 0x10B7, 0x5055, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
 	{"3c556 Laptop Hurricane",	0,
@@ -161,21 +210,22 @@ struct pcidev_entry {
 	{"3Com 3CCFE556", "3ccfe556 Roadrunner PCMCIA",
 	 0x10B7, 0x0556, 0xffff, ROADRUNNER, 32, vortex_diag},
 	{"3c575 CardBus", 0,
-	 0x10B7, 0x5057, 0xffff, HAS_FLASH_BUG, 32, vortex_diag},
+	 0x10B7, 0x5057, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 32, vortex_diag},
 	{"3CCFE575BT CardBus", 0,
-	 0x10B7, 0x5157, 0xffff, HAS_FLASH_BUG, 128, vortex_diag},
+	 0x10B7, 0x5157, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 128, vortex_diag},
 	{"3CCFE575CT CardBus", 0,
-	 0x10B7, 0x5257, 0xffff, HAS_FLASH_BUG, 128, vortex_diag},
+	 0x10B7, 0x5257, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 128, vortex_diag},
 	{"3C575 CardBus (unknown version)", 0,
 	 0x10B7, 0x5057, 0x0fff, HAS_FLASH_BUG, 32, vortex_diag},
 	{"3ccfe656 Cyclone CardBus", 0,
-	 0x10B7, 0x6560, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x6560, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3Com 3CCFE656B Cyclone+WinModem CardBus", 0,
-	 0x10B7, 0x6562, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x6562, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3Com 3CCFE656C Tornado CardBus", 0,
-	 0x10B7, 0x6564, 0xffff, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x6564, 0xffff, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
 	{"3Com 3CCFE656 Series", 0,
-	 0x10B7, 0x6560, 0xfff0, HAS_FLASH_BUG, 256, vortex_diag},
+	 0x10B7, 0x6560, 0xfff0, HAS_FLASH_BUG | IS_CYCLONE, 256, vortex_diag},
+	/* Added most new/unverified entries here. */
 	{ 0, 0, 0, 0},
 };
 
@@ -192,6 +242,7 @@ int nway_advertise = 0, fixed_speed = -1;
 int new_default_media = -1;
 static unsigned int opt_flash_show = 0;
 static char	*opt_flash_dumpfile = NULL, *opt_flash_loadfile = NULL;
+
 static unsigned char new_hwaddr[6], set_hwaddr = 0;
 static int emergency_rewrite = 0;
 static unsigned set_ee_rom = 0;
@@ -211,7 +262,7 @@ main(int argc, char **argv)
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:gG:mp:PqrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:gG:hH:mp:PqrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -227,7 +278,8 @@ main(int argc, char **argv)
 				errflag++;
 			break;
 		case 'g': opt_dma_diag++;	break;
-		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 0); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -459,6 +511,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -466,7 +519,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtoul(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -476,37 +530,35 @@ static int parse_media_type(const char *capabilities)
    0x0800	Power up autosense (check speed only once)
    0x8000	Dynamic Autosense
 */
-/* A table of media names to indices.  This matches the Digital Tulip
-   SROM numbering, primarily because that is the most complete list.
-   Other chips will have to map these number to their internal values.
+/* A table of media names to EEPROM setting.
+   These match the 3Com internal values.
 */
-struct { char *name; int value; } mediamap[] = {
-	{ "10baseT", 0 },
-	{ "10base2", 1 },
-	{ "AUI", 2 },
-	{ "100baseTx", 3 },
-	{ "10baseT-FDX", 0x204 },
-	{ "100baseTx-FDX", 0x205 },
-	{ "100baseT4", 6 },
-	{ "100baseFx", 7 },
-	{ "100baseFx-FDX", 8 },
-	{ "MII", 11 },
-	{ "Autosense", 0x0800 },
+struct { int value; char *name; } mediamap[] = {
+	{ 0, "10baseT-serial"},
+	{ 1, "10Mbps-AUI"},
+	{ 3, "10base2"},
+	{ 5, "100baseFx"},
+	{ 6, "MII/100baseT4"},
+	{ 8, "Autonegotiation"},
 	{ 0, 0 },
 };
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
@@ -514,15 +566,11 @@ static int get_media_index(const char *name)
 /* Chip-specific section. */
 
 /* The 3Com Vortex/Boomerang/Cyclone designs. */
-static int read_eeprom(long ioaddr, int addrsize, int location);
-static void write_eeprom(long ioaddr, int addrsize, int index, int value);
-static int do_update(long ioaddr, int addrlen, 
-					 unsigned short *ee_values, unsigned short *old_ee_values);
+static void interpret_eeprom(long ioaddr);
 static void mdio_sync(long ioaddr);
 int mdio_read(long ioaddr, int phy_id, int location);
 void mdio_write(long ioaddr, int phy_id, int location, int value);
 
-static void parse_eeprom(unsigned short *ee_data);
 static void show_dma_state(long ioaddr);
 int vortex_flash_in(long addr, int offset);
 void vortex_flash_out(long addr, int offset, int val);
@@ -540,7 +588,8 @@ enum vortex_cmd {
 	SetStatusEnb = 15<<11, SetRxFilter = 16<<11, SetRxThreshold = 17<<11,
 	SetTxThreshold = 18<<11, SetTxStart = 19<<11,
 	StartDMAUp = 20<<11, StartDMADown = (20<<11)+1, StatsEnable = 21<<11,
-	StatsDisable = 22<<11, StopCoax = 23<<11,};
+	StatsDisable = 22<<11, StopCoax = 23<<11, SetFilterBit = 25<<11,
+};
 
 /* Bits in the general status register. */
 enum vortex_status {
@@ -561,7 +610,7 @@ const char *intr_names[13] ={
 enum Window0 {
 	Wn0EepromCmd = 10,		/* Window 0: EEPROM command register. */
 	Wn0EepromData = 12,		/* Window 0: EEPROM results register. */
-	IntrStatus=0x0E,			/* Valid in all windows. */
+	IntrStatus=0x0E,		/* Valid in all windows. */
 };
 #define MAX_EEPROM_SIZE 256
 /* In undocumented bogusness, these value depend on the EEPROM size. */
@@ -580,6 +629,10 @@ enum Window3 {			/* Window 3: MAC/config bits. */
 enum Window4 {		/* Window 4: Xcvr/media bits. */
 	Wn4_FIFODiag = 4, Wn4_NetDiag = 6, Wn4_PhysicalMgmt=8, Wn4_Media = 10,
 };
+enum Window5 {
+	Wn5_TxThreshold = 0, Wn5_RxFilter = 8,
+};
+
 /* Media names in Wn3_Config bits 24:20. */
 const char *el_medianames[16] ={
 	"10baseT", "10Mbs AUI", "undefined", "10base2",
@@ -587,6 +640,22 @@ const char *el_medianames[16] ={
 	"Autonegotiate", "MII-External", "undefined-10", "undefined-11",
 	"undefined-12", "undefined-13", "undefined-14", "undefined-15",
 };
+/* Media available in Wn3_Options and EEPROM. */
+const char *medias[] = {
+	"100baseT4", "100baseTx", "100baseFx", "10baseT",
+	"10base2", "AUI", "MII", "",   "10baseFL"
+};
+
+struct config_name { int val, mask; const char*name;}
+/* Bits in Wn5_RxFilter. */
+static rcvr_mode[] = {
+	{0x05, 0x1f, "Normal unicast"},
+	{0x15, 0x1f, "Normal unicast and hashed multicast"},
+	{0x07, 0x0f, "Normal unicast and all multicast"},
+	{0x08, 0x08, "Promiscuous"},
+	{0x00, 0x00, "Unknown/invalid"},
+};
+
 
 /* Values read from the EEPROM, and the new image. */
 unsigned short eeprom_contents[MAX_EEPROM_SIZE];
@@ -596,8 +665,8 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 {
 	int chip_active = 0;
 	int saved_window = inw(ioaddr + EL3_CMD) >> 13;
-	int internal_config;
-	int eeaddrlen, eesize, ee_tbl_offset = 0;
+	int flags = pcidev_tbl[part_idx].flags;
+	int internal_config, rx_mode;
 	int i;
 
 	/* It's mostly safe to examine the registers and EEPROM during
@@ -607,6 +676,16 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 		chip_active = 1;
 #endif
 
+	EL3WINDOW(2);
+	printf(" Station address %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x.\n",
+		   inb(ioaddr + 0), inb(ioaddr + 1), inb(ioaddr + 2),
+		   inb(ioaddr + 3), inb(ioaddr + 4), inb(ioaddr + 5));
+	EL3WINDOW(5);
+	rx_mode = inb(ioaddr + Wn5_RxFilter);
+	for (i = 0; rcvr_mode[i].mask; i++)
+		if ((rx_mode & rcvr_mode[i].mask) == rcvr_mode[i].val) break;
+	printf("  Receive mode is 0x%2.2x: %s.\n", rx_mode, rcvr_mode[i].name);
+
 	if (opt_a) {
 		int io_size = pcidev_tbl[part_idx].io_size;
 		int j;
@@ -632,9 +711,8 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 			printf(".\n");
 		}
 		printf("Vortex chip registers at %#lx", ioaddr);
-		for (i = 0x10;
-			 i < (io_size > 0x40 ? 0x40 : io_size);
-			 i += 4) {
+		if (io_size > 0x80) io_size = 0x80;
+		for (i = 0x10; i < io_size; i += 4) {
 			if ((i & 0x0f) == 0)
 				printf("\n  0x%3.3lX:", ioaddr + i);
 			if (! opt_f) {
@@ -643,14 +721,32 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 				else if (i == 0x1c)
 					printf(" *STATUS*");
 				else
-					printf(" %8.8x", inl(ioaddr + i));
+					printf(" %8.8x", (int)inl(ioaddr + i));
 			} else
-				printf(" %8.8x", inl(ioaddr + i));
+				printf(" %8.8x", (int)inl(ioaddr + i));
 		}
 		printf("\n");
+		if (opt_dma_diag || (flags & IS_CYCLONE))
+			show_dma_state(ioaddr);
+	}
+	if (opt_G) {
+#if defined(OPTG_MULTICAST)
+		printf("Setting filter bit, %4.4x.\n", SetFilterBit|0x0400|opt_GPIO);
+		outw(SetRxFilter | 0x15, ioaddr + EL3_CMD);
+		for (i = 0; i < 255; i++)
+			outw(SetFilterBit|0x0000|i, ioaddr + EL3_CMD);
+		outw(SetFilterBit|0x0400|opt_GPIO, ioaddr + EL3_CMD);
+#elif defined(OPTG_RXTHRESH)
+		printf("Setting the Rx Priority Thrshold to %d bytes.\n",
+			   opt_GPIO);
+		outb((opt_GPIO + 0x1f)>>5, ioaddr + 0x3C);
+		printf("Setting the Rx Priority Thrshold to %d bytes.\n",
+			   inb(ioaddr + 0x3C)<<5);
+#else
+		EL3WINDOW(2);
+		outw(opt_GPIO, ioaddr + Wn2_ResetOptions);
+#endif
 	}
-	if (opt_dma_diag)
-		show_dma_state(ioaddr);
 	EL3WINDOW(3);
 	internal_config = inl(ioaddr + Wn3_Config);
 	if (verbose > 1 || opt_a) {
@@ -666,8 +762,6 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 					printf("   %s indication.\n", intr_names[i]);
 		}
 		{
-			const char *medias[] = {"100baseT4", "100baseTx", "100baseFx",
-			"10baseT", "10base2", "AUI", "MII", "", "10baseFL"};
 			int MediaOptions, MacCtrl;
 			EL3WINDOW(3);
 			MediaOptions = inw(ioaddr + Wn3_Options);
@@ -675,7 +769,7 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 			for (i = 0; i < 8; i++)
 				if (MediaOptions & 1<<i)
 					printf(" %s", medias[i]);
-			printf("%s.\n", (MediaOptions&0xE010)==0x0010 ? "10baseFL" : "");
+			printf("%s.\n", (MediaOptions&0xE100)==0x0100 ? " 10baseFL" : "");
 			printf("Transceiver type in use:  %s.\n",
 				   el_medianames[(internal_config >> 20) & 15]);
 			MacCtrl = inw(ioaddr + Wn3_MAC_Ctrl);
@@ -696,84 +790,8 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 		printf(" Configuration options %4.4x.\n",
 			   inw(ioaddr + Wn2_ResetOptions));
 	}
-	/* Read the EEPROM. */
-	EL3WINDOW(0);
 
-	outw(0x5555, ioaddr + Wn0EepromData);
-	if (read_eeprom(ioaddr, 6, 0) != 0x5555) {
-		if (debug)
-			printf(" EEPROM Address length is 6 bits (%4.4x).\n",
-				   read_eeprom(ioaddr, 6, 0));
-		eeaddrlen = 6;
-	} else {
-		int val = read_eeprom(ioaddr, 6, 0);
-		int e6 = read_eeprom(ioaddr, 6, 0);
-		int e8 = read_eeprom(ioaddr, 8, 0);
-		if (debug)
-			printf(" EEPROM address sizing read returned %4.4x/%4.4x/%4.4x.\n",
-				   e6, e8, val);
-		eeaddrlen = (e6 != 0xffff) ? 8 : 6;
-	}
-	eesize = 1 << eeaddrlen;
-
-	if (debug)
-		printf(" EEPROM address size is %d bits.\n", eeaddrlen);
-	for (i = 0; i < eesize; i++)
-		eeprom_contents[i] = read_eeprom(ioaddr, eeaddrlen, i);
-
-	if (eeprom_contents[0x37] == 0x6d50) {
-		ee_tbl_offset = 0x30;
-	} else if (eeprom_contents[0] == 0x10b7  ||  eeprom_contents[0] == 0x1570
-		|| eeprom_contents[0] == 0x2978) {
-		printf("  Questionable EEPROM offset for software information!\n");
-		ee_tbl_offset = 0x30;
-	}
-	if (set_hwaddr) {
-		unsigned short sum = 0;
-		memcpy(new_ee_contents, eeprom_contents, eesize << 1);
-		for (i = 0; i < 3; i++)
-			new_ee_contents[ee_tbl_offset + 10 + i] =
-				(new_hwaddr[i*2]<<8) + new_hwaddr[i*2+1];
-		/* Recalculate the checksum: Cyclone only! */
-		for (i = 0; i < 0x1A; i++)
-			sum ^= new_ee_contents[ee_tbl_offset + i];
-		new_ee_contents[ee_tbl_offset + 0x20] = (sum ^ (sum>>8)) & 0xff;
-		do_update(ioaddr, eeaddrlen, new_ee_contents, eeprom_contents);
-		for (i = 0; i < eesize; i++)
-			eeprom_contents[i] = read_eeprom(ioaddr, eeaddrlen, i);
-	}
-
-	if (set_ee_rom) {
-		unsigned short sum = 0;
-		memcpy(new_ee_contents, eeprom_contents, eesize << 1);
-		new_ee_contents[9] = 0x3001;
-		/* Recalculate the checksum: Cyclone only! */
-		for (i = 0; i < 0x1B; i++)
-			sum ^= new_ee_contents[i];
-		new_ee_contents[0x20] = (sum ^ (sum>>8)) & 0xff;
-		printf("Setting the EEPROM BIOS ROM field to %4.4x, new checksum "
-			   "%2.2x.\n", new_ee_contents[9], new_ee_contents[0x20]);
-		do_update(ioaddr, eeaddrlen, new_ee_contents, eeprom_contents);
-		for (i = 0; i < eesize; i++)
-			eeprom_contents[i] = read_eeprom(ioaddr, eeaddrlen, i);
-	}
-	if (verbose + show_eeprom > 2) {
-		unsigned short sum = 0;
-		printf("EEPROM contents (%d words, offset %#x):",
-			   eesize, ee_tbl_offset);
-		for (i = 0; i < eesize; i++) {
-			if (i % 8 == 0)
-				printf("\n 0x%3.3x:", i);
-			printf(" %4.4x", eeprom_contents[i]);
-			sum += eeprom_contents[i];
-		}
-		printf("\n The word-wide EEPROM checksum is %#4.4x.\n", sum);
-	}
-
-	/* The user will usually want to see the interpreted EEPROM contents. */
-	if (verbose > 1 || show_eeprom) {
-		parse_eeprom(eeprom_contents);
-	}
+	interpret_eeprom(ioaddr);
 
 	/* Show up to four (not just the on-board) PHYs. */
 	if (verbose > 1 || show_mii) {
@@ -844,7 +862,7 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 	   to read the flash on one Hurricane rev. */
 	/* Turn on the MII transceiver for some cards. */
 	{
-		if (pcidev_tbl[part_idx].flags & HAS_FLASH_BUG) {
+		if (flags & HAS_FLASH_BUG) {
 			EL3WINDOW(3);
 			outl((internal_config & ~0x00f00000)|0x00600000,
 				 ioaddr + Wn3_Config);
@@ -872,7 +890,7 @@ int vortex_diag(int vendor_id, int dev_id, long ioaddr, int part_idx)
 				EL3WINDOW(saved_window);
 				return 4;
 			}
-		if (pcidev_tbl[part_idx].flags & HAS_FLASH_BUG) {
+		if (flags & HAS_FLASH_BUG) {
 			EL3WINDOW(3);
 			outl(internal_config, ioaddr + Wn3_Config);
 		}
@@ -912,8 +930,11 @@ static int read_eeprom(long ioaddr, int addrlen, int location)
 {
 	int timer;
 
+	if (addrlen == 6 && location > 0x3f)
+		location = (location & 0x3f) + ((location<<2) & 0x0f00);
+
 	outw((2 << addrlen) + location, ioaddr + Wn0EepromCmd);
- 	/* Wait for the read to take place, worst-case 162 us. */
+	/* Wait for the read to take place, worst-case 162 us. */
 	for (timer = 1620; timer >= 0; timer--) {
 		if ((inw(ioaddr + Wn0EepromCmd) & 0x8000) == 0)
 			break;
@@ -969,7 +990,7 @@ error_return:
 			index, value);
 }
 
-static int do_update(long ioaddr, int addrlen, 
+static int do_update(long ioaddr, int addrlen,
 					 unsigned short *ee_values, unsigned short *old_ee_values)
 {
 	int i;
@@ -986,7 +1007,7 @@ static int do_update(long ioaddr, int addrlen,
 						   i, ee_values[i]);
 				write_eeprom(ioaddr, addrlen, i, ee_values[i]);
 			} else
-				printf(" Would write new %d entry 0x%4.4x (old value 0x%4.4x).\n", 
+				printf(" Would write new %d entry 0x%4.4x (old value 0x%4.4x).\n",
 					   i, ee_values[i], old_ee_values[i]);
 		}
 	}
@@ -995,6 +1016,202 @@ static int do_update(long ioaddr, int addrlen,
 		;
 	return 0;
 }
+
+static void parse_eeprom(unsigned short *eeprom)
+{
+	unsigned char *p = (void *)eeprom;		/* Assumes little-endian */
+	u16 *ee = eeprom;
+	int i, sum = 0;
+
+	printf("Saved EEPROM settings of a 3Com Vortex/Boomerang:\n");
+	if (eeprom[0x37] == 0x6d50) {
+		printf(" The CardBus product ID is %4.4x %4.4x.\n",
+			   eeprom[0], eeprom[1]);
+		p += 0x60;
+		ee += 0x30;
+	}
+	printf(" 3Com Node Address ");
+	for (i = 0; i < 5; i++)
+		printf("%2.2X:", p[i^1]);
+	printf("%2.2X (used as a unique ID only).\n", p[i^1]);
+	printf(" OEM Station address %2.2X", p[1 + 20]);
+	for (i = 1; i < 6; i++)
+		printf(":%2.2X", p[(i^1) + 20]);
+	printf(" (used as the ethernet address).\n");
+	/* The original 3Com layout did not consider post-Y2K production. */
+	printf("  Device ID %4.4x,  Manufacturer ID %4.4x.\n", ee[3], ee[7]);
+	printf("  Manufacture date (MM/DD/YYYY) %d/%d/%d, division %c,"
+		   " product %c%c.\n", ((p[8]>>5) & 7) + ((p[9]<<3) & 8),
+		   p[8] & 31, (p[9]>>1) + ((p[9]>>1) < 95 ? 2000 : 1900),
+		   p[10], p[12], p[13]);
+	if (ee[9] & 0x0800)
+		printf("  A BIOS ROM of size %dKx8 is expected.\n",
+			   (ee[9]>>12)*64);
+	else
+		printf("  No BIOS ROM is present.\n");
+	printf(" Transceiver selection: %s.\n",
+		   el_medianames[(ee[19] >> 4) & 15]);
+	printf("   Options: %s duplex, link beat %s.\n",
+		   ee[13] & 0x8000 ? "force full" : "negotiated",
+		   ee[13] & 0x4000 ? "check disabled" : "required");
+	if (ee[0x18]) {
+		int MediaOptions = ee[0x19];
+		printf(" PCI Subsystem IDs: Vendor %4.4x Device %4.4x.\n",
+			   ee[0x17], ee[0x18]);
+		for (i = 0; i < 8; i++)
+			if (MediaOptions & 1<<i)
+				printf(" %s", medias[i]);
+		printf("%s.\n", (MediaOptions&0xE100)==0x0100 ? "10baseFL" : "");
+	}
+	for (i = 0; i < 0x16; i++)
+		sum ^= ee[i];
+	printf("  Vortex format checksum is %scorrect (%2.2x vs. %2.2x).\n",
+		   ((sum ^ (sum>>8)) & 0xff) == ee[0x17] ? "" : "in",
+		   (sum ^ (sum>>8)) & 0xff, ee[0x17]);
+	for ( ; i < 0x1A; i++)
+		sum ^= ee[i];
+	sum ^= sum>>8;
+	printf("  Cyclone format checksum is %scorrect (%#2.2x vs. %#2.2x).\n",
+		   (sum  & 0xff) == (ee[0x20] & 0xff) ? "" : "in",
+		   sum & 0xff, ee[0x20] & 0xff);
+	for (sum = 0, i = 0; i < 0x20*2; i++)
+		sum ^= ((unsigned char *)ee)[i];
+	printf("  Hurricane format checksum is %scorrect (%#2.2x vs. %#2.2x).\n",
+		   sum == (ee[0x20] & 0xff) ? "" : "in",
+		   sum, ee[0x20] & 0xff);
+	return;
+}
+
+static void interpret_eeprom(long ioaddr)
+{
+	int eeaddrlen, eesize, ee_tbl_offset = 0;
+	int e6, eeprom_update = 0;
+	int i;
+
+	/* Read the EEPROM. */
+	EL3WINDOW(0);
+
+	outw(0x5555, ioaddr + Wn0EepromData);
+	e6 = read_eeprom(ioaddr, 6, 0);
+	if (e6 != 0x5555) {
+		if (debug)
+			printf(" EEPROM Address length is 6 bits (%4.4x).\n",
+				   read_eeprom(ioaddr, 6, 0));
+		eeaddrlen = 6;
+	} else {
+		int e8 = read_eeprom(ioaddr, 8, 0);
+		if (debug)
+			printf(" EEPROM address sizing read returned %4.4x/%4.4x.\n",
+				   e6, e8);
+		eeaddrlen = (e8 != 0xffff) ? 8 : 6;
+	}
+	eesize = 1 << eeaddrlen;
+
+	if (debug)
+		printf(" EEPROM address size is %d bits.\n", eeaddrlen);
+	for (i = 0; i < eesize; i++)
+		eeprom_contents[i] = read_eeprom(ioaddr, eeaddrlen, i);
+
+	if (eeprom_contents[0x37] == 0x6d50) {
+		ee_tbl_offset = 0x30;
+	} else if (eeprom_contents[0] == 0x10b7  ||  eeprom_contents[0] == 0x1570
+		|| eeprom_contents[0] == 0x2978) {
+		printf("  Questionable EEPROM offset for software information!\n");
+		ee_tbl_offset = 0x30;
+	}
+
+	memcpy(new_ee_contents, eeprom_contents, eesize << 1);
+	if (set_hwaddr) {
+		for (i = 0; i < 3; i++)
+			new_ee_contents[ee_tbl_offset + 10 + i] =
+				(new_hwaddr[i*2]<<8) + new_hwaddr[i*2+1];
+		eeprom_update++;
+	}
+	if (set_ee_rom) {
+		new_ee_contents[ee_tbl_offset + 9] = 0x3001;
+		eeprom_update++;
+	}
+	if (new_default_media >= 0) {
+		new_ee_contents[ee_tbl_offset + 19] &= 0x00ff;
+		new_ee_contents[ee_tbl_offset + 19] |= (new_default_media << 8);
+		eeprom_update++;
+	}
+	if (eeprom_update) {
+		u16 *ee = eeprom_contents + ee_tbl_offset;
+		unsigned short sum = 0;
+		int sum_range = 0x1A, sum_location = 0x20;
+
+		/* Recalculate the checksum:
+		   First find the checksum range and location. */
+		for (i = 0; i < 0x16; i++)
+			sum ^= ee[i];
+		if (((sum ^ (sum>>8)) & 0xff) == ee[0x17]) {
+			sum_range = 0x16;
+			sum_location = 0x17;
+		}
+		for ( ; i < 0x1A; i++)
+			sum ^= ee[i];
+		if (((sum ^ (sum>>8)) & 0xff) == (ee[0x20] & 0xff)) {
+			sum_range = 0x1A;
+			sum_location = 0x20;
+		}			
+		for ( ; i < 0x20; i++)
+			sum ^= ee[i];
+		if (((sum ^ (sum>>8)) & 0xff) == (ee[0x20] & 0xff)) {
+			sum_range = 0x20;
+			sum_location = 0x20;
+		}			
+		sum = 0;
+		for (i = 0; i < sum_range; i++)
+			sum ^= new_ee_contents[ee_tbl_offset + i];
+		new_ee_contents[ee_tbl_offset + sum_location] =
+			(sum ^ (sum>>8)) & 0xff;
+		do_update(ioaddr, eeaddrlen, new_ee_contents, eeprom_contents);
+		for (i = 0; i < eesize; i++)
+			eeprom_contents[i] = read_eeprom(ioaddr, eeaddrlen, i);
+	}
+
+	if (verbose > 2 || show_eeprom > 1) {
+		unsigned short sum = 0;
+		int previous_row_same = 0;
+
+		printf("EEPROM format %dx16, configuration table at offset %#x:\n",
+			   eesize, ee_tbl_offset);
+		for (i = 0; i < eesize; i += 8) {
+			int j;
+			if (i == 0 ||		/* Always show first row. */
+				(memcmp(eeprom_contents+i, eeprom_contents+i - 8, 16) != 0)) {
+				previous_row_same = 0;
+				printf("  %#4.2x:", i);
+				for (j = 0; j < 8; j++) {
+					printf(" %4.4x", eeprom_contents[i + j]);
+					sum += eeprom_contents[i + j];
+				}
+				if (show_eeprom > 2) {
+					printf("  ");
+					for (j = 0; j < 8; j++) {
+						int ew = eeprom_contents[i + j];
+						printf("%c%c",
+							   isprint(ew & 0xff) ? ew & 0xff : '_',
+							   isprint(ew >>   8) ? ew >> 8   : '_' );
+					}
+				}
+				printf("\n");
+			} else if ( ! previous_row_same) {
+				previous_row_same = 1;
+				printf("      ...\n");
+			}
+		}
+		printf("\n The word-wide EEPROM checksum is %#4.4x.\n", sum);
+	}
+
+	/* The user will usually want to see the interpreted EEPROM contents. */
+	if (verbose > 1 || show_eeprom) {
+		parse_eeprom(eeprom_contents);
+	}
+	return;
+}
+
 
 
 /* Read and write the MII registers using software-generated serial
@@ -1015,7 +1232,7 @@ static void mdio_sync(long ioaddr)
 {
 	long mdio_addr = ioaddr + Wn4_PhysicalMgmt;
 	int i;
-	/* Establish sync by sending at least 32 logic ones. */ 
+	/* Establish sync by sending at least 32 logic ones. */
 	for (i = 32; i >= 0; i--) {
 		outw(MDIO_DATA_WRITE1, mdio_addr);
 		mdio_delay(mdio_addr);
@@ -1109,54 +1326,6 @@ void vortex_flash_out(long addr, int offset, int val)
 	outb(val, addr + 8);
 }
 
-
-static void parse_eeprom(unsigned short *eeprom)
-{
-	unsigned char *p = (void *)eeprom;
-	u16 *ee = eeprom;
-	int i, sum = 0;
-
-	printf("Saved EEPROM settings of a 3Com Vortex/Boomerang:\n");
-	if (eeprom[0x37] == 0x6d50) {
-		printf(" The CardBus product ID is %4.4x %4.4x.\n",
-			   eeprom[0], eeprom[1]);
-		p += 0x60;
-		ee += 0x30;
-	}
-	printf(" 3Com Node Address ");
-	for (i = 0; i < 5; i++)
-		printf("%2.2X:", p[i^1]);
-	printf("%2.2X (used as a unique ID only).\n", p[i^1]);
-	printf(" OEM Station address %2.2x", p[1 + 20]);
-	for (i = 1; i < 6; i++)
-		printf(":%2.2X", p[(i^1) + 20]);
-	printf(" (used as the ethernet address).\n");
-	/* The original 3Com layout did not consider post-Y2K production. */
-	printf(" Manufacture date (MM/DD/YYYY) %d/%d/%d, division %c,"
-		   " product %c%c.\n", ((p[8]>>5) & 7) + ((p[9]<<3) & 8),
-		   p[8] & 31, (p[9]>>1) + ((p[9]>>1) < 95 ? 2000 : 1900),
-		   p[10], p[12], p[13]);
-	printf(" Options: %s duplex, link beat %s.\n",
-		   ee[13] & 0x8000 ? "force full" : "negotiated",
-		   ee[13] & 0x4000 ? "check disabled" : "required");
-	for (i = 0; i < 0x16; i++)
-		sum ^= ee[i];
-	printf("  Vortex format checksum is %scorrect (%4.4x vs. %4.4x).\n",
-		   ((sum ^ (sum>>8)) & 0xff) == ee[0x17] ? "" : "in",
-		   (sum ^ (sum>>8)) & 0xff, ee[0x17]);
-	for ( ; i < 0x1A; i++)
-		sum ^= ee[i];
-	sum ^= sum>>8;
-	printf("  Cyclone format checksum is %scorrect (%#2.2x vs. %#2.2x).\n",
-		   (sum  & 0xff) == (ee[0x20] & 0xff) ? "" : "in",
-		   sum & 0xff, ee[0x20] & 0xff);
-	for (sum = 0, i = 0; i < 0x20*2; i++)
-		sum ^= ((unsigned char *)ee)[i];
-	printf("  Hurricane format checksum is %scorrect (%#2.2x vs. %#2.2x).\n",
-		   sum == (ee[0x20] & 0xff) ? "" : "in",
-		   sum, ee[0x20] & 0xff);
-	return;
-}
 
 static void show_dma_state(long ioaddr)
 {
@@ -1165,18 +1334,21 @@ static void show_dma_state(long ioaddr)
 	if (dma_ctrl & 0x0080) {
 		outw(DownStall, ioaddr + EL3_CMD);
 		printf("   DMA control register is %8.8x (during Tx Stall).\n",
-			   inl(ioaddr + 0x20));
+			   (int)inl(ioaddr + 0x20));
 		outw(DownUnstall, ioaddr + EL3_CMD);
 	}
 	printf("   Tx list starts at %8.8x.\n",
-		   inl(ioaddr + 0x24));
+		   (int)inl(ioaddr + 0x24));
 	printf("   Tx FIFO thresholds: min. burst %d bytes, "
 		   "priority with %d bytes to empty.\n",
-		   inb(ioaddr + 0x2A)<<5, inl(ioaddr + 0x2C)<<5);
-	printf("   Tx poll period %d0 ns.\n",
-		   inb(ioaddr + 0x2D)<<5);
-	printf("   Tx maximum burst recorded %d.\n",
-		   inw(ioaddr + 0x78));
+		   inb(ioaddr + 0x2A)<<5, inb(ioaddr + 0x2C)<<5);
+	printf("   Rx FIFO thresholds: min. burst %d bytes, "
+		   "priority with %d bytes to full.\n",
+		   inb(ioaddr + 0x3E)<<5, inb(ioaddr + 0x3C)<<5);
+	printf("   Poll period Tx %d0 ns.,  Rx %d ns.\n",
+		   inb(ioaddr + 0x2D)<<5, inb(ioaddr + 0x3D)<<5);
+	printf("   Maximum burst recorded Tx %d,  Rx %d.\n",
+		   inw(ioaddr + 0x78), inw(ioaddr + 0x7A));
 }
 
 
