@@ -5,7 +5,7 @@
 	  Winbond w89c840
    This file also contains the chip-independent diagnostic code.
 
-   Copyright 1998-2000 by Donald Becker.
+	Copyright 1998-2002 by Donald Becker.
 	This software may be used and distributed according to the terms of
 	the GNU General Public License (GPL), incorporated herein by reference.
 	Contact the author for use under other terms.
@@ -29,10 +29,48 @@
 */
 
 static char *version_msg =
-"winbond-diag.c:v2.00 4/19/2000 Donald Becker (becker@scyld.com)\n"
+"winbond-diag.c:v2.01 12/17/2002 Donald Becker (becker@scyld.com)\n"
 " http://www.scyld.com/diag/index.html\n";
 static char *usage_msg =
-"Usage: etherdiag [-aEefFmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n";
+"Usage: etherdiag [-aEefFGhmqrRtvVwW] [-p <IOport>] [-[AF] <media>]\n"
+"  For details and other options see http://www.scyld.com/diag/index.html\n";
+
+static const char long_usage_msg[] =
+"Usage: %s [-aDfrRvVw] [-AF <speed+duplex>] [-#<BoardNum>]
+
+   Show the internal state of a network adapter.
+
+   The common usage is
+      diag -aem
+
+ Frequently used options are
+   -a  --show_all_registers	Print all registers.
+   -e  --show-eeprom	Dump EEPROM contents, \"-ee\" shows the details.
+   -m  --show_mii		Print the MII transceiver state
+					Using -mm monitors the link.
+   -f  --force		Perform operation, even on a running NIC.
+
+ To operate on a single NIC, or one that hasn't been automatically found:
+   -#  --card_num	INDEX	Operate on the specified card index.
+   -p  --port-base	IOADDR	Assume an adapter at the specified I/O address.
+   -t  --chip-type	TYPE	Specify adapter type with '-p', use '-1' to list.
+
+ To change the persistent EEPROM settings
+   -G  --parameters PARMS	Set adapter-specific parameters.
+   -H  --new-hwaddr 01:23:45:67:ab:cd
+			Set a new hardware station address.  Typically diabled, 
+   -w  --write-EEPROM	 Actually write the new settings into the EEPROM.
+ To read and write the boot BIOS extension Flash ROM
+   -B       Show the first few bytes of the ROM
+   -L FILE  Load the Flash from FILE.
+   -S FILE  Store the Flash image to FILE.
+
+   -D  --debug
+   -v  --verbose	Report each action taken.
+   -V  --version	Emit version information.
+
+   -A  --advertise <speed|setting>  (See the mii-diag manual page.)
+";
 
 #if ! defined(__OPTIMIZE__)
 #warning  You must compile this program with the correct options!
@@ -46,6 +84,8 @@ static char *usage_msg =
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/types.h>
+#include <ctype.h>
 
 #if defined(__linux__)  &&  __GNU_LIBRARY__ == 1
 #include <asm/io.h>			/* Newer libraries use <sys/io.h> instead. */
@@ -53,7 +93,7 @@ static char *usage_msg =
 #include <sys/io.h>
 #endif
 
-/* No libmii.h or libflash.h yet. */
+/* No libmii.h or libflash.h yet, thus the declarations here. */
 extern int show_mii_details(long ioaddr, int phy_id);
 extern int monitor_mii(long ioaddr, int phy_id);
 
@@ -64,17 +104,18 @@ extern int (*flash_in_hook)(long addr, int offset);
 extern void (*flash_out_hook)(long addr, int offset, int val);
 
 /* We should use __u8 .. __u32, but they are not always defined. */
-typedef unsigned int u32;
-typedef unsigned short u16;
-typedef unsigned char u8;
+typedef u_int32_t u32;
+typedef u_int16_t u16;
+typedef u_int8_t u8;
 
 struct option longopts[] = {
  /* { name  has_arg  *flag  val } */
+	{"card-num", 1, 0, '#'},		/* Operate on the specified card index. */
 	{"Advertise", 1, 0, 'A'},
 	{"base-address", 1, 0, 'p'},
 	{"show_all_registers",	0, 0, 'a'},	/* Print all registers. */
-	{"help",	0, 0, 'h'},	/* Give help */
-	{"show-eeprom",  0, 0, 'e'}, /* Dump EEPROM contents (-ee valid). */
+	{"help",	0, 0, 'h'},			/* Print a long usage message. */
+	{"show-eeprom",  0, 0, 'e'},	/* Dump EEPROM contents (-ee valid). */
 	{"emergency-rewrite",  0, 0, 'E'}, /* Re-write a corrupted EEPROM.  */
 	{"force-detection",  0, 0, 'f'},
 	{"new-interface",  1, 0, 'F'},	/* New interface (built-in, AUI, etc.) */
@@ -97,21 +138,21 @@ extern int winbond_diag(int vend_id, int dev_id, long ioaddr, int part_idx);
 
 /* The table of known chips.
    Because of the bogus /proc/pci interface we must have both the exact
-   name and a PCI vendor/device IDs.
+   name from the kernel, a common name and the PCI vendor/device IDs.
    This table is searched in order: place specific entries followed by
    'catch-all' general entries. */
 struct pcidev_entry {
-	char *proc_pci_name;
-	char *part_name;
+	const char *part_name;
+	const char *proc_pci_name;
 	int vendor, device, device_mask;
 	int flags;
 	int io_size;
-	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_num);
+	int (*diag_func)(int vendor_id, int device_id, long ioaddr, int part_idx);
 } pcidev_tbl[] = {
-	{ "Winbond W89c840", "Winbond W89c840",
+	{ "Winbond W89c840", 0,
 	  0x1050, 0x0840, 0xffff, 0, 128, winbond_diag },
-	{ "Compex Unknown device", "Compex RL100-ATX", 0x11F6, 0x2011, 0xffff,
-	  0, 128, winbond_diag },
+	{ "Compex RL100-ATX", "Compex Unknown device",
+	  0x11F6, 0x2011, 0xffff, 0, 128, winbond_diag },
 	{ 0, 0, 0, 0},
 };
 
@@ -131,6 +172,7 @@ static unsigned int opt_flash_show = 0;
 static char	*opt_flash_dumpfile = NULL, *opt_flash_loadfile = NULL;
 
 static unsigned char new_hwaddr[6], set_hwaddr = 0;
+static int emergency_rewrite = 0;
 
 static int scan_proc_pci(int card_num);
 static int parse_media_type(const char *capabilities);
@@ -143,12 +185,11 @@ main(int argc, char **argv)
 {
 	int port_base = 0, chip_type = 0;
 	int errflag = 0, show_version = 0;
-	int emergency_rewrite = 0;
 	int c, longind;
 	int card_num = 0;
 	extern char *optarg;
 
-	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:mp:qrRst:vVwWH:BL:S:",
+	while ((c = getopt_long(argc, argv, "#:aA:DeEfF:G:hH:mp:qrRt:vVwWBL:S:",
 							longopts, &longind))
 		   != -1)
 		switch (c) {
@@ -163,7 +204,8 @@ main(int argc, char **argv)
 			if (new_default_media < 0)
 				errflag++;
 			break;
-		case 'G': opt_G++; opt_GPIO = strtol(optarg, NULL, 16); break;
+		case 'G': opt_G++; opt_GPIO = strtoul(optarg, NULL, 16); break;
+		case 'h': fprintf(stderr, long_usage_msg, argv[0]); return 0;
 		case 'H':
 			{
 				int hwaddr[6], i;
@@ -179,7 +221,7 @@ main(int argc, char **argv)
 			}
 		case 'm': show_mii++;	 break;
 		case 'p':
-			port_base = strtol(optarg, NULL, 16);
+			port_base = strtoul(optarg, NULL, 16);
 			break;
 		case 'q': if (verbose) verbose--;		 break;
 		case 'r': opt_restart++;	break;
@@ -253,7 +295,8 @@ static char bogus_iobase[] =
 static char bogus_irq[] =
 "This chip has not been assigned a valid IRQ, and will not function.\n"
 " This must be fixed in the PCI BIOS setup.  The device driver has no way\n"
-" of changing the PCI IRQ settings.\n";
+" of changing the PCI IRQ settings.\n"
+" See  http://www.scyld.com/expert/irq-conflict.html  for more information.\n";
 
 static int scan_proc_bus_pci(int card_num)
 {
@@ -275,8 +318,9 @@ static int scan_proc_bus_pci(int card_num)
 				   &pci_bus, &pci_devid, &irq, &pciaddr0, &pciaddr1) <= 0)
 			break;
 		for (i = 0; pcidev_tbl[i].vendor; i++) {
-			if (pci_devid !=
-				(pcidev_tbl[i].vendor << 16) + pcidev_tbl[i].device)
+			if ((pci_devid >> 16) != pcidev_tbl[i].vendor
+				|| (pci_devid & pcidev_tbl[i].device_mask) !=
+				pcidev_tbl[i].device)
 				continue;
 			chip_idx = i;
 			card_cnt++;
@@ -334,8 +378,9 @@ static int scan_proc_pci(int card_num)
 					int i;
 					if (debug)
 						printf("Named ethernet controller %s.\n", chip_name);
-					for (i = 0; pcidev_tbl[i].proc_pci_name; i++)
-						if (strncmp(pcidev_tbl[i].proc_pci_name, chip_name,
+					for (i = 0; pcidev_tbl[i].part_name; i++)
+						if (pcidev_tbl[i].proc_pci_name  &&
+							strncmp(pcidev_tbl[i].proc_pci_name, chip_name,
 									strlen(pcidev_tbl[i].proc_pci_name))
 							== 0) {
 							state = 2;
@@ -391,6 +436,7 @@ static int parse_media_type(const char *capabilities)
 		"100baseT4", "100baseTx", "100baseTx-FD", "100baseTx-HD",
 		"10baseT", "10baseT-FD", "10baseT-HD", 0,
 	};
+	char *endptr;
 	int cap_map[] = { 0x0200, 0x0180, 0x0100, 0x0080, 0x0060, 0x0040, 0x0020,};
 	int i;
 	if (debug)
@@ -398,7 +444,8 @@ static int parse_media_type(const char *capabilities)
 	for (i = 0; mtypes[i]; i++)
 		if (strcasecmp(mtypes[i], capabilities) == 0)
 			return cap_map[i];
-	if ((i = strtol(capabilities, NULL, 16)) <= 0xffff)
+	i = strtoul(capabilities, &endptr, 16);
+	if (*endptr == 0  &&  0 < i  &&  i <= 0xffff)
 		return i;
 	fprintf(stderr, "Invalid media advertisement '%s'.\n", capabilities);
 	return 0;
@@ -429,16 +476,20 @@ struct { char *name; int value; } mediamap[] = {
 
 static int get_media_index(const char *name)
 {
+	char *endptr;
 	int i;
+
+	if (! name)
+		return -1;
 	for (i = 0; mediamap[i].name; i++)
 		if (strcasecmp(name, mediamap[i].name) == 0)
 			return i;
-	if (name  &&  atoi(name) >= 00)
-		return atoi(name);
-	fprintf(stderr, "Invalid interface specified: it must be one of\n  ");
+	i = strtol(name, &endptr, 0);
+	if (*endptr == 0)
+		return i;
+	fprintf(stderr, "Invalid interface specified.  It must be one of\n");
 	for (i = 0; mediamap[i].name; i++)
-		fprintf(stderr, "  %s", mediamap[i].name);
-	fprintf(stderr, ".\n");
+		fprintf(stderr, "  %d %s\n", mediamap[i].value, mediamap[i].name);
 	return -1;
 }
 
